const { getBrowser } = require('../util/puppeteerUtil');
const { authenticateWithEscapia,
        setAndGetConfigEscapia,
        logoutOfEscapia,
        searchForUserByEmail } = require('../util/escapiaUtil');
const { sleep, resultsEnum } = require('../util/misc');
const logger = require('@turnkeyvr/logger');
const { PendingXHR } = require('pending-xhr-puppeteer');
const csv = require('fast-csv');
const fs = require('fs');
const path = require('path');
const createCsvWriter = require('csv-writer').createObjectCsvWriter;
const eachSeries = require('async/eachSeries');

//const csvFileName = 'Non-Authorized ACH Withdrawals 2019 - November 2019.csv'; // temp for testing
const csvFileName = 'test.csv'; // temp for testing

const loadCSV = async () => {
    let records = [];
    return new Promise((resolve, reject) => {
        let csvPath = path.resolve(__dirname, './files/' + csvFileName);
        logger.debug('csvPath: ', csvPath);
        fs.createReadStream(csvPath)
            .pipe(csv.parse({headers: true}))
            .on('data', row => {
                records.push(row);
            })
            .on('error', error => {
                logger.error(error);
                reject();
            })
            .on('end', () => {
                resolve(records);
            });
    });
};

const getCSVRecordHeaders = csvRecord => {
    return Object.keys(csvRecord).map(key => key);
};

// technically only Email + New Balance are required, but if the others are missing that indicates a potential problem
const validateCSVHeaders = csvHeaders => {
    let requiredHeaders = ['Email', 'New Balance', 'Balance Pulled', 'Reference #', 'Notes'];
    for (const header of requiredHeaders) {
        if (!csvHeaders.includes(header)) {
            return false;
        }
    }
    return true;
};

const validateCSVRecord = csvRecord => {
    return csvRecord['Email'] && 
        csvRecord['Email'].trim() !== '' && 
        csvRecord['New Balance'] && 
        csvRecord['New Balance'].trim() !== '' && 
        csvRecord['New Balance'].includes('-') && 
        csvRecord['New Balance'].includes('$') && 
        csvRecord['New Balance'].includes('.');
};

const searchForUserByEmailTest = async (page, email, role) => {
    // open contact search page
    let url = role ? 'https://turnkeyvr.escapia.com/WebAgency/FindContacts/Find?Role=' + role : 'https://turnkeyvr.escapia.com/WebAgency/FindContacts/Find?Role=None';
    await page.goto(url, {waitUntil: 'networkidle0'});

    // input email
    await page.focus('#txtEmail');
    await page.keyboard.type(email);

    // click search button
    await page.click('#butSearch', {waitUntil: 'networkidle0'});

    await sleep(5000); // terrible, bullshit hack; need to replace this

    // click row
    let rowCount = await page.evaluate(() => {
        let elements = document.getElementsByClassName('clickable');
        if (!elements.length || elements.length !== 1) {
            return elements.length;
        }
        elements[0].children[1].click();
        return 1;
    });

    // error handling for no results or multiple results found; overridden to test for JB
    if (rowCount !== 1) {
        if (email.toLowerCase() === 'john.banczak@gmail.com') {
            page.evaluate(() => {
                let elements = document.getElementsByClassName('clickable');
                elements[1].children[1].click();
            });
        }
        else {
            return resultsEnum.FAILURE;
        }
    }

    return page;
};

const processACHWithdrawal = async (record, browser, page, csvWriter) => {
    // validate csv record
    if (!validateCSVRecord(record)) {
        record['Notes'] = 'Robot error: could not validate csv record (missing information)';
        await csvWriter.writeRecords(new Array(record));
        return;
    }

    let ownerData = {
        firstName: null,
        lastName: null,
        payChecksTo: null,
        routingNumber: null,
        accountNumber: null,
        paymentAmount: record['New Balance'].replace('-', '').replace('$', '').trim()
    };

    // open contact to retrieve data
    let ownerSearchResults = await searchForUserByEmailTest(page, record['Email'], 'Owner'); // 'brinkmanclan@comcast.net'
    if (ownerSearchResults === resultsEnum.FAILURE) {
        record['Notes'] = 'Robot error: could not find owner (either no search results or too many results)';
        await csvWriter.writeRecords(new Array(record));
        return;
    }
    else {
        await page.waitForNavigation();
    }

    // get first and last name as saved in escapia
    let firstNameInput = await page.$('#ctlContactWizard_ctl02_txtNameFirst');
    let firstNameValueHandle = await firstNameInput.getProperty('value');
    ownerData.firstName = (await firstNameValueHandle.jsonValue()).trim();

    let lastNameInput = await page.$('#ctlContactWizard_ctl02_txtNameLast');
    let lastNameValueHandle = await lastNameInput.getProperty('value');
    ownerData.lastName = (await lastNameValueHandle.jsonValue()).trim();    


    // click "Owner" tab
    await page.click('a[href="#tab-1"]');

    await sleep(10000);

    // using puppeteer jshandles to scrape required values for processing
    let payChecksToInput = await page.$('#ctlContactWizard_ctl04_txtChecksPayableTo');
    let payChecksToValueHandle = await payChecksToInput.getProperty('value');
    ownerData.payChecksTo = (await payChecksToValueHandle.jsonValue()).trim();
    if (!ownerData.payChecksTo || ownerData.payChecksTo === '') {
        ownerData.payChecksTo = ownerData.firstName + ' ' + ownerData.lastName;
    }

    let routingNumberInput = await page.$('#ctlContactWizard_ctl04_txtBankRoutingNumber');
    let routingNumberValueHandle = await routingNumberInput.getProperty('value');
    ownerData.routingNumber = (await routingNumberValueHandle.jsonValue()).trim();

    let accountNumberInput = await page.$('#ctlContactWizard_ctl04_txtBankAccountNumber');
    let accountNumberValueHandle = await accountNumberInput.getProperty('value');
    ownerData.accountNumber = (await accountNumberValueHandle.jsonValue()).trim();

    logger.debug(ownerData);

    // validate owner bank info
    if (!ownerData.routingNumber || ownerData.routingNumber === '' || !ownerData.accountNumber || ownerData.accountNumber === '') {
        record['Notes'] = 'Error: missing owner bank info (routing number and/or account number)';
        logger.error(record['Notes']);
        await csvWriter.writeRecords(new Array(record));
        return;
    }

    // open receive payment page
    await page.goto('https://turnkeyvr.escapia.com/webagency/Accounting/AccountsReceivable/ReceivePayment.aspx', {waitUntil: 'networkidle0'});

    // set received from owner
    await page.select('#ctlReceivePayment_lstSubledgerType', '1');
    await page.waitForSelector('#ctlReceivePayment_txtSubledgerEntityName', {visible: true});

    // set payment type to eCheck
    await page.select('#ctlReceivePayment_lstPaymentType', '5');
    await page.waitForSelector('#ctlReceivePayment_txtSubledgerEntityName', {visible: true});

    // do search
    let searchComplete = false;
    let popupSearch = async target => {
        if (target.url() === 'https://turnkeyvr.escapia.com/webagency/Accounting/SelectSubledgerEntity.aspx?SubledgerType=1&ParentIDControl=ctlReceivePayment_txtSubledgerEntityID&ParentNameControl=ctlReceivePayment_txtSubledgerEntityName&isPopupDialog=1') {

            // do search
            let popupPage = await target.page();

            await popupPage.waitForSelector('#inpLName', {visible: true});

            let pendingXHR = new PendingXHR(popupPage);
            
            // input last name
            await popupPage.focus('#inpLName');
            await popupPage.keyboard.type(ownerData.lastName);

            // input first name
            await popupPage.focus('#inpFName');
            await popupPage.keyboard.type(ownerData.firstName);

            // wait for xhr requests to finish
            await pendingXHR.waitForAllXhrFinished();

            await sleep(10000);

            // click row
            let rowCount = await popupPage.evaluate(() => {
                // there are always 10 rows in the table, but some may not be visible; need to filter those out
                let elements = document.getElementsByClassName('clickable');
                elements = Array.from(elements); // convert HTMLCollection to array

                // early exit if no results for some reason (although as previous stated, there should always be 10 rows)
                if (!elements.length) {
                    return elements.length;
                }

                elements = elements.filter(td => td.parentElement.style.display !== 'none');

                // TODO: if multiple records can attempt to determine the correct element; not sure best way to proceed
                // early exit if no visible results or multiple visible results
                if (!elements.length || elements.length !== 1) {
                    //return elements.length; // temp commented out to test JB!
                }

                elements[0].click();
                return 1;
            });

            // error handling for no results or multiple results found
            if (rowCount !== 1) {
                logger.error('error on rowCount: ', rowCount);
                let errorMessage = rowCount === 0 ? 'No search results found' : 'Too many search results found';
                throw new Error(errorMessage);
            }
            else {
                logger.debug('clicked');
            }

            searchComplete = true;
        }
    };

    // add event listener
    browser.on('targetcreated', popupSearch);

    await page.click('#ctlReceivePayment_txtSubledgerEntityName');

    while (!searchComplete) {
        await sleep(2000);
    }
    
    // remove listener
    browser.removeListener('targetcreated', popupSearch);

    await page.waitForSelector('#ctlReceivePayment_txtBankAccountFullName', {visible: true});

    // fill in account holder name
    await page.focus('#ctlReceivePayment_txtBankAccountFullName');
    await page.keyboard.type(ownerData.payChecksTo);

    // fill in bank account number
    await page.focus('#ctlReceivePayment_txtBankAccountNumber');
    await page.keyboard.type(ownerData.accountNumber);

    // temp commented out for dialog popup testing
    // fill in routing number
    await page.focus('#ctlReceivePayment_txtBankRoutingNumber');
    await page.keyboard.type(ownerData.routingNumber);

    // temp commented out for testing
    // fill in payment amount
    await page.focus('#ctlReceivePayment_txtAmount');
    await page.keyboard.type(ownerData.paymentAmount);

    // uncheck "save bank account information"
    await page.$eval('#ctlReceivePayment_chkSaveAccountNumber', c => c.checked = false);

    // capture alert if it happens
    let alertDetected = false;
    let alertMessage = '';
    let captureAlerts = async dialog => {
        alertDetected = true;
        alertMessage = dialog.message();
        logger.debug('alert on save: ', alertMessage);
        await dialog.dismiss();
    };
    page.on('dialog', captureAlerts);

    // click save
    await page.click('#btnSave', {waitUntil: 'networkidle0'}); // temp commented out for testing

    // wait for loading icon to appear and disappear; alert is captured during this time if it appears
    // todo: I think it won't appear if form invalid, test
    try {
        await page.waitForSelector('#divPleaseWait', {visible: true});
        await page.waitForSelector('#divPleaseWait', {hidden: true});  
    }
    catch (e) {
        if (!alertDetected) {
            record['Notes'] = 'Error: payment may or may not have been submitted, manual review required';
            logger.error(record['Notes']);
            await csvWriter.writeRecords(new Array(record));
            return;
        }
    }

    page.removeListener('dialog', captureAlerts);

    if (alertDetected && alertMessage.toLowerCase() !== 'updating payment account failed due to db timeout or deadlock, try again later.') {
        // log this as error and cancel for now
        logger.debug('cancel!');
        await page.click('#btnCancel', {waitUntil: 'networkidle0'});
        record['Notes'] = alertMessage;
        await csvWriter.writeRecords(new Array(record));
        return;
    }

    //await sleep(10000);


    // todo: get reference number
    // search again for user
    await searchForUserByEmailTest(page, record['Email'], 'Owner');
    await page.waitForNavigation();

    // click "Owner Trans." tab
    await page.click('a[href="#tab-5"]');

    // click "Search" button
    await page.click('#ctlContactWizard_ctl12_ctlSubledgerEntityTransactionFinder_btnSearch', {waitUntil: 'networkidle0'});

    // huge PITA: need to iterate though all table rows and find the credit
    let referenceNumber = await page.evaluate((paymentAmount) => {
        let elements = document.getElementsByClassName('listData num');
        if (!elements.length) {
            return null;
        }

        let chosenTD = null;
        for (let i = 0; i < elements.length; i++) {
            let td = elements[i];
            if (!td.textContent || td.textContent.trim() === '') {
                continue;
            }
            if (td.textContent.trim() !== paymentAmount) {
                continue;
            }
            // if another TD with the same value found, early exit as error, manual intervention required
            if (chosenTD) {
                return null;
            }
            chosenTD = td;
        }

        if (!chosenTD) {
            return null;
        }

        // scrape reference number value
        let rowElement = chosenTD.parentElement;
        let paymentAmountTDIndex = null;
        for (let i = 0; i < rowElement.children.length; i++) {
            let td = rowElement.children[i];
            if (!td.textContent || td.textContent.trim() !== paymentAmount) {
                continue;
            }
            paymentAmountTDIndex = i;
            break;
        }

        if (!paymentAmountTDIndex) {
            return null;
        }

        return rowElement.children[paymentAmountTDIndex - 1].textContent.trim();
    }, ownerData.paymentAmount);

    if (!referenceNumber) {
        record['Notes'] = 'Error: could not retrieve reference number after posting payment';
        logger.error(record['Notes']);
        await csvWriter.writeRecords(new Array(record));
        return;
    }

    await sleep(10000); // todo: is this necessary?

    // set record values;
    record['Reference #'] = referenceNumber;
    record['Notes'] = 'Withdrawn';
    record['Balance Pulled'] = '$' + ownerData.paymentAmount;

    // append results to file
    await csvWriter.writeRecords(new Array(record));
};

const handler = async () => {
    let records = await loadCSV();
    logger.debug('load CSV complete');
    logger.debug(records.length);
    logger.debug(records[0]);

    if (records.length === 0) {
        logger.error('No records to process');
        return;
    }

    let csvHeaders = getCSVRecordHeaders(records[0]);

    if (!validateCSVHeaders(csvHeaders)) {
        logger.error('CSV not in expected format (missing/incorrect headers)');
        return;
    }

    // create csv
    const csvWriter = createCsvWriter({
        path: './ach/files/' + csvFileName.replace('.csv', ' - RESULTS.csv'),
        header: csvHeaders.map(header => {
            return {id: header, title: header};
        })
    });

    // load browser
    let browser = await getBrowser();
    if (!browser) {
        return;
    }

    // authenticate with escapia
    setAndGetConfigEscapia('1655'); // only running against the main instance
    let page = await authenticateWithEscapia(await browser.newPage());
    if (!page) {
        await browser.close();
        return;
    }

    // core processing loop
    await eachSeries(records, async record => await processACHWithdrawal(record, browser, page, csvWriter));

    // logout of escapia then exit browser
    await logoutOfEscapia(page);

    await browser.close();
    logger.debug('done');

    return;
};

exports.handler = handler;