// override logger settings
process.env.NODE_CONFIG = JSON.stringify({
    logger: {
        name: 'airbnb-rate-reconciliation',
        transports: {
            file: false,
            console: false,
            papertrail: {
                level: 'verbose'
            }
        },
    },
});

const { getBrowser } = require('../util/puppeteerUtil');
const { authenticateWithEscapia,
        setAndGetConfigEscapia,
        applyHostFeeCreditToFolio,
        handleAirbnbEscapiaPercentComparison,
        removeGuestPhoneNumberFromReservationNote,
        removePendingReconciliationReservationNote,
        getReservationNumberFromReservation,
        getIsCancelledFromReservation,
        checkForPendingReconciliationReservationNote,
        logoutOfEscapia2, 
        handleTaxEdits } = require('../util/escapiaUtil');
const { getRateReconciliationCases, 
        getEscapiaUnitReservationIdFromCaseDescription, 
        getEscapiaQuotePercentComparisonFromCaseDescription, 
        getAirbnbAmountFromCaseDescription,
        getListingHostFeeFromCaseDescription,
        getGuestPhoneNumberFromCaseDescription,
        determineRequiredTaxEdits,
        updateCase,
        chatterNew } = require('../util/sfUtil');
const { getAirbnbAccountInfoFromCase,
        getMessageThreadIdFromReservation,
        sendMessage } = require('../util/airbnbUtil');
const { sleep, resultsEnum } = require('../util/misc');
const eachSeries = require('async/eachSeries');
const logger = require('@turnkeyvr/logger');
const https = require('https');
const axios = require('axios');
const MongoDb = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('rate-reconciliation-prod');
const moment = require('moment');

const isProd = (process.env.NODE_ENV.toLowerCase() === 'prod' || process.env.NODE_ENV.toLowerCase() === 'production');

const airbnbMessage = 'Thanks for booking! What is your direct email address? '
    + 'We use that address to send the arrival info and door code via email about 10 days prior to arrival, '
    + 'as well as a link to the TurnKey Guest Portal which may include some additional steps required to finalize the booking with TurnKey.'
    + '\n\n'
    + 'Feel free to call us with any questions (24/7) at 888-512-0498 or SMS us through the SMS message you will receive from us shortly.'
    + '\n\n'
    + 'Thanks,'
    + '\n\n'
    + 'The TurnKey Team';

const handler = async () => {
    // only runs in production
    if (!isProd) {
        logger.debug('rate reconciliation only runs in production');
        return;
    }

    while (true) {
        logger.info('Airbnb Rate Reconciliation is running'); // used for monitoring

        let cases = await getRateReconciliationCases('airbnb');

        // no results to process; early exit
        if (!cases || cases.length === 0) {
            logger.info('No cases to process');
            await sleep(1000 * 60 * 10); // 10 minutes
            continue;
        }

        // load browser
        let browser = await getBrowser();
        if (!browser) {
            await sleep(1000 * 60 * 10); // 10 minutes
            continue;
        }

        // authenticate with escapia
        setAndGetConfigEscapia('1655'); // only running against the main instance
        let page = await authenticateWithEscapia(await browser.newPage());
        if (!page) {
            await browser.close();
            await sleep(1000 * 60 * 10); // 10 minutes
            continue;
        }

        // mark all cases as in progress
        let inProgressCases = [];
        await eachSeries(cases, async (caseRecord) => {
            // mark the Case as "In Progress"
            let inProgressResults = await updateCase({
                Id: caseRecord.Id,
                Status: 'In Progress'
            });
            if (inProgressResults === resultsEnum.FAILURE) {
                await handleCaseProcessingFailure(caseRecord, 'Failed to mark Case ' + caseRecord.Id + ' as In Progress');
            }
            else {
                inProgressCases.push(caseRecord);
            }
        });

        // process each case
        await eachSeries(inProgressCases, async (caseRecord) => await processCase(browser, page, caseRecord));

        try {
            logger.info('starting logout of escapia');
            if (page && !page.isClosed()) {
                await logoutOfEscapia2(page);
            }

            logger.info('logged out of escapia'); // added for debugging

            logger.info('starting close of browser');
            if (browser && browser.isConnected()) {
                await browser.close();
            }

            logger.info('browser closed'); // added for debugging
        }
        catch (e) {
            logger.error(`logout failed: ${e.message}`);
        }
    }
};

const processCase = async (browser, page, caseRecord) => {
    try {
        logger.info('Airbnb Rate Reconciliation is running'); // used for monitoring
        logger.info('caseRecord Id: ', caseRecord.Id);

        let processCaseSuccess = resultsEnum.SUCCESS; // default to success
        let processCaseNotes = '';

        // if it's a "Marketing Only" Case, let Smiles process it -- higher likelihood of having a security deposit
        if (caseRecord.OpportunityCase__r.Is_Marketing_Only__c) {
            await handleCaseProcessingFailure(caseRecord, 'Robotify does not process Cases for marketing-only Rentals');
            return;
        }

        let dataResults = await getRequiredDataToProcessCase(page, caseRecord);
        if (dataResults.success === resultsEnum.FAILURE) {
            await handleCaseProcessingFailure(caseRecord, 'Failed to retrieve data for Case ' + caseRecord.Id + ': ' + dataResults.notes);
            return;
        }
        logger.debug(dataResults.data);

        // if the reservation in Escapia is cancelled, flag it as an error
        if (dataResults.data.isCancelled) {
            await handleCaseProcessingFailure(caseRecord, 'Reservation is cancelled in Escapia');
            return;
        }

        let { escapiaUnitReservationId, guestPhoneNumber, listingHostFee, airbnbAmountPlusListingHostFee, escapiaQuotePercent, airbnbAccountInfo, airbnbMessageThreadId, reservationNumber } = dataResults.data;

        // if the reservation in Escapia doesn't have a pending reconiliation note, flag it for human review
        let pendingReconciliationNotePresentResults = await checkForPendingReconciliationReservationNote(page, escapiaUnitReservationId)
        if (pendingReconciliationNotePresentResults === resultsEnum.FAILURE) {
            await handleCaseProcessingFailure(caseRecord, 'No "Pending Reconciliation" note Escapia, please review');
            return;
        }

        // handle tax edits
        let handleTaxEditsResults = await handleTaxEdits(browser, page, escapiaUnitReservationId, determineRequiredTaxEdits(caseRecord));
        logger.debug('handleTaxEditsResults', handleTaxEditsResults);
        if (handleTaxEditsResults === resultsEnum.FAILURE) {
            await handleCaseProcessingFailure(caseRecord, 'Tax edits for Case ' + caseRecord.Id + ' failed');
            return;
        }
        else if (handleTaxEditsResults === resultsEnum.REVIEW_REQUIRED) {
            processCaseSuccess = resultsEnum.REVIEW_REQUIRED;
            processCaseNotes += '- Tax edits for Case ' + caseRecord.Id + ' require review\n';
            logger.warn('Tax edits for Case ' + caseRecord.Id + ' require review');
        }
        else {
            processCaseNotes += '- Tax edits complete\n';
        }
        
        // handle escapia % comparison from case (also removes security deposit if present)
        let handleAirbnbEscapiaPercentComparisonResults = await handleAirbnbEscapiaPercentComparison(browser, page, escapiaUnitReservationId, escapiaQuotePercent, airbnbAmountPlusListingHostFee);
        if (handleAirbnbEscapiaPercentComparisonResults === resultsEnum.FAILURE) {
            await handleCaseProcessingFailure(caseRecord, 'Unable to handle Airbnb + Escapia percent comparison logic for Case ' + caseRecord.Id);
            return;
        }
        processCaseNotes += '- Airbnb/Escapia percent comparison logic complete\n';

        // apply the host fee as a credit to the reservation folio
        let applyHostFeeCreditToFolioResults = await applyHostFeeCreditToFolio(page, escapiaUnitReservationId, listingHostFee);
        if (applyHostFeeCreditToFolioResults === resultsEnum.FAILURE) {
            await handleCaseProcessingFailure(caseRecord, 'Unable to apply host fee credit to folio for Case ' + caseRecord.Id);
            return;
        }
        processCaseNotes += '- Host fee applied as a credit to the reservation folio\n';

        // message the guest via airbnb
        let airbnbMessageResults = await sendMessage(airbnbAccountInfo, airbnbMessageThreadId, airbnbMessage);
        if (airbnbMessageResults === resultsEnum.FAILURE) {
            await handleCaseProcessingFailure(caseRecord, 'Airbnb message sending failed for Case ' + caseRecord.Id);
            return;
        }
        processCaseNotes += '- Guest messaged on Airbnb\n';

        // remove the guest's phone number from escapia reservation notes
        let removeGuestPhoneNumberFromReservationNoteResults = await removeGuestPhoneNumberFromReservationNote(page, escapiaUnitReservationId, guestPhoneNumber);
        if (removeGuestPhoneNumberFromReservationNoteResults === resultsEnum.FAILURE) {
            await handleCaseProcessingFailure(caseRecord, 'Unable to remove guest phone number from Escapia Reservation notes for Case ' + caseRecord.Id);
            return;
        }

        // remove the pending reconciliation note from escapia reservation notes
        let removePendingReconciliationReservationNoteResults = await removePendingReconciliationReservationNote(page, escapiaUnitReservationId);
        if (removePendingReconciliationReservationNoteResults === resultsEnum.FAILURE) {
            await handleCaseProcessingFailure(caseRecord, 'Unable to remove "Pending Reconciliation" note from Escapia Reservation notes for Case ' + caseRecord.Id);
            return;
        }
        processCaseNotes += '- "Pending Reconciliation" note removed from reservation in Escapia\n';

        // make sure pending reconciliation note was remove from escapia reservation
        // if the reservation in Escapia doesn't have a pending reconiliation note, flag it for human review
        let pendingReconciliationNotePresentFinalResults = await checkForPendingReconciliationReservationNote(page, escapiaUnitReservationId)
        if (pendingReconciliationNotePresentFinalResults !== resultsEnum.FAILURE) {
            await handleCaseProcessingFailure(caseRecord, 'Unable to remove "Pending Reconciliation" note from Escapia Reservation notes for Case ' + caseRecord.Id);
            return;
        }

        // add case comment
        let chatterData = { 
            body: {
                messageSegments: [{
                    type: 'Text',
                    text: 'INT Completed ' + reservationNumber
                }]
            },
            feedElementType: 'FeedItem',
            subjectId: caseRecord.Id
        };
        let chatterResults = await chatterNew(chatterData);
        if (chatterResults === resultsEnum.FAILURE) {
            await handleCaseProcessingFailure(caseRecord, 'Failed to leave chatter message on Case ' + caseRecord.Id);
            return;
        }
        processCaseNotes += '- Chatter comment added to Case\n';

        // update the case for success/review required
        let caseUpdateRecord = {
            Id: caseRecord.Id
        };
        if (processCaseSuccess === resultsEnum.REVIEW_REQUIRED) {
            caseUpdateRecord.Subject = caseRecord.Subject += ' - - REVIEW REQUIRED';
            caseUpdateRecord.Description = caseRecord.Description += '\n\n##### ROBOTIFY NOTES #####\n' + processCaseNotes;
            caseUpdateRecord.Status = 'New';
        }
        else {
            caseUpdateRecord.Description = caseRecord.Description += '\n\n##### ROBOTIFY NOTES #####\n' + processCaseNotes;
            caseUpdateRecord.Status = 'Closed';

            // add mongo record for later recheck logic
            let newRecord = {
                reconciliationDate: moment().toDate(),
                escapiaUnitReservationId: escapiaUnitReservationId,
                sfdcCaseId: caseRecord.Id,
                attempts: 0,
                source: 'Airbnb'
            };
            await MongoDb.insertMany('recheck', [newRecord]);
        }
        await updateCase(caseUpdateRecord);
    }
    catch(e) {
        logger.error(`Process individual case exception: ${JSON.stringify(e)}`);

        // attempt to update the case as failed
        await handleCaseProcessingFailure(caseRecord, 'Generic error for Case Id: ' + caseRecord.Id);
    }
};

const getRequiredDataToProcessCase = async (page, caseRecord) => {
    let results = {
        success: resultsEnum.SUCCESS,
        notes: '',
        data: {
            escapiaUnitReservationId: null,
            guestPhoneNumber: null,
            airbnbAmount: null,
            listingHostFee: null,
            airbnbAmountPlusListingHostFee: null,
            escapiaQuotePercent: null,
            airbnbAccountInfo: null,
            airbnbMessageThreadId: null,
            reservationNumber: null,
            isCancelled: false
        }
    };

    results.data.escapiaUnitReservationId = getEscapiaUnitReservationIdFromCaseDescription(caseRecord.Description);
    if (!results.data.escapiaUnitReservationId) {
        results.success = resultsEnum.FAILURE;
        results.notes += 'Unable to extract escapiaUnitReservationId from Case Description';
        return results;
    }

    results.data.guestPhoneNumber = getGuestPhoneNumberFromCaseDescription(caseRecord.Description);
    if (!results.data.guestPhoneNumber) {
        results.success = resultsEnum.FAILURE;
        results.notes += 'Unable to extract guest phone number from Case Description';
        return results;
    }

    results.data.airbnbAmount = getAirbnbAmountFromCaseDescription(caseRecord.Description);
    if (!results.data.airbnbAmount) {
        results.success = resultsEnum.FAILURE;
        results.notes += 'Unable to retrieve airbnbAmount from Case Description';
        return results;
    }

    results.data.listingHostFee = getListingHostFeeFromCaseDescription(caseRecord.Description);
    if (!results.data.listingHostFee) {
        results.success = resultsEnum.FAILURE;
        results.notes += 'Unable to retrieve listingHostFee from Case Description';
        return results;
    }

    results.data.airbnbAmountPlusListingHostFee = results.data.airbnbAmount + results.data.listingHostFee;
    results.data.airbnbAmountPlusListingHostFee = Number(Math.round(results.data.airbnbAmountPlusListingHostFee + 'e2') + 'e-2');

    results.data.escapiaQuotePercent = getEscapiaQuotePercentComparisonFromCaseDescription(caseRecord.Description);
    if (!results.data.escapiaQuotePercent) {
        results.success = resultsEnum.FAILURE;
        results.notes += 'Unable to retrieve escapiaQuotePercent from Case Description';
        return results;
    }

    results.data.airbnbAccountInfo = getAirbnbAccountInfoFromCase(caseRecord);
    if (!results.data.airbnbAccountInfo) {
        results.success = resultsEnum.FAILURE;
        results.notes += 'Unable to retrieve airbnbAccountInfo from Case';
        return results;
    }

    results.data.airbnbMessageThreadId = await getMessageThreadIdFromReservation(caseRecord);
    if (!results.data.airbnbMessageThreadId) {
        results.success = resultsEnum.FAILURE;
        results.notes += 'Unable to retrieve airbnbMessageThreadId';
        return results;
    }

    results.data.reservationNumber = await getReservationNumberFromReservation(page, results.data.escapiaUnitReservationId);
    if (!results.data.reservationNumber) {
        results.success = resultsEnum.FAILURE;
        results.notes += 'Unable to retrieve reservationNumber';
        return results;
    }

    results.data.isCancelled = await getIsCancelledFromReservation(page, results.data.escapiaUnitReservationId);

    return results;
};

const handleCaseProcessingFailure = async (caseRecord, errorMessage) => {
    logger.error(errorMessage);

    let caseUpdateRecord = {
        Id: caseRecord.Id,
        Status: 'New',
        Subject: caseRecord.Subject + ' - - ROBOTIFY FAILURE',
        Description: caseRecord.Description + '\n\n##### ROBOTIFY NOTES #####\n' + errorMessage
    };

    await updateCase(caseUpdateRecord);
};

// just a test for alerts
const alertTest = () => {
    logger.debug('robotify alert test');
    // get ip
    https.get('https://api.ipify.org?format=json', (resp) => {
        let data = '';

        // A chunk of data has been recieved.
        resp.on('data', (chunk) => {
            data += chunk;
        });

        // The whole response has been received. Print out the result.
        resp.on('end', () => {
            logger.debug(JSON.parse(data));
            logger.debug('fin');
        });

    }).on("error", (err) => {
        console.log("Error: " + err.message);
    });
};

const getTest = async () => {
    try {
        let response = await axios.get('https://api.ipify.org?format=json');
        console.log('outboundIPAddress: ', response);
        let results = {
            test: 'test success',
            outboundIPAddress: response.data
        };
        return {
            statusCode: 200,
            headers: {
                "Access-Control-Allow-Origin": "*",
                "Access-Control-Allow-Credentials": true
            },
            body: JSON.stringify(results)
        };
    }
    catch (e) {
        console.log(e);
        return {
            statusCode: 500,
            headers: {
                "Access-Control-Allow-Origin": "*",
                "Access-Control-Allow-Credentials": true
            },
            body: JSON.stringify(e.message)
        };
    }
};


exports.handler = handler;
exports.alertTest = alertTest;
exports.getTest = getTest;