// override logger settings
process.env.NODE_CONFIG = JSON.stringify({
    logger: {
        name: 'coupon-code-sync',
        transports: {
            file: false,
            console: false,
            papertrail: {
                level: 'verbose'
            }
        },
    },
});

const moment = require('moment');
const logger = require('@turnkeyvr/logger');
const MongoDbCouponCodes = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('coupon-codes-prod');
const { customRequestSF } = require('./util/sfdc');
const { resultsEnum, sleep } = require('../util/misc');

// scraping
const scrapeSeasons = require('./scraping/scrapeSeasons');
const scrapeUnitGroups = require('./scraping/scrapeUnitGroups');
const scrapeCouponCodesAndPricingPlans = require('./scraping/scrapeCouponCodesAndPricingPlans');
const scrapeRateRules = require('./scraping/scrapeRateRules');

// processing
const processSeasons = require('./processing/processSeasons');
const processUnitGroups = require('./processing/processUnitGroups');
const processCouponCodesAndPricingPlans = require('./processing/processCouponCodesAndPricingPlans');
const processRateRules = require('./processing/processRateRules');

// syncing
const sync = require('./syncing/sync');

const syncCouponCodes = async () => {
    logger.verbose('COUPON CODE SYNC START');

    let today = moment().format('MM-DD-YYYY');
    let syncVersion = 1;
    let syncOperationsCollectionName;
    let escapiaCollectionName;
    let sfdcCollectionName;
    
    // create mongo sync collections
    try {
        // determine sync version for the day
        let mongoConn = await MongoDbCouponCodes.connect();
        let dbo = mongoConn.db();
        let existingCollectionNames = (await dbo.collections()).map(c => c.collectionName);
        while (existingCollectionNames.includes(`coupon-code-sync-operations-${today}-v${syncVersion}`)) {
            syncVersion += 1;
        }

        syncOperationsCollectionName = `coupon-code-sync-operations-${today}-v${syncVersion}`;
        let syncOperationsCollection = await dbo.createCollection(syncOperationsCollectionName);
        await syncOperationsCollection.createIndex('action', { unique: false });
        
        escapiaCollectionName = `coupon-code-sync-escapia-${today}-v${syncVersion}`;
        let escapiaCollection = await dbo.createCollection(escapiaCollectionName);
        await escapiaCollection.createIndex('sObjectType', { unique: false });
        
        sfdcCollectionName = `coupon-code-sync-sfdc-${today}-v${syncVersion}`;
        let sfdcCollection = await dbo.createCollection(sfdcCollectionName);
        await sfdcCollection.createIndex('sObjectType', { unique: false });

        mongoConn.close();
    }
    catch (e) {
        logger.error('Failed to create mongo sync collections');
        return resultsEnum.FAILURE;
    }

    // scrape seasons from escapia
    let scrapeSeasonsResult = await scrapeSeasons(escapiaCollectionName);
    if (scrapeSeasonsResult !== resultsEnum.SUCCESS) {
        logger.error('Scraping Seasons failed');
        return resultsEnum.FAILURE;
    }

    // process seasons
    let processSeasonsResult = await processSeasons(escapiaCollectionName, sfdcCollectionName, syncOperationsCollectionName);
    if (processSeasonsResult !== resultsEnum.SUCCESS) {
        logger.error('Processing Seasons failed');
        return resultsEnum.FAILURE;
    }

    // scrape unit groups from escapia
    let scrapeUnitGroupsResult = await scrapeUnitGroups(escapiaCollectionName);
    if (scrapeUnitGroupsResult !== resultsEnum.SUCCESS) {
        logger.error('Scraping Unit Groups failed');
        return resultsEnum.FAILURE;
    }

    // process unit groups
    let processUnitGroupsResult = await processUnitGroups(escapiaCollectionName, sfdcCollectionName, syncOperationsCollectionName);
    if (processUnitGroupsResult !== resultsEnum.SUCCESS) {
        logger.error('Processing Unit Groups failed');
        return resultsEnum.FAILURE;
    }

    // scrape coupon codes and pricing plans from escapia
    let scrapeCouponCodesAndPricingPlansResult = await scrapeCouponCodesAndPricingPlans(escapiaCollectionName);
    if (scrapeCouponCodesAndPricingPlansResult !== resultsEnum.SUCCESS) {
        logger.error('Scraping Coupon Codes failed');
        return resultsEnum.FAILURE;
    }

    // process coupon codes and pricing plans
    let processCouponCodesAndPricingPlansResult = await processCouponCodesAndPricingPlans(escapiaCollectionName, sfdcCollectionName, syncOperationsCollectionName);
    if (processCouponCodesAndPricingPlansResult !== resultsEnum.SUCCESS) {
        logger.error('Processing Coupon Codes failed');
        return resultsEnum.FAILURE;
    }

    // scrape rate rules from escapia (has to happen after scraping pricing plans)
    let scrapeRateRulesResult = await scrapeRateRules(escapiaCollectionName);
    if (scrapeRateRulesResult !== resultsEnum.SUCCESS) {
        logger.error('Scraping Rate Rules failed');
        return resultsEnum.FAILURE;
    }

    // process rate rules
    let processRateRulesResult = await processRateRules(escapiaCollectionName, sfdcCollectionName, syncOperationsCollectionName);
    if (processRateRulesResult !== resultsEnum.SUCCESS) {
        logger.error('Processing Rate Rules failed');
        return resultsEnum.FAILURE;
    }

    // sync all
    let syncResult = await sync(syncOperationsCollectionName);
    if (syncResult !== resultsEnum.SUCCESS) {
        logger.error('Sync failed!!!');
        return resultsEnum.FATAL;
    }

    logger.verbose('COUPON CODE SYNC COMPLETE');
    return resultsEnum.SUCCESS;
};

const handler = async () => {
    let retryLimit = 3;
    let retryCount = 0;
    let syncResult;

    while (retryCount < retryLimit && syncResult !== resultsEnum.SUCCESS) {
        syncResult = await syncCouponCodes();
        if (syncResult === resultsEnum.FATAL) {
            logger.alert('Syncing coupon codes from mongo to sfdc failed!');
            break;
        }
        retryCount += 1;
    }


    logger.verbose(`syncResult: ${syncResult}`); // temp jps test

    if (syncResult === resultsEnum.FAILURE) {
        logger.alert(`All ${retryLimit} coupon code sync attempts failed; stopping for now`);
    }

    process.exit(0);
};

const invokeCouponCodeRESTEndpoint = async () => {
    const reservationData = {
        unitId: 'RLTVL131',
        code: '1000265817702',
        bookingDate: '2020-03-23',
        startDate: '2020-05-05',
        endDate: '2020-05-12',
        adults: 2,
        children: 0,
        pets: 1,
        totalGuests: 2,
        nights: [
            {
                nightDate: '2020-05-05',
                baseNightlyRate: 100,
                couponCodeAdjustedRate: null,
                dayOfWeek: 'Wed',
                nightNumber: 1
            },
            {
                nightDate: '2020-05-06',
                baseNightlyRate: 100,
                couponCodeAdjustedRate: null,
                dayOfWeek: 'Thurs',
                nightNumber: 1
            },
            {
                nightDate: '2020-05-07',
                baseNightlyRate: 100,
                couponCodeAdjustedRate: null,
                dayOfWeek: 'Fri',
                nightNumber: 2
            },
            {
                nightDate: '2020-05-08',
                baseNightlyRate: 100,
                couponCodeAdjustedRate: null,
                dayOfWeek: 'Sat',
                nightNumber: 3
            },
            {
                nightDate: '2020-05-09',
                baseNightlyRate: 100,
                couponCodeAdjustedRate: null,
                dayOfWeek: 'Sun',
                nightNumber: 4
            },
            {
                nightDate: '2020-05-10',
                baseNightlyRate: 100,
                couponCodeAdjustedRate: null,
                dayOfWeek: 'Mon',
                nightNumber: 5
            },
            {
                nightDate: '2020-05-11',
                baseNightlyRate: 100,
                couponCodeAdjustedRate: null,
                dayOfWeek: 'Tue',
                nightNumber: 6
            }
        ]
    };

    let results;
    try {
        results = await customRequestSF({
            method: 'POST',
            url: '/services/apexrest/CouponCode',
            body: {
                reservationData: reservationData
            }
        });
    }
    catch (e) {
        logger.error(`coupon code rest failed: ${e.message}`);
    }

    logger.verbose(`got results: ${JSON.stringify(results)}`);

    process.exit(0);
};

const test = async () => {
    console.log('TEESSSSTTTT');
    await sleep(30000);
    process.exit(0);
};

exports.handler = handler;
exports.invokeCouponCodeRESTEndpoint = invokeCouponCodeRESTEndpoint;
exports.test = test;