const logger = require('@turnkeyvr/logger');
const MongoDbCouponCodes = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('coupon-codes-prod');
const { querySF } = require('../util/sfdc');

// global utility methods
const { sleep, resultsEnum, convertArrayToAnArrayOfArrays } = require('../../util/misc');

const loadPricingPlanRecordsFromSFDCIntoMongo = async (sfdcCollectionName) => {
    // query sfdc records
    let sfdcQueryResults = await querySF('SELECT Id, Name, Escapia_Id__c FROM Pricing_Plan__c');

    // format records
    let sfdcRecords = sfdcQueryResults.records.map(sfdcRecord => {
        // add sObjectType
        sfdcRecord.sObjectType = 'Pricing_Plan__c';
        return sfdcRecord;
    });

    if (sfdcRecords.length > 0) {
        await MongoDbCouponCodes.insertMany(sfdcCollectionName, sfdcRecords);
    }

    return resultsEnum.SUCCESS;
};

// todo: add Description__c
const loadCouponCodeRecordsFromSFDCIntoMongo = async (sfdcCollectionName) => {
    // query sfdc records
    let sfdcQueryResults = await querySF('SELECT Id, Name, Escapia_Id__c, Pricing_Plan__r.Escapia_Id__c, Status__c, Effective_Date__c, Expiration_Date__c, Limit__c, Used__c FROM Coupon_Code__c');

    // format records
    let sfdcRecords = sfdcQueryResults.records.map(sfdcRecord => {
        if (sfdcRecord.Pricing_Plan__r && sfdcRecord.Pricing_Plan__r.attributes) {
            delete sfdcRecord.Pricing_Plan__r.attributes;
        }
        // add sObjectType
        sfdcRecord.sObjectType = 'Coupon_Code__c';
        return sfdcRecord;
    });

    if (sfdcRecords.length > 0) {
        await MongoDbCouponCodes.insertMany(sfdcCollectionName, sfdcRecords);
    }

    return resultsEnum.SUCCESS;
};

module.exports = async function(escapiaCollectionName, sfdcCollectionName, syncOperationsCollectionName) {
    try {
        // first load sfdc records from sfdc into mongo
        await loadPricingPlanRecordsFromSFDCIntoMongo(sfdcCollectionName);
        await loadCouponCodeRecordsFromSFDCIntoMongo(sfdcCollectionName);

        // then being processing
        let sfdcRecords = await MongoDbCouponCodes.find(sfdcCollectionName, {sObjectType: { '$in': [ 'Coupon_Code__c', 'Pricing_Plan__c' ] }});
        let escapiaRecords = await MongoDbCouponCodes.find(escapiaCollectionName, {sObjectType: { '$in': [ 'Coupon_Code__c', 'Pricing_Plan__c' ] }});

        logger.verbose('sfdcCouponCodesAndPricingPlans.length: ', sfdcRecords.length);
        logger.verbose('escapiaCouponCodesAndPricingPlans.length: ', escapiaRecords.length);

        let sfdcPricingPlanRecords = sfdcRecords.filter(r => r.sObjectType === 'Pricing_Plan__c');
        let escapiaPricingPlanRecords = escapiaRecords.filter(r => r.sObjectType === 'Pricing_Plan__c');
        let sfdcCouponCodeRecords = sfdcRecords.filter(r => r.sObjectType === 'Coupon_Code__c');
        let escapiaCouponCodeRecords = escapiaRecords.filter(r => r.sObjectType === 'Coupon_Code__c');

        let sfdcPricingPlanRecordsMap = new Map();
        let escapiaPricingPlanRecordsMap = new Map();
        let sfdcCouponCodeRecordsMap = new Map();
        let escapiaCouponCodeRecordsMap = new Map();

        sfdcPricingPlanRecords.forEach(record => {
            delete record.sObjectType; // required for processing
            sfdcPricingPlanRecordsMap.set(record.Escapia_Id__c, record);
        });
        escapiaPricingPlanRecords.forEach(record => {
            delete record.sObjectType; // required for processing
            escapiaPricingPlanRecordsMap.set(record.Escapia_Id__c, record);
        });
        sfdcCouponCodeRecords.forEach(record => {
            delete record.sObjectType; // required for processing
            sfdcCouponCodeRecordsMap.set(record.Escapia_Id__c, record);
        });
        escapiaCouponCodeRecords.forEach(record => {
            delete record.sObjectType; // required for processing
            escapiaCouponCodeRecordsMap.set(record.Escapia_Id__c, record);
        });

        // {
        //     action: 'TREE INSERT | INSERT | DELETE | UPDATE | UPSERT W/EXTERNAL ID',
        //     sObjectType: 'SFDC Object Name',
        //     record: 'Object Record',
        //     status: 'PENDING'
        // }
        let dmlOperations = [];

        // first process pricing plans; look for records that need to be deleted from sfdc
        Array.from(sfdcPricingPlanRecordsMap.keys()).forEach(escapiaId => {
            let sfdcRecord = sfdcPricingPlanRecordsMap.get(escapiaId);
            if (!escapiaPricingPlanRecordsMap.has(escapiaId)) {
                dmlOperations.push({
                    action: 'DELETE',
                    sObjectType: 'Pricing_Plan__c',
                    record: { Id: sfdcRecord.Id },
                    status: 'PENDING'
                });
                sfdcPricingPlanRecordsMap.delete(escapiaId);
            }
        });
        // otherwise do upserts on all the remaining escapia pricing plans
        if (escapiaPricingPlanRecordsMap.size > 0) {
            Array.from(escapiaPricingPlanRecordsMap.keys()).forEach(escapiaId => {
                let record = escapiaPricingPlanRecordsMap.get(escapiaId);
                delete record._id;
                dmlOperations.push({
                    action: 'UPSERT W/EXTERNAL ID',
                    sObjectType: 'Pricing_Plan__c',
                    record: record,
                    status: 'PENDING'
                });
                escapiaPricingPlanRecordsMap.delete(escapiaId);
            });
        }


        // first process pricing plans; look for records that need to be deleted from sfdc
        Array.from(sfdcCouponCodeRecordsMap.keys()).forEach(escapiaId => {
            let sfdcRecord = sfdcCouponCodeRecordsMap.get(escapiaId);
            if (!escapiaCouponCodeRecordsMap.has(escapiaId)) {
                dmlOperations.push({
                    action: 'DELETE',
                    sObjectType: 'Coupon_Code__c',
                    record: { Id: sfdcRecord.Id },
                    status: 'PENDING'
                });
                sfdcCouponCodeRecordsMap.delete(escapiaId);
            }
        });
        // otherwise do upserts on all the remaining escapia pricing plans
        if (escapiaCouponCodeRecordsMap.size > 0) {
            Array.from(escapiaCouponCodeRecordsMap.keys()).forEach(escapiaId => {
                let record = escapiaCouponCodeRecordsMap.get(escapiaId);
                delete record._id;
                dmlOperations.push({
                    action: 'UPSERT W/EXTERNAL ID',
                    sObjectType: 'Coupon_Code__c',
                    record: escapiaCouponCodeRecordsMap.get(escapiaId),
                    status: 'PENDING'
                });
                escapiaCouponCodeRecordsMap.delete(escapiaId);
            });
        }

        logger.verbose('coupon codes + pricing plans dmlOperations.length: ', dmlOperations.length);

        let upsertWithExternalIdRecords = dmlOperations.filter(r => r.action === 'UPSERT W/EXTERNAL ID');
        let deleteRecords = dmlOperations.filter(r => r.action === 'DELETE');


        logger.verbose(`COUPON CODES + PRICING PLANS:: upsertWithExternalIdRecords: ${upsertWithExternalIdRecords.length} | deleteRecords: ${deleteRecords.length}`);


        // insert dml operations into mongo collection
        if (dmlOperations.length > 0) {
            await MongoDbCouponCodes.insertMany(syncOperationsCollectionName, dmlOperations);
        }

        await MongoDbCouponCodes.close();
        return resultsEnum.SUCCESS;
    }
    catch (e) {
        logger.error(e);
        await MongoDbCouponCodes.close();
        return resultsEnum.FAILURE;
    }
}