const logger = require('@turnkeyvr/logger');
const MongoDbCouponCodes = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('coupon-codes-prod');
const { querySF } = require('../util/sfdc');

// global utility methods
const { sleep, resultsEnum, convertArrayToAnArrayOfArrays } = require('../../util/misc');

const rateRuleQuery = 'SELECT Id, \
    Name, \
    Description__c, \
    Escapia_Id__c, \
    Status__c, \
    Sequence_Number__c, \
    Pricing_Plan__r.Escapia_Id__c \
    FROM Rate_Rule__c';

const rateRuleConditionQuery = 'SELECT Id, \
    RecordTypeId, \
    Rate_Rule__r.Escapia_Id__c, \
    Between_In_or_Within__c, \
    Coupon_Code_Season__r.Escapia_Id__c, \
    End_Date__c, \
    Friday__c, \
    Includes__c, \
    Is_or_Is_Not__c, \
    Maximum__c, \
    Minimum__c, \
    Monday__c, \
    Nights_or_Weeks__c, \
    Night_Type__c, \
    Reservation_Type__c, \
    Saturday__c, \
    Start_Date__c, \
    Sunday__c, \
    Thursday__c, \
    Tuesday__c, \
    Wednesday__c, \
    ( \
        SELECT Id, \
        Unit_Group_Name__c \
        FROM Rate_Rule_Condition_Unit_Groups__r \
        ORDER BY Unit_Group_Name__c ASC \
    ), \
    ( \
        SELECT Id, \
        Rental_Name__c \
        FROM Coupon_Code_Units__r \
        ORDER BY Rental_Name__c ASC \
    ) \
    FROM Rate_Rule_Condition__c';

const rateRuleActionQuery = 'SELECT Id, \
    RecordTypeId, \
    Adjusted_Rate_Duration__c, \
    Amount__c, \
    Amount_Unit_of_Measure__c, \
    Applicable_to_Friday__c, \
    Applicable_to_Monday__c, \
    Applicable_to_Saturday__c, \
    Applicable_to_Sunday__c, \
    Applicable_to_Thursday__c, \
    Applicable_to_Tuesday__c, \
    Applicable_to_Wednesday__c, \
    Month_Rate__c, \
    Night_Rate__c, \
    Period_From__c, \
    Period_To__c, \
    Period_Type__c, \
    Rate_Rule__r.Escapia_Id__c, \
    Scope__c, \
    Sequence_Number__c, \
    Specific_Date__c, \
    Term_Rate__c, \
    Week_Rate__c \
    FROM Rate_Rule_Action__c \
    ORDER BY Sequence_Number__c ASC';

const cleanSFDCRecords = (obj) => {
    var propNames = Object.getOwnPropertyNames(obj);
    for (var i = 0; i < propNames.length; i++) {
        var propName = propNames[i];
        if (obj[propName] === null) {
            delete obj[propName];
        }
    }
}

// this one is more complex than unit groups b/c have to query records separately, then join together
// ...then, before inserting will need to break up again
const loadRecordsFromSFDCIntoMongo = async (sfdcCollectionName) => {
    // rate rule records
    let sfdcRateRules = (await querySF(rateRuleQuery)).records;

    // rate rule conditions + rate rule condition unit groups + coupon code units
    let sfdcRateRuleConditions = (await querySF(rateRuleConditionQuery)).records;

    // rate rule actions
    let sfdcRateRuleActions = (await querySF(rateRuleActionQuery)).records;

    // need to piece together the records to match the schema in mongo
    let rateRuleIdMap = new Map();
    for (let i = 0; i < sfdcRateRules.length; i++) {
        let sfdcRateRule = sfdcRateRules[i];
        sfdcRateRule.sObjectType = 'Rate_Rule__c'; // add sObjectType
        sfdcRateRule.Rate_Rule_Conditions__r = [];
        sfdcRateRule.Rate_Rule_Actions__r = [];

        // modify some sfdc attributes (misc from sfgen query)
        if (sfdcRateRule.Pricing_Plan__r && sfdcRateRule.Pricing_Plan__r.attributes) {
            delete sfdcRateRule.Pricing_Plan__r.attributes;
        }

        // other misc data transformations
        if (sfdcRateRule.Description__c === null) {
            sfdcRateRule.Description__c = '';
        }

        rateRuleIdMap.set(sfdcRateRule.Escapia_Id__c, sfdcRateRule);
    }

    for (let i = 0; i < sfdcRateRuleConditions.length; i++) {
        let sfdcRateRuleCondition = sfdcRateRuleConditions[i];
        let sfdcRateRule = rateRuleIdMap.get(sfdcRateRuleCondition.Rate_Rule__r.Escapia_Id__c);

        // modify some sfdc attributes (misc from sfgen query)
        if (sfdcRateRuleCondition.Rate_Rule__r) {
            delete sfdcRateRuleCondition.Rate_Rule__r;
        }

        if (sfdcRateRuleCondition.Coupon_Code_Season__r && sfdcRateRuleCondition.Coupon_Code_Season__r.attributes) {
            delete sfdcRateRuleCondition.Coupon_Code_Season__r.attributes;
        }

        if (!sfdcRateRuleCondition.Rate_Rule_Condition_Unit_Groups__r) {
            sfdcRateRuleCondition.Rate_Rule_Condition_Unit_Groups__r = [];
        }
        else {
            sfdcRateRuleCondition.Rate_Rule_Condition_Unit_Groups__r = sfdcRateRuleCondition.Rate_Rule_Condition_Unit_Groups__r.records.map(r => {
                return {
                    Unit_Group_Name__c: r.Unit_Group_Name__c
                };
            });
        }

        if (!sfdcRateRuleCondition.Coupon_Code_Units__r) {
            sfdcRateRuleCondition.Coupon_Code_Units__r = [];
        }
        else {
            sfdcRateRuleCondition.Coupon_Code_Units__r = sfdcRateRuleCondition.Coupon_Code_Units__r.records.map(r => {
                return {
                    Rental_Name__c: r.Rental_Name__c
                };
            });
        }

        // remove null object properties (won't be present in escapia scraping data)
        cleanSFDCRecords(sfdcRateRuleCondition);


        sfdcRateRule.Rate_Rule_Conditions__r.push(sfdcRateRuleCondition);
        rateRuleIdMap.set(sfdcRateRule.Escapia_Id__c, sfdcRateRule);
    }

    for (let i = 0; i < sfdcRateRuleActions.length; i++) {
        let sfdcRateRuleAction = sfdcRateRuleActions[i];
        let sfdcRateRule = rateRuleIdMap.get(sfdcRateRuleAction.Rate_Rule__r.Escapia_Id__c);

        // modify some sfdc attributes (misc from sfgen query)
        if (sfdcRateRuleAction.Rate_Rule__r) {
            delete sfdcRateRuleAction.Rate_Rule__r;
        }

        if (sfdcRateRuleAction.Amount__c) {
            sfdcRateRuleAction.Amount__c = parseFloat(parseFloat(sfdcRateRuleAction.Amount__c).toFixed(2));
        }
        if (sfdcRateRuleAction.Month_Rate__c) {
            sfdcRateRuleAction.Month_Rate__c = parseFloat(parseFloat(sfdcRateRuleAction.Month_Rate__c).toFixed(2));
        }
        if (sfdcRateRuleAction.Night_Rate__c) {
            sfdcRateRuleAction.Night_Rate__c = parseFloat(parseFloat(sfdcRateRuleAction.Night_Rate__c).toFixed(2));
        }
        if (sfdcRateRuleAction.Term_Rate__c) {
            sfdcRateRuleAction.Term_Rate__c = parseFloat(parseFloat(sfdcRateRuleAction.Term_Rate__c).toFixed(2));
        }
        if (sfdcRateRuleAction.Week_Rate__c) {
            sfdcRateRuleAction.Week_Rate__c = parseFloat(parseFloat(sfdcRateRuleAction.Week_Rate__c).toFixed(2));
        }


        // remove null object properties (won't be present in escapia scraping data)
        cleanSFDCRecords(sfdcRateRuleAction);

        sfdcRateRule.Rate_Rule_Actions__r.push(sfdcRateRuleAction);
        rateRuleIdMap.set(sfdcRateRule.Escapia_Id__c, sfdcRateRule);
    }


    if (rateRuleIdMap.size > 0) {
        await MongoDbCouponCodes.insertMany(sfdcCollectionName, Array.from(rateRuleIdMap.values()));
    }

    return resultsEnum.SUCCESS;
};

module.exports = async function(escapiaCollectionName, sfdcCollectionName, syncOperationsCollectionName) {
    try {
        // first load sfdc records from sfdc into mongo
        await loadRecordsFromSFDCIntoMongo(sfdcCollectionName);

        // then being processing
        let sfdcRecords = await MongoDbCouponCodes.find(sfdcCollectionName, {sObjectType: 'Rate_Rule__c'});
        let escapiaRecords = await MongoDbCouponCodes.find(escapiaCollectionName, {sObjectType: 'Rate_Rule__c'});

        logger.verbose('sfdcRateRules.length: ', sfdcRecords.length);
        logger.verbose('escapiaRateRules.length: ', escapiaRecords.length);

        let sfdcRecordsMap = new Map();
        let escapiaRecordsMap = new Map();

        sfdcRecords.forEach(record => {
            delete record.sObjectType; // required for processing
            sfdcRecordsMap.set(record.Escapia_Id__c, record);
        });
        escapiaRecords.forEach(record => {
            delete record.sObjectType; // required for processing
            escapiaRecordsMap.set(record.Escapia_Id__c, record);
        });

        let dmlOperations = [];

        Array.from(sfdcRecordsMap.keys()).forEach(escapiaId => {
            let sfdcRecord = sfdcRecordsMap.get(escapiaId);

            // delete records from sfdc using custom rest endpoint
            if (!escapiaRecordsMap.has(escapiaId)) {
                dmlOperations.push({
                    action: 'RATE RULE DELETE',
                    sObjectType: 'Rate_Rule__c',
                    record: { Id: sfdcRecord.Id },
                    status: 'PENDING'
                });
                return;
            }
        });

        Array.from(escapiaRecordsMap.keys()).forEach(escapiaId => {
            let sfdcRecord = escapiaRecordsMap.get(escapiaId);

            // complicated upsert operation using custom rest endpoint
            dmlOperations.push({
                action: 'RATE RULE UPSERT',
                sObjectType: 'Rate_Rule__c',
                record: sfdcRecord,
                status: 'PENDING'
            });
        });

        logger.verbose('rate rules dmlOperations.length: ', dmlOperations.length);

        let upsertRecords = dmlOperations.filter(r => r.action === 'RATE RULE UPSERT');
        let deleteRecords = dmlOperations.filter(r => r.action === 'RATE RULE DELETE');

        logger.verbose(`UNITGROUPS:: rateRuleUpsert records: ${upsertRecords.length} | deleteRecords: ${deleteRecords.length}`);

        // insert dml operations into mongo collection
        if (dmlOperations.length > 0) {
            await MongoDbCouponCodes.insertMany(syncOperationsCollectionName, dmlOperations);
        }

        await MongoDbCouponCodes.close();
        return resultsEnum.SUCCESS;
    }
    catch (e) {
        logger.error(e);
        await MongoDbCouponCodes.close();
        return resultsEnum.FAILURE;
    }
}