const logger = require('@turnkeyvr/logger');
const MongoDbCouponCodes = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('coupon-codes-prod');
const moment = require('moment');
const diff = require('deep-diff');
const { querySF } = require('../util/sfdc');

// global utility methods
const { sleep, resultsEnum, convertArrayToAnArrayOfArrays } = require('../../util/misc');

const loadSeasonsFromSFDCIntoMongo = async (sfdcCollectionName) => {
    // query sfdc season records
    let seasonRecordsQueryResults = await querySF('SELECT Id, Name, Escapia_Id__c, (SELECT Id, Start_Date__c, End_Date__c FROM Coupon_Code_Season_Date_Ranges__r ORDER BY Start_Date__c ASC, End_Date__c ASC) FROM Coupon_Code_Season__c');

    // format records
    let seasonRecords = seasonRecordsQueryResults.records.map(season => {
        // add sObjectType
        season.sObjectType = 'Coupon_Code_Season__c';
        if (!season.Coupon_Code_Season_Date_Ranges__r) {
            season.Coupon_Code_Season_Date_Ranges__r = [];
            return season;
        }
        season.Coupon_Code_Season_Date_Ranges__r = season.Coupon_Code_Season_Date_Ranges__r.records.map(dateRange => {
            return {
                Id: dateRange.Id,
                Start_Date__c: dateRange.Start_Date__c,
                End_Date__c: dateRange.End_Date__c
            };
        });

        return season;
    });

    if (seasonRecords.length > 0) {
        await MongoDbCouponCodes.insertMany(sfdcCollectionName, seasonRecords);
    }

    return resultsEnum.SUCCESS;
};

module.exports = async function(escapiaCollectionName, sfdcCollectionName, syncOperationsCollectionName) {
    try {
        // first load sfdc seasons from sfdc into mongo
        await loadSeasonsFromSFDCIntoMongo(sfdcCollectionName);

        // then being processing
        let sfdcRecords = await MongoDbCouponCodes.find(sfdcCollectionName, {sObjectType: 'Coupon_Code_Season__c'});
        let escapiaRecords = await MongoDbCouponCodes.find(escapiaCollectionName, {sObjectType: 'Coupon_Code_Season__c'});

        logger.verbose('sfdcSeasons.length: ', sfdcRecords.length);
        logger.verbose('escapiaSeasons.length: ', escapiaRecords.length);

        let sfdcRecordsMap = new Map();
        let escapiaRecordsMap = new Map();

        sfdcRecords.forEach(record => {
            delete record.sObjectType; // required for processing
            sfdcRecordsMap.set(record.Escapia_Id__c, record);
        });
        escapiaRecords.forEach(record => {
            delete record.sObjectType; // required for processing
            escapiaRecordsMap.set(record.Escapia_Id__c, record);
        });

        // {
        //     action: 'TREE INSERT | INSERT | DELETE | UPDATE',
        //     sObjectType: 'SFDC Object Name',
        //     record: 'Object Record',
        //     status: 'PENDING'
        // }
        let dmlOperations = [];

        // first compare sfdc to escapia
        Array.from(sfdcRecordsMap.keys()).forEach(escapiaId => {
            let sfdcRecord = sfdcRecordsMap.get(escapiaId);
            
            // delete record from sfdc
            if (!escapiaRecordsMap.has(escapiaId)) {
                dmlOperations.push({
                    action: 'DELETE',
                    sObjectType: 'Coupon_Code_Season__c',
                    record: { Id: sfdcRecord.Id },
                    status: 'PENDING'
                });
                sfdcRecord.Coupon_Code_Season_Date_Ranges__r.forEach(record => {
                    dmlOperations.push({
                        action: 'DELETE',
                        sObjectType: 'Coupon_Code_Season_Date_Range__c',
                        record: { Id: record.Id },
                        status: 'PENDING'
                    });
                });
                sfdcRecordsMap.delete(escapiaId);
                return;
            }

            let escapiaRecord = escapiaRecordsMap.get(escapiaId);

            // sort escapia season's dateranges
            escapiaRecord.Coupon_Code_Season_Date_Ranges__r = escapiaRecord.Coupon_Code_Season_Date_Ranges__r.sort((a, b) => {
                let aStartDate = moment(a.Start_Date__c);
                let aEndDate = moment(a.End_Date__c);
                let bStartDate = moment(b.Start_Date__c);
                let bEndDate = moment(b.End_Date__c);
                if (aStartDate.isSame(bStartDate)) {
                    if (aEndDate.isSame(bEndDate)) {
                        return 0;
                    }
                    if (aEndDate.isAfter(bEndDate)) {
                        return 1;
                    }
                    return -1;
                }
                if (aStartDate.isAfter(bStartDate)) {
                    return 1;
                }
                return -1;
            });

            // need to ignore certain id values
            let diffValues = diff(sfdcRecord, escapiaRecord, {
                prefilter: (path, key) => {
                    if (key === 'Id' || key === '_id') {
                        return true;
                    }
                    return false;
                }
            });

            // no difference between records; no actions required
            if (!diffValues) {
                // remove escapiaId from both maps and early exit
                sfdcRecordsMap.delete(escapiaId);
                escapiaRecordsMap.delete(escapiaId);
                return;
            }

            let resyncChildRecords = false;
            for (let i = 0; i < diffValues.length; i++) {
                let diffValue = diffValues[i];
                // if the record's Name changed, update in SFDC
                if (diffValue.kind === 'E' && diffValue.path.length === 1 && diffValue.path[0] === 'Name') {
                    dmlOperations.push({
                        action: 'UPDATE',
                        sObjectType: 'Coupon_Code_Season__c',
                        record: { Id: sfdcRecord.Id, Name: diffValue.rhs },
                        status: 'PENDING'
                    });
                    continue;
                }
                // the only other type of change is with Date Range's; doing full resync in this case
                resyncChildRecords = true;
            }

            if (resyncChildRecords) {
                sfdcRecord.Coupon_Code_Season_Date_Ranges__r.forEach(dateRange => {
                    dmlOperations.push({
                        action: 'DELETE',
                        sObjectType: 'Coupon_Code_Season_Date_Range__c',
                        record: { Id: dateRange.Id },
                        status: 'PENDING'
                    });
                });
                escapiaRecord.Coupon_Code_Season_Date_Ranges__r.forEach(dateRange => {
                    dmlOperations.push({
                        action: 'INSERT',
                        sObjectType: 'Coupon_Code_Season_Date_Range__c',
                        record: { 
                            Coupon_Code_Season__c: sfdcRecord.Id, 
                            Start_Date__c: dateRange.Start_Date__c, 
                            End_Date__c: dateRange.End_Date__c 
                        },
                        status: 'PENDING'
                    });
                });
            }

            // remove escapiaId from both maps
            sfdcRecordsMap.delete(escapiaId);
            escapiaRecordsMap.delete(escapiaId);
        });

        // these are new records from escapia that need to be inserted into sfdc
        if (escapiaRecordsMap.size > 0) {
            logger.verbose(`escapiaRecordsMapLength: ${escapiaRecordsMap.size}`);
            Array.from(escapiaRecordsMap.keys()).forEach(escapiaId => {
                dmlOperations.push({
                    action: 'TREE INSERT',
                    sObjectType: 'Coupon_Code_Season__c',
                    record: escapiaRecordsMap.get(escapiaId),
                    status: 'PENDING'
                });
                escapiaRecordsMap.delete(escapiaId);
            });
        }

        logger.verbose('seasons dmlOperations.length: ', dmlOperations.length);

        let treeInsertRecords = dmlOperations.filter(r => r.action === 'TREE INSERT');
        let insertRecords = dmlOperations.filter(r => r.action === 'INSERT');
        let deleteRecords = dmlOperations.filter(r => r.action === 'DELETE');
        let updateRecords = dmlOperations.filter(r => r.action === 'UPDATE');

        logger.verbose(`SEASONS:: treeInsertRecords: ${treeInsertRecords.length} | insertRecords: ${insertRecords.length} | deleteRecords: ${deleteRecords.length} | updateRecords: ${updateRecords.length}`);


        // insert dml operations into mongo collection
        if (dmlOperations.length > 0) {
            await MongoDbCouponCodes.insertMany(syncOperationsCollectionName, dmlOperations);
        }

        await MongoDbCouponCodes.close();
        return resultsEnum.SUCCESS;
    }
    catch (e) {
        logger.error(e);
        await MongoDbCouponCodes.close();
        return resultsEnum.FAILURE;
    }
}