const logger = require('@turnkeyvr/logger');
const diff = require('deep-diff');
const { querySF } = require('../util/sfdc');

// global utility methods
const { sleep, resultsEnum, convertArrayToAnArrayOfArrays } = require('../../util/misc');

const loadRecordsFromSFDCIntoMongo = async (sfdcCollectionName) => {
    // query sfdc records
    const MongoDbCouponCodes = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('coupon-codes-prod');
    let sfdcQueryResults = await querySF('SELECT Id, Name, Escapia_Id__c, Status__c, (SELECT Id, Rental_Name__c FROM Coupon_Code_Units__r ORDER BY Rental_Name__c ASC) FROM Coupon_Code_Unit_Group__c');

    // format records
    let sfdcRecords = sfdcQueryResults.records.map(sfdcRecord => {
        // add sObjectType
        sfdcRecord.sObjectType = 'Coupon_Code_Unit_Group__c';
        if (!sfdcRecord.Coupon_Code_Units__r) {
            sfdcRecord.Coupon_Code_Units__r = [];
            return sfdcRecord;
        }
        sfdcRecord.Coupon_Code_Units__r = sfdcRecord.Coupon_Code_Units__r.records.map(childRecord => {
            return {
                Id: childRecord.Id,
                Rental_Name__c: childRecord.Rental_Name__c
            };
        });
        return sfdcRecord;
    });

    if (sfdcRecords.length > 0) {
        await MongoDbCouponCodes.insertMany(sfdcCollectionName, sfdcRecords);
    }

    await MongoDbCouponCodes.close()

    return resultsEnum.SUCCESS;
};

module.exports = async function(escapiaCollectionName, sfdcCollectionName, syncOperationsCollectionName) {
    const MongoDbCouponCodes = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('coupon-codes-prod');
    
    try {
        // first load sfdc records from sfdc into mongo
        
        await loadRecordsFromSFDCIntoMongo(sfdcCollectionName);

        // then being processing
        let sfdcRecords = await MongoDbCouponCodes.find(sfdcCollectionName, {sObjectType: 'Coupon_Code_Unit_Group__c'});
        let escapiaRecords = await MongoDbCouponCodes.find(escapiaCollectionName, {sObjectType: 'Coupon_Code_Unit_Group__c'});

        logger.verbose('sfdcUnitGroups.length: ', sfdcRecords.length);
        logger.verbose('escapiaUnitGroups.length: ', escapiaRecords.length);

        let sfdcRecordsMap = new Map();
        let escapiaRecordsMap = new Map();

        sfdcRecords.forEach(record => {
            delete record.sObjectType; // required for processing
            sfdcRecordsMap.set(record.Escapia_Id__c, record);
        });
        escapiaRecords.forEach(record => {
            delete record.sObjectType; // required for processing
            escapiaRecordsMap.set(record.Escapia_Id__c, record);
        });

        // {
        //     action: 'TREE INSERT | INSERT | DELETE | UPDATE',
        //     sObjectType: 'SFDC Object Name',
        //     record: 'Object Record',
        //     status: 'PENDING'
        // }
        let dmlOperations = [];

        // first compare sfdc to escapia
        Array.from(sfdcRecordsMap.keys()).forEach(escapiaId => {
            let sfdcRecord = sfdcRecordsMap.get(escapiaId);
            
            // delete record from sfdc
            if (!escapiaRecordsMap.has(escapiaId)) {
                dmlOperations.push({
                    action: 'DELETE',
                    sObjectType: 'Coupon_Code_Unit_Group__c',
                    record: { Id: sfdcRecord.Id },
                    status: 'PENDING'
                });
                sfdcRecord.Coupon_Code_Units__r.forEach(record => {
                    dmlOperations.push({
                        action: 'DELETE',
                        sObjectType: 'Coupon_Code_Unit__c',
                        record: { Id: record.Id },
                        status: 'PENDING'
                    });
                });
                sfdcRecordsMap.delete(escapiaId);
                return;
            }

            let escapiaRecord = escapiaRecordsMap.get(escapiaId);

            // sort escapia unit group's units
            escapiaRecord.Coupon_Code_Units__r = escapiaRecord.Coupon_Code_Units__r.sort((a, b) => {
                if (a.Rental_Name__c === b.Rental_Name__c) {
                    return 0;
                }
                if (a.Rental_Name__c > b.Rental_Name__c) {
                    return 1;
                }
                return -1;
            });

            // need to ignore certain id values
            let diffValues = diff(sfdcRecord, escapiaRecord, {
                prefilter: (path, key) => {
                    if (key === 'Id' || key === '_id') {
                        return true;
                    }
                    return false;
                }
            });

            // no difference between records; no actions required
            if (!diffValues) {
                // remove escapiaId from both maps and early exit
                sfdcRecordsMap.delete(escapiaId);
                escapiaRecordsMap.delete(escapiaId);
                return;
            }

            let resyncChildRecords = false;
            for (let i = 0; i < diffValues.length; i++) {
                let diffValue = diffValues[i];
                if (diffValue.kind === 'E' && diffValue.path.length === 1 && diffValue.path[0] === 'Name') {
                    dmlOperations.push({
                        action: 'UPDATE',
                        sObjectType: 'Coupon_Code_Unit_Group__c',
                        record: { Id: sfdcRecord.Id, Name: diffValue.rhs },
                        status: 'PENDING'
                    });
                }
                else if (diffValue.kind === 'E' && diffValue.path.length === 1 && diffValue.path[0] === 'Status__c') {
                    dmlOperations.push({
                        action: 'UPDATE',
                        sObjectType: 'Coupon_Code_Unit_Group__c',
                        record: { Id: sfdcRecord.Id, Status__c: diffValue.rhs },
                        status: 'PENDING'
                    });
                }
                else {
                    // the only other type of change is with Date Range's; doing full resync in this case
                    resyncChildRecords = true;
                }
            }

            if (resyncChildRecords) {
                sfdcRecord.Coupon_Code_Units__r.forEach(childRecord => {
                    dmlOperations.push({
                        action: 'DELETE',
                        sObjectType: 'Coupon_Code_Unit__c',
                        record: { Id: childRecord.Id },
                        status: 'PENDING'
                    });
                });
                escapiaRecord.Coupon_Code_Units__r.forEach(childRecord => {
                    dmlOperations.push({
                        action: 'INSERT',
                        sObjectType: 'Coupon_Code_Unit__c',
                        record: { 
                            Coupon_Code_Unit_Group__c: sfdcRecord.Id, 
                            Rental_Name__c: childRecord.Rental_Name__c
                        },
                        status: 'PENDING'
                    });
                });
            }

            // remove escapiaId from both maps
            sfdcRecordsMap.delete(escapiaId);
            escapiaRecordsMap.delete(escapiaId);
        });

        // these are new records from escapia that need to be inserted into sfdc
        if (escapiaRecordsMap.size > 0) {
            logger.verbose(`escapiaRecordsMapLength: ${escapiaRecordsMap.size}`);
            Array.from(escapiaRecordsMap.keys()).forEach(escapiaId => {
                dmlOperations.push({
                    action: 'TREE INSERT',
                    sObjectType: 'Coupon_Code_Unit_Group__c',
                    record: escapiaRecordsMap.get(escapiaId),
                    status: 'PENDING'
                });
                escapiaRecordsMap.delete(escapiaId);
            });
        }

        logger.verbose('unit groups dmlOperations.length: ', dmlOperations.length);

        let treeInsertRecords = dmlOperations.filter(r => r.action === 'TREE INSERT');
        let insertRecords = dmlOperations.filter(r => r.action === 'INSERT');
        let deleteRecords = dmlOperations.filter(r => r.action === 'DELETE');
        let updateRecords = dmlOperations.filter(r => r.action === 'UPDATE');

        logger.verbose(`UNITGROUPS:: treeInsertRecords: ${treeInsertRecords.length} | insertRecords: ${insertRecords.length} | deleteRecords: ${deleteRecords.length} | updateRecords: ${updateRecords.length}`);


        // insert dml operations into mongo collection
        if (dmlOperations.length > 0) {
            await MongoDbCouponCodes.insertMany(syncOperationsCollectionName, dmlOperations);
        }

        return resultsEnum.SUCCESS;
    }
    catch (e) {
        await logoutOfEscapia(page);
        await browser.close();
    }

    await MongoDbCouponCodes.close()
}