const logger = require('@turnkeyvr/logger');
const MongoDbCouponCodes = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('coupon-codes-prod');
const eachSeries = require('async/eachSeries');
const moment = require('moment');
const { Cluster } = require('puppeteer-cluster');

// global utility methods
const { getBrowser,
    getSelectedValueFromSelectHandle,
    getSelectValue,
    getCheckboxValue,
    getInputValue } = require('../../util/puppeteerUtil');
const { authenticateWithEscapia, setAndGetConfigEscapia, logoutOfEscapia } = require('../../util/escapiaUtil');
const { sleep, resultsEnum, convertArrayToAnArrayOfArrays } = require('../../util/misc');
const { retry } = require('../../util/retry');

// local utility methods
const { IS_PROD } = require('../util/constants');

const scrapeCouponCodeIds = async (page) => {
    let couponCodeIds = [];
    await page.goto('https://turnkeyvr.escapia.com/webagency/Property/Setup/Rates/ManageCouponCodes.aspx', {waitUntil: 'networkidle0'});
    let gatheredAllResults = false;
    while (!gatheredAllResults) {
        let elementHandleArray = await page.$$('.listRow');
        let firstIdValue = await page.evaluate(e => e.getAttribute('rowarg'), elementHandleArray[0]);
        if (couponCodeIds.includes(firstIdValue)) {
            gatheredAllResults = true;
        }
        else {
            await eachSeries(elementHandleArray, async e => {
                let idValue = await page.evaluate(e => e.getAttribute('rowarg'), e);
                couponCodeIds.push(idValue);
            });
            // click next
            let paginationSpanHandle = await page.$('#ctlDataPager');
            // await paginationSpanHandle.$$eval('a', nodes => nodes.map(n => {
            //     if (n.textContent !== 'NEXT') {
            //         return;
            //     }
            //     n.click();
            // }));
            // await page.waitForNavigation();

            await Promise.all([
                paginationSpanHandle.$$eval('a', nodes => nodes.map(n => {
                    if (n.textContent !== 'NEXT') {
                        return;
                    }
                    n.click();
                })),
                page.waitForNavigation()
            ]);
        }
    }
    return couponCodeIds;
};

// todo: scrape Description__c
const getCouponCodeRecords = async (couponCodeIds) => {
    let couponCodeRecords = [];
    let scrapingErrors = [];

    // concurrency cluster
    const cluster = await Cluster.launch({
        concurrency: Cluster.CONCURRENCY_CONTEXT,
        maxConcurrency: 10,
        puppeteerOptions: {
            headless: IS_PROD,
        },
        timeout: 60 * 1000 * 30
    });
    
    let couponCodeIdArrayArray = convertArrayToAnArrayOfArrays(couponCodeIds, 100); // an array of arrays

    const scrapeCouponCodeRecords = async ({ page, data }) => {
        let { couponCodeIds, couponCodeRecords, scrapingErrors } = data;
        try {
            setAndGetConfigEscapia('1655'); // only running against the main instance
            page = await authenticateWithEscapia(page);

            await eachSeries(couponCodeIds, async couponId => {
                try {
                    let couponCodeRecord = {
                        Escapia_Id__c: couponId
                    };
                    await page.goto('https://turnkeyvr.escapia.com/webagency/Property/Setup/Rates/OpenCouponCode.aspx?id=' + couponId, {waitUntil: 'networkidle0', timeout: 0});

                    couponCodeRecord.Name = await getInputValue(page, '#txtCouponCode');
            
                    let pricingPlanSelectHandle = await page.$('#lstPricingPlan');
                    let pricingPlanSelectValueHandle = await pricingPlanSelectHandle.getProperty('value');
                    couponCodeRecord.pricingPlanId = (await pricingPlanSelectValueHandle.jsonValue()).trim();
                    couponCodeRecord.pricingPlanName = await pricingPlanSelectHandle.$$eval('option', (nodes, pricingPlanId) => {
                        for (let i = 0; i < nodes.length; i++) {
                            if (nodes[i].value === pricingPlanId) {
                                return nodes[i].textContent;
                            }
                        }
                    }, couponCodeRecord.pricingPlanId);
            
                    let statusSelectHandle = await page.$('#lstStatus');
                    let statusSelectValueHandle = await statusSelectHandle.getProperty('value');
                    let statusSelectValue = (await statusSelectValueHandle.jsonValue()).trim();
                    couponCodeRecord.Status__c = await statusSelectHandle.$$eval('option', (nodes, statusSelectValue) => {
                        for (let i = 0; i < nodes.length; i++) {
                            if (nodes[i].value === statusSelectValue) {
                                return nodes[i].textContent;
                            }
                        }
                    }, statusSelectValue);
            
                    couponCodeRecord.Effective_Date__c = await getInputValue(page, 'input[name="dpStartDate"');
                    if (couponCodeRecord.Effective_Date__c && couponCodeRecord.Effective_Date__c !== '') {
                        couponCodeRecord.Effective_Date__c = moment(couponCodeRecord.Effective_Date__c).format('YYYY-MM-DD');
                    }
                    couponCodeRecord.Expiration_Date__c = await getInputValue(page, 'input[name="dpEndDate"');
                    if (couponCodeRecord.Expiration_Date__c && couponCodeRecord.Expiration_Date__c !== '') {
                        couponCodeRecord.Expiration_Date__c = moment(couponCodeRecord.Expiration_Date__c).format('YYYY-MM-DD');
                    }
                    couponCodeRecord.Limit__c = await getInputValue(page, '#txtRedemptionLimit');
                    couponCodeRecord.Used__c = await getInputValue(page, '#txtRedemptionCount');
            
                    couponCodeRecords.push(couponCodeRecord);
                }
                catch (e) {
                    scrapingErrors.push(e);
                }
            });

            await logoutOfEscapia(page);
        }
        catch (e) {
            scrapingErrors.push(e);
        }
    };

    for (let i = 0; i < couponCodeIdArrayArray.length; i++) {
        cluster.queue({ couponCodeIds: couponCodeIdArrayArray[i], couponCodeRecords: couponCodeRecords, scrapingErrors: scrapingErrors }, scrapeCouponCodeRecords);
    }

    await cluster.idle();
    await cluster.close();

    if (scrapingErrors.length > 0) {
        throw scrapingErrors[0];
    }

    return couponCodeRecords;
};


module.exports = async function(escapiaCollectionName) {
    let browser;
    let page;
    try {
        // load browser
        browser = await getBrowser();
        if (!browser) {
            return;
        }

        // authenticate with escapia
        setAndGetConfigEscapia('1655'); // only running against the main instance
        page = await authenticateWithEscapia(await browser.newPage());
        if (!page) {
            await browser.close();
            return;
        }

        let couponCodeIds = await scrapeCouponCodeIds(page);

        // temp testing
        // let tempIds = [];
        // for (let i = 0; i < 10; i++) {
        //     tempIds.push(couponCodeIds[i]);
        // }
        // couponCodeIds = tempIds;


        let couponCodeRecords = await getCouponCodeRecords(couponCodeIds);

        // this one is a little different because the pricing plan data is in the couponCodeRecords...
        let pricingPlanEscapiaIds = [];
        let pricingPlanRecords = [];
        couponCodeRecords.forEach(coupon => {
            if (pricingPlanEscapiaIds.includes(coupon.pricingPlanId)) {
                return;
            }
            pricingPlanEscapiaIds.push(coupon.pricingPlanId);
            pricingPlanRecords.push({
                Name: coupon.pricingPlanName,
                Escapia_Id__c: coupon.pricingPlanId,
                sObjectType: 'Pricing_Plan__c'
            });
        });

        logger.verbose(`pricingPlanRecords.length: ${pricingPlanRecords.length}`);
        logger.verbose(`pricingPlanRecords record 0: ${JSON.stringify(pricingPlanRecords[0])}`);

        // finish formatting coupon codes
        couponCodeRecords = couponCodeRecords.map(couponCode => {
            couponCode.Pricing_Plan__r = { Escapia_Id__c: couponCode.pricingPlanId };
            couponCode.sObjectType = 'Coupon_Code__c';
            delete couponCode.pricingPlanId;
            delete couponCode.pricingPlanName;
            return couponCode;
        });

        logger.verbose(`couponCodeRecords.length: ${couponCodeRecords.length}`);
        logger.verbose(`couponCodeRecords record 0: ${JSON.stringify(couponCodeRecords[0])}`);

        // push these records to mongodb
        await MongoDbCouponCodes.insertMany(escapiaCollectionName, pricingPlanRecords);
        await MongoDbCouponCodes.insertMany(escapiaCollectionName, couponCodeRecords);

        await logoutOfEscapia(page);
        await browser.close();
        logger.verbose('done scraping coupon codes + pricing plans');
        await MongoDbCouponCodes.close();
        return resultsEnum.SUCCESS;
    }
    catch (e) {
        try {
            await logoutOfEscapia(page);
            await browser.close();
        }
        finally {
            logger.error(e);
            await MongoDbCouponCodes.close();
            return resultsEnum.FAILURE;
        }
    }
};