const logger = require('@turnkeyvr/logger');
const MongoDbCouponCodes = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('coupon-codes-prod');
const eachSeries = require('async/eachSeries');
const moment = require('moment');
const { Cluster } = require('puppeteer-cluster');

// global utility methods
const { getBrowser,
    getSelectedValueFromSelectHandle,
    getSelectValue,
    getCheckboxValue,
    getInputValue } = require('../../util/puppeteerUtil');
const { authenticateWithEscapia, setAndGetConfigEscapia, logoutOfEscapia } = require('../../util/escapiaUtil');
const { sleep, resultsEnum, convertArrayToAnArrayOfArrays } = require('../../util/misc');
const { retry } = require('../../util/retry');

// local utility methods
const { customRequestSF } = require('../util/sfdc');

// local constants
const { IS_PROD, DAYS_OF_WEEK } = require('../util/constants');

const getRateRuleConditionRecordTypesMap = async () => {
    let recordTypeNameToIdMap = new Map();
    let rateRuleConditionDescribe = await customRequestSF({
        method: 'GET', 
        url: `/services/data/v47.0/sobjects/Rate_Rule_Condition__c/describe/`
    });
    rateRuleConditionDescribe.recordTypeInfos.forEach(rti => {
        recordTypeNameToIdMap.set(rti.name, rti.recordTypeId);
    });
    return recordTypeNameToIdMap;
};

const getRateRuleActionRecordTypesMap = async () => {
    let recordTypeNameToIdMap = new Map();
    let rateRuleActionDescribe = await customRequestSF({
        method: 'GET', 
        url: `/services/data/v47.0/sobjects/Rate_Rule_Action__c/describe/`
    });
    rateRuleActionDescribe.recordTypeInfos.forEach(rti => {
        recordTypeNameToIdMap.set(rti.name, rti.recordTypeId);
    });
    return recordTypeNameToIdMap;
};

const getSelectId = async (page, selector) => {
    await retry(
        () => Promise.all([
            page.waitForSelector(selector, {visible: true})
        ]), 5
    );
    let selectHandle = await page.$(selector);
    let selectValue = await selectHandle.$$eval('option', nodes => {
        for (let i = 0; i < nodes.length; i++) {
            if (nodes[i].selected) {
                return nodes[i].value;
            }
        }
    });
    return selectValue;
};

// jps note: no records exist (I don't think) with dateType === 'in', unable to test
const scrapeRateRuleBookingDatesSection = async (page, rateRuleConditionRecordTypesMap) => {
    await page.waitForSelector('#upnlBookingDate > table', {visible: true});
    let tableHandle = await page.$('#upnlBookingDate > table');
    let rateRuleConditionRecords = [];
    let rowCount = await page.evaluate(table => table.rows.length, tableHandle);
    if (rowCount > 0) {
        // notice offset for loop; done so string expressions are simpler
        for (let i = 1; i < rowCount + 1; i++) {
            let rateRuleConditionRecord = {};
            let isActive = await getCheckboxValue(page, `#rptBookingDate_ctl0${i}_chkBookingDateIsActive`);

            // if not active, we don't sync
            if (!isActive) {
                continue;
            }

            rateRuleConditionRecord.Is_or_Is_Not__c = await getSelectValue(page, `#rptBookingDate_ctl0${i}_lstBookingIsWithin`);
            rateRuleConditionRecord.Between_In_or_Within__c = await getSelectValue(page, `#rptBookingDate_ctl0${i}_lstBookingDateType`);

            if (rateRuleConditionRecord.Between_In_or_Within__c.toLowerCase() === 'between') {
                rateRuleConditionRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Booking Dates - Date Range');
                rateRuleConditionRecord.Start_Date__c = await getInputValue(page, `input[name="rptBookingDate$ctl0${i}$dpBookingFromDate"`);
                if (rateRuleConditionRecord.Start_Date__c && rateRuleConditionRecord.Start_Date__c !== '') {
                    rateRuleConditionRecord.Start_Date__c = moment(rateRuleConditionRecord.Start_Date__c).format('YYYY-MM-DD');
                }
                rateRuleConditionRecord.End_Date__c = await getInputValue(page, `input[name="rptBookingDate$ctl0${i}$dpBookingToDate"`);
                if (rateRuleConditionRecord.End_Date__c && rateRuleConditionRecord.End_Date__c !== '') {
                    rateRuleConditionRecord.End_Date__c = moment(rateRuleConditionRecord.End_Date__c).format('YYYY-MM-DD');
                }
            }
            else if (rateRuleConditionRecord.Between_In_or_Within__c.toLowerCase() === 'within') {
                rateRuleConditionRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Booking Dates - First Night Of Stay');
                rateRuleConditionRecord.Minimum__c = await getInputValue(page, `#rptBookingDate_ctl0${i}_txtBookingPeriodLength`);
                if (rateRuleConditionRecord.Minimum__c) {
                    rateRuleConditionRecord.Minimum__c = parseInt(rateRuleConditionRecord.Minimum__c);
                }
                rateRuleConditionRecord.Nights_or_Weeks__c = await getSelectValue(page, `#rptBookingDate_ctl0${i}_lstBookingPeriodType`);
            }
            else {
                // rateRuleConditionRecord.Between_In_or_Within__c.toLowerCase() === 'in'
                rateRuleConditionRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Booking Dates - Season');
                rateRuleConditionRecord.Coupon_Code_Season__r = {};
                rateRuleConditionRecord.Coupon_Code_Season__r.Escapia_Id__c = await getSelectId(page, `#rptBookingDate_ctl0${i}_lstBookingDateSeason`);
            }

            rateRuleConditionRecords.push(rateRuleConditionRecord);
        }
    }

    return rateRuleConditionRecords;
};

const scrapeRateRuleStayDatesSection = async (page, rateRuleConditionRecordTypesMap) => {
    // warning! dynamic fields
    await page.waitForSelector('#upnlTravelDates > table');
    let tableHandle = await page.$('#upnlTravelDates > table');
    let stayDateRecords = [];
    let rowCount = await page.evaluate(table => table.rows.length, tableHandle);
    if (rowCount > 0) {
        // notice offset for loop; done so string expressions are simpler
        for (let i = 1; i < rowCount + 1; i++) {
            let stayDateRecord = {};
            let isActive = await getCheckboxValue(page, `#rptTravelDates_ctl0${i}_chkTravelDateIsActive`);
            
            // if not active, we don't sync
            if (!isActive) {
                continue;
            }
            
            stayDateRecord.Night_Type__c = await getSelectValue(page, `#rptTravelDates_ctl0${i}_lstTravelNightType`);
            stayDateRecord.Is_or_Is_Not__c = await getSelectValue(page, `#rptTravelDates_ctl0${i}_lstTravelIsWithin`);
            stayDateRecord.Between_In_or_Within__c = await getSelectValue(page, `#rptTravelDates_ctl0${i}_lstTravelDateType`);

            if (stayDateRecord.Between_In_or_Within__c.toLowerCase() === 'between') {
                stayDateRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Stay Dates - Date Range');
                stayDateRecord.Start_Date__c = await getInputValue(page, `input[name="rptTravelDates$ctl0${i}$dpTravelFromDate"`);
                if (stayDateRecord.Start_Date__c && stayDateRecord.Start_Date__c !== '') {
                    stayDateRecord.Start_Date__c = moment(stayDateRecord.Start_Date__c).format('YYYY-MM-DD');
                }
                stayDateRecord.End_Date__c = await getInputValue(page, `input[name="rptTravelDates$ctl0${i}$dpTravelToDate"`);
                if (stayDateRecord.End_Date__c && stayDateRecord.End_Date__c !== '') {
                    stayDateRecord.End_Date__c = moment(stayDateRecord.End_Date__c).format('YYYY-MM-DD');
                }
            }
            else {
                stayDateRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Stay Dates - Season');
                stayDateRecord.Coupon_Code_Season__r = {};
                stayDateRecord.Coupon_Code_Season__r.Escapia_Id__c = await getSelectId(page, `#rptTravelDates_ctl0${i}_lstTravelDateSeason`);
            }

            stayDateRecords.push(stayDateRecord);
        }
    }

    return stayDateRecords;
};

// jps note: no records exist with dateType === 'in', unable to test
const scrapeRateRuleStayDateRestrictionsSection = async (page, rateRuleConditionRecordTypesMap) => {
    await page.waitForSelector('#upnlTravelDateRestrictions > table', {visible: true});
    let tableHandle = await page.$('#upnlTravelDateRestrictions > table');
    let stayDateRestrictionsRecords = [];
    let rowCount = await page.evaluate(table => table.rows.length, tableHandle);
    if (rowCount > 0) {
        // notice offset for loop; done so string expressions are simpler
        for (let i = 1; i < rowCount + 1; i++) {
            let stayDateRestrictionsRecord = {};
            let isActive = await getCheckboxValue(page, `#rptTravelDateRestrictions_ctl0${i}_chkTravelDateRestrictionIsActive`);
            
            // if not active, we don't sync
            if (!isActive) {
                continue;
            }
            
            stayDateRestrictionsRecord.Night_Type__c = await getSelectValue(page, `#rptTravelDateRestrictions_ctl0${i}_lstTravelDateRestrictionNightType`);
            stayDateRestrictionsRecord.Between_In_or_Within__c = await getSelectValue(page, `#rptTravelDateRestrictions_ctl0${i}_lstTravelDateRestrictionDateType`);

            if (stayDateRestrictionsRecord.Between_In_or_Within__c.toLowerCase() === 'between') {
                stayDateRestrictionsRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Stay Date Restrictions - Date Range');
                stayDateRestrictionsRecord.Start_Date__c = await getInputValue(page, `input[name="rptTravelDateRestrictions$ctl0${i}$dpTravelDateRestrictionFromDate"`);
                if (stayDateRestrictionsRecord.Start_Date__c && stayDateRestrictionsRecord.Start_Date__c !== '') {
                    stayDateRestrictionsRecord.Start_Date__c = moment(stayDateRestrictionsRecord.Start_Date__c).format('YYYY-MM-DD');
                }
                stayDateRestrictionsRecord.End_Date__c = await getInputValue(page, `input[name="rptTravelDateRestrictions$ctl0${i}$dpTravelDateRestrictionToDate"`);
                if (stayDateRestrictionsRecord.End_Date__c && stayDateRestrictionsRecord.End_Date__c !== '') {
                    stayDateRestrictionsRecord.End_Date__c = moment(stayDateRestrictionsRecord.End_Date__c).format('YYYY-MM-DD');
                }
            }
            else {
                // this value would be "in"
                stayDateRestrictionsRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Stay Date Restrictions - Season');
                stayDateRestrictionsRecord.Coupon_Code_Season__r = {};
                stayDateRestrictionsRecord.Coupon_Code_Season__r.Escapia_Id__c = await getSelectId(page, `#rptTravelDateRestrictions_ctl0${i}_lstTravelDateRestrictionSeason`);
            }

            stayDateRestrictionsRecords.push(stayDateRestrictionsRecord);
        }
    }

    return stayDateRestrictionsRecords;
};

const scrapeRateRuleStayDaysSection = async (page, rateRuleConditionRecordTypesMap) => {
    let stayDaysRecords = [];

    let daysToSFDCDayFields = {
        'Monday': 'Monday__c',
        'Tuesday': 'Tuesday__c',
        'Wednesday': 'Wednesday__c',
        'Thursday': 'Thursday__c',
        'Friday': 'Friday__c',
        'Saturday': 'Saturday__c',
        'Sunday': 'Sunday__c'
    };

    // Stay Days - Nights
    let travelNightsIsActive = await getCheckboxValue(page, '#chkTravelNightsIsActive');
    if (travelNightsIsActive) {
        let stayDaysRecord = {};
        stayDaysRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Stay Days - Nights');
        stayDaysRecord.Includes__c = await getSelectValue(page, '#lstTravelNightsOperator');
        await eachSeries(DAYS_OF_WEEK, async day => {
            let dayCheckboxSelector = '#chkTravelNights' + day;
            stayDaysRecord[daysToSFDCDayFields[day]] = await getCheckboxValue(page, dayCheckboxSelector);
        });
        stayDaysRecords.push(stayDaysRecord);
    }

    // Stay Days - First Night
    let travelFirstNightIsActive = await getCheckboxValue(page, '#chkTravelFirstNightIsActive');
    if (travelFirstNightIsActive) {
        let stayDaysRecord = {};
        stayDaysRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Stay Days - First Night');
        stayDaysRecord.Is_or_Is_Not__c = await getSelectValue(page, '#lstTravelFirstNightOperator');
        await eachSeries(DAYS_OF_WEEK, async day => {
            let dayCheckboxSelector = '#chkTravelFirstNight' + day;
            stayDaysRecord[daysToSFDCDayFields[day]]  = await getCheckboxValue(page, dayCheckboxSelector);
        });
        stayDaysRecords.push(stayDaysRecord);
    }

    // Stay Days - Last Night
    let travelLastNightIsActive = await getCheckboxValue(page, '#chkTravelLastNightIsActive');
    if (travelLastNightIsActive) {
        let stayDaysRecord = {};
        stayDaysRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Stay Days - Last Night');
        stayDaysRecord.Is_or_Is_Not__c = await getSelectValue(page, '#lstTravelLastNightOperator');
        await eachSeries(DAYS_OF_WEEK, async day => {
            let dayCheckboxSelector = '#chkTravelLastNight' + day;
            stayDaysRecord[daysToSFDCDayFields[day]]  = await getCheckboxValue(page, dayCheckboxSelector);
        });
        stayDaysRecords.push(stayDaysRecord);
    }

    // Stay Days - Length Of Stay
    let travelLengthOfStayIsActive = await getCheckboxValue(page, '#chkTravelLengthOfStayIsActive');
    if (travelLengthOfStayIsActive) {
        let stayDaysRecord = {};
        stayDaysRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Stay Days - Length Of Stay');
        stayDaysRecord.Minimum__c = await getInputValue(page, '#txtTravelLengthOfStayFrom');
        if (stayDaysRecord.Minimum__c) {
            stayDaysRecord.Minimum__c = parseInt(stayDaysRecord.Minimum__c);
        }
        stayDaysRecord.Maximum__c = await getInputValue(page, '#txtTravelLengthOfStayTo');
        if (stayDaysRecord.Maximum__c) {
            stayDaysRecord.Maximum__c = parseInt(stayDaysRecord.Maximum__c);
        }
        stayDaysRecords.push(stayDaysRecord);
    }

    return stayDaysRecords;
};

const scrapeRateRuleOtherSection = async (page, rateRuleConditionRecordTypesMap) => {   
    let otherRecords = [];

    // Other - Reservation Type
    let reservationTypeIsActive = await getCheckboxValue(page, '#chkReservationTypeIsActive');
    if (reservationTypeIsActive) {
        let otherRecord = {};
        otherRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Other - Reservation Type');
        otherRecord.Reservation_Type__c = '';
        let escapiaReservationTypeToSFDCPicklistValue = {
            'Renter': 'Renter',
            'Owner': 'Owner',
            'TravelAgent': 'Travel Agent',
            'LongTerm': 'Long Term',
            'Group': 'Group',
            'OwnerGuest': 'Owner Guest',
            'Referral': 'Owner Referral',
            'Complimentary': 'Complimentary'
        };
        await eachSeries(['Renter', 'Owner', 'TravelAgent', 'LongTerm', 'Group', 'OwnerGuest', 'Referral', 'Complimentary'], async reservationType => {
            let isChecked = await getCheckboxValue(page, `#chkReservationType${reservationType}`);
            if (isChecked) {
                let value = escapiaReservationTypeToSFDCPicklistValue[reservationType];
                if (otherRecord.Reservation_Type__c === '') {
                    otherRecord.Reservation_Type__c += value;
                }
                else {
                    otherRecord.Reservation_Type__c += `;${value}`;
                }
            }
        });
        otherRecords.push(otherRecord);
    }

    // Other - Unit
    let unitIsActive = await getCheckboxValue(page, '#chkUnitIsActive');
    if (unitIsActive) {
        let otherRecord = {};
        otherRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Other - Unit');
        otherRecord.Is_or_Is_Not__c = await getSelectValue(page, '#lstUnitOperator');
        otherRecord.Coupon_Code_Units__r = []; // units (ex: { Rental_Name__c: '14CPLVR5'})
        otherRecord.Rate_Rule_Condition_Unit_Groups__r = []; // unit groups (ex: { Unit_Group_Name__c: 'AL - Gulf Shores'})
        await page.waitForSelector('#lbxUnits', {visible: true});
        let unitsSelectHandle = await page.$('#lbxUnits');
        let selectedUnitsAndGroups = await unitsSelectHandle.$$eval('option', nodes => {
            let unitsAndGroups = [];
            for (let i = 0; i < nodes.length; i++) {
                unitsAndGroups.push(nodes[i].text.trim()); // using "text" b/c textContent and value include extra white spaces sometimes -- weird escapia thing
            }
            return unitsAndGroups;
        });
        selectedUnitsAndGroups.forEach(u => {
            if (u.startsWith('*')) {
                otherRecord.Rate_Rule_Condition_Unit_Groups__r.push({ Unit_Group_Name__c: u.split(/\*(.+)/)[1] }); // unit group
            }
            else {
                otherRecord.Coupon_Code_Units__r.push({ Rental_Name__c: u.split(/\s-(.+)/)[0] }); // unit
            }
        });
        otherRecords.push(otherRecord);
    }

    // JPS NOTE: as of Feb 5 2020 there are no occupany records; not supporting for now
    // this one is a total nightmare due to changing inputs
    let occupancyIsActive = await getCheckboxValue(page, '#chkOccupancyIsActive');
    if (occupancyIsActive) {
        logger.alert('Scraping Rate Rules: Occupancy values active');
    }

    // Other - Number Of Guests
    let numberOfGuestsIsActive = await getCheckboxValue(page, '#chkNumberOfGuestsIsActive');
    if (numberOfGuestsIsActive) {
        let otherRecord = {};
        otherRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Other - Number Of Guests');
        otherRecord.Minimum__c = await getInputValue(page, '#txtNumberOfGuestsFrom');
        if (otherRecord.Minimum__c) {
            otherRecord.Minimum__c = parseInt(otherRecord.Minimum__c);
        }
        otherRecord.Maximum__c = await getInputValue(page, '#txtNumberOfGuestsTo');
        if (otherRecord.Maximum__c) {
            otherRecord.Maximum__c = parseInt(otherRecord.Maximum__c);
        }
        otherRecords.push(otherRecord);
    }

    // 	Other - Number Of Adults
    let numberOfAdultsIsActive = await getCheckboxValue(page, '#chkNumberOfAdultsIsActive');
    if (numberOfAdultsIsActive) {
        let otherRecord = {};
        otherRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Other - Number Of Adults');
        otherRecord.Minimum__c = await getInputValue(page, '#txtNumberOfAdultsFrom');
        if (otherRecord.Minimum__c) {
            otherRecord.Minimum__c = parseInt(otherRecord.Minimum__c);
        }
        otherRecord.Maximum__c = await getInputValue(page, '#txtNumberOfAdultsTo');
        if (otherRecord.Maximum__c) {
            otherRecord.Maximum__c = parseInt(otherRecord.Maximum__c);
        }
        otherRecords.push(otherRecord);
    }

    // Other - Number Of Children
    let numberOfChildrenIsActive = await getCheckboxValue(page, '#chkNumberOfChildrenIsActive');
    if (numberOfChildrenIsActive) {
        let otherRecord = {};
        otherRecord.RecordTypeId = rateRuleConditionRecordTypesMap.get('Other - Number Of Children');
        otherRecord.Minimum__c = await getCheckboxValue(page, '#txtNumberOfChildrenFrom');
        if (otherRecord.Minimum__c) {
            otherRecord.Minimum__c = parseInt(otherRecord.Minimum__c);
        }
        otherRecord.Maximum__c = await getCheckboxValue(page, '#txtNumberOfChildrenTo');
        if (otherRecord.Maximum__c) {
            otherRecord.Maximum__c = parseInt(otherRecord.Maximum__c);
        }
        otherRecords.push(otherRecord);
    }

    return otherRecords;
};

// JPS NOTE: right now (Feb 2020) we only support one action, and it has to be one of the following record types: Decrease Rate By, Set Rate To
// however, semi-coding in potential support for multiple actions in the future
const scrapeRateRuleActionsSection = async (page, rateRuleActionRecordTypesMap) => {
    let actionRecords = [];
    let actionRecord = {};

    let daysToSFDCDayFields = {
        'Monday': 'Applicable_to_Monday__c',
        'Tuesday': 'Applicable_to_Tuesday__c',
        'Wednesday': 'Applicable_to_Wednesday__c',
        'Thursday': 'Applicable_to_Thursday__c',
        'Friday': 'Applicable_to_Friday__c',
        'Saturday': 'Applicable_to_Saturday__c',
        'Sunday': 'Applicable_to_Sunday__c'
    };

    await retry(
        () => Promise.all([
            page.waitForSelector('#rptStandardActions_ctl00_lstActionType', {visible: true})
        ]), 5
    );
    let actionTypeSelectHandle = await page.$('#rptStandardActions_ctl00_lstActionType');
    let actionTypeSelectValue = await actionTypeSelectHandle.$$eval('option', getSelectedValueFromSelectHandle);

    if (!actionTypeSelectValue || !['Decrease Rate By', 'Set Rate To'].includes(actionTypeSelectValue)) {
        logger.alert('Unrecognized rate rule action type: ', actionTypeSelectValue);
        return [];
    }

    actionRecord.RecordTypeId = rateRuleActionRecordTypesMap.get(actionTypeSelectValue);
    actionRecord.Sequence_Number__c = parseInt(await getSelectValue(page, '#rptStandardActions_ctl00_lstStandardActionSequence'));

    if (actionTypeSelectValue === 'Decrease Rate By') {
        actionRecord.Amount__c = await getInputValue(page, '#rptStandardActions_ctl00_txtAmount');
        if (actionRecord.Amount__c) {
            actionRecord.Amount__c = parseFloat(parseFloat(actionRecord.Amount__c).toFixed(2));
        }
        actionRecord.Amount_Unit_of_Measure__c = await getSelectValue(page, '#rptStandardActions_ctl00_lstAdjustedRateUOM');
        if (actionRecord.Amount_Unit_of_Measure__c === '$') {
            actionRecord.Adjusted_Rate_Duration__c = await getSelectValue(page, '#rptStandardActions_ctl00_lstAdjustRateDuration');
        }

        actionRecord.Scope__c = await getSelectValue(page, '#rptStandardActions_ctl00_lstRateActionScope');
        actionRecord.Period_Type__c = await getSelectValue(page, '#rptStandardActions_ctl00_lstRateActionPeriodType');

        // actionRecord.Period_Type__c sets dynamic fields
        if (actionRecord.Period_Type__c === 'Specific Nights' || actionRecord.Period_Type__c === 'Specific Weeks') {
            actionRecord.Period_From__c = await getInputValue(page, '#rptStandardActions_ctl00_txtActionPeriodFrom');
            if (actionRecord.Period_From__c) {
                actionRecord.Period_From__c = parseInt(actionRecord.Period_From__c);
            }
            actionRecord.Period_to__c = await getInputValue(page, '#rptStandardActions_ctl00_txtActionPeriodTo');
            if (actionRecord.Period_to__c) {
                actionRecord.Period_to__c = parseInt(actionRecord.Period_to__c);
            }
        }
        else if (actionRecord.Period_Type__c === 'Specific Date') {
            actionRecord.Specific_Date__c = await getInputValue(page, `input[name="rptStandardActions$ctl00$dpActionPeriodDate"`);
        }

        await eachSeries(DAYS_OF_WEEK, async day => {
            actionRecord[daysToSFDCDayFields[day]] = await getCheckboxValue(page, `#rptStandardActions_ctl00_chkApplicable${day}`);
        });
    }
    else if (actionTypeSelectValue === 'Set Rate To') {
        actionRecord.Night_Rate__c = await getInputValue(page, '#rptStandardActions_ctl00_txtNightRate');
        if (actionRecord.Night_Rate__c) {
            actionRecord.Night_Rate__c = parseFloat(parseFloat(actionRecord.Night_Rate__c).toFixed(2));
        }
        actionRecord.Week_Rate__c = await getInputValue(page, '#rptStandardActions_ctl00_txtWeekRate');
        if (actionRecord.Week_Rate__c) {
            actionRecord.Week_Rate__c = parseFloat(parseFloat(actionRecord.Week_Rate__c).toFixed(2));
        }
        actionRecord.Month_Rate__c = await getInputValue(page, '#rptStandardActions_ctl00_txtMonthRate');
        if (actionRecord.Month_Rate__c) {
            actionRecord.Month_Rate__c = parseFloat(parseFloat(actionRecord.Month_Rate__c).toFixed(2));
        }
        actionRecord.Term_Rate__c = await getInputValue(page, '#rptStandardActions_ctl00_txtTermRate');
        if (actionRecord.Term_Rate__c) {
            actionRecord.Term_Rate__c = parseFloat(parseFloat(actionRecord.Term_Rate__c).toFixed(2));
        }

        actionRecord.Scope__c = await getSelectValue(page, '#rptStandardActions_ctl00_lstRateActionScope');
        actionRecord.Period_Type__c = await getSelectValue(page, '#rptStandardActions_ctl00_lstRateActionPeriodType');

        // actionRecord.Period_Type__c sets dynamic fields
        if (actionRecord.Period_Type__c === 'Specific Nights' || actionRecord.Period_Type__c === 'Specific Weeks') {
            actionRecord.Period_From__c = await getInputValue(page, '#rptStandardActions_ctl00_txtActionPeriodFrom');
            if (actionRecord.Period_From__c) {
                actionRecord.Period_From__c = parseInt(actionRecord.Period_From__c);
            }
            actionRecord.Period_to__c = await getInputValue(page, '#rptStandardActions_ctl00_txtActionPeriodTo');
            if (actionRecord.Period_to__c) {
                actionRecord.Period_to__c = parseInt(actionRecord.Period_to__c);
            }
        }
        else if (actionRecord.Period_Type__c === 'Specific Date') {
            actionRecord.Specific_Date__c = await getInputValue(page, `input[name="rptStandardActions$ctl00$dpActionPeriodDate"`);
        }

        await eachSeries(DAYS_OF_WEEK, async day => {
            actionRecord[daysToSFDCDayFields[day]] = await getCheckboxValue(page, `#rptStandardActions_ctl00_chkApplicable${day}`);
        });
    }

    actionRecords.push(actionRecord);

    return actionRecords;
};

// this one is a little different, need to return a map of pricing plan id -> array of rate rule ids
const scrapeRateRuleIds = async (page, escapiaCollectionName) => {
    // get pricing plan escapia ids from escapia mongo collection
    let escapiaPricingPlanRecords = await MongoDbCouponCodes.find(escapiaCollectionName, {sObjectType: 'Pricing_Plan__c'});
    let pricingPlanIds = escapiaPricingPlanRecords.map(r => r.Escapia_Id__c);

    // temp testing
    // let tempIds = [];
    // for (let i = 0; i < 5; i++) {
    //     tempIds.push(pricingPlanIds[i]);
    // }
    // pricingPlanIds = tempIds;

    let pricingPlanIdToRateRuleIdsMap = new Map();
    for (let i = 0; i < pricingPlanIds.length; i++) {
        let pricingPlanId = pricingPlanIds[i];

        await retry(
            () => Promise.all([
                page.goto('https://turnkeyvr.escapia.com/WebAgency/PricingPlan/Edit?id=' + pricingPlanId + '&travelDate=&unitID=0', {waitUntil: 'networkidle0'}),
                page.waitForSelector('#RateRuleTable')
            ]), 5
        );

        let rateRuleTableHandle = await page.$('#RateRuleTable');
        let rateRuleIds = await rateRuleTableHandle.$$eval('tbody tr.clickable', async nodes => {
            return nodes.map(tr => tr.dataset['uniqueid']);
        });

        pricingPlanIdToRateRuleIdsMap.set(pricingPlanId, rateRuleIds);
    }

    return pricingPlanIdToRateRuleIdsMap;
}

const scrapeRateRuleRecords = async ({ page, data }) => {
    // idMapArray: [Map(Pricing Plan Id, [Rate Rule Ids])]
    let { idMapArray, rateRuleRecords, scrapingErrors } = data;
    try {
        // get Rate_Rule_Condition__c record types map (name -> id)
        let rateRuleConditionRecordTypesMap = await getRateRuleConditionRecordTypesMap();
        // get Rate_Rule_Action__c record types map (name -> id)
        let rateRuleActionRecordTypesMap = await getRateRuleActionRecordTypesMap();

        setAndGetConfigEscapia('1655'); // only running against the main instance
        page = await authenticateWithEscapia(page);

        await eachSeries(idMapArray, async idMap => {
            try {
                let pricingPlanId = Array.from(idMap.keys())[0];
                let rateRuleIds = idMap.get(pricingPlanId);

                for (let i = 0; i < rateRuleIds.length; i++) {
                    let rateRuleId = rateRuleIds[i];
                    await page.goto(`https://turnkeyvr.escapia.com/WebAgency/Property/Setup/Rates/OpenRateRule.aspx?planID=${pricingPlanId}&ruleID=${rateRuleId}&ruleType=0&travelDate=&unitID=0&seasonID=0`, {waitUntil: 'networkidle0'});

                    let rateRuleRecord = {
                        sObjectType: 'Rate_Rule__c',
                        Pricing_Plan__r: { Escapia_Id__c: pricingPlanId },
                        Escapia_Id__c: rateRuleId
                    };

                    await retry(
                        () => Promise.all([
                            page.waitForSelector('#imgRateDetails', {visible: true})
                        ]), 5
                    );
                    let rateRuleInfoExpansionHandler = await page.$('#imgRateDetails');
                    await retry(
                        () => Promise.all([
                            rateRuleInfoExpansionHandler.click(),
                            page.waitForSelector('#txtName', {visible: true}),
                            page.waitForSelector('#lstStatus', {visible: true}),
                            page.waitForSelector('#txtDescription', {visible: true})
                        ]), 5
                    );
                    rateRuleRecord.Name = await getInputValue(page, '#txtName');
                    rateRuleRecord.Status__c = await getSelectValue(page, '#lstStatus');
                    rateRuleRecord.Description__c = await getInputValue(page, '#txtDescription');
                    rateRuleRecord.Sequence_Number__c = await page.$$eval('.lineDesc', nodes => {
                        for (let i = 0; i < nodes.length; i++) {
                            if (nodes[i].textContent.trim() !== 'Sequence') {
                                continue;
                            }
                            return parseInt(nodes[i].parentElement.querySelector('.lineAmt').textContent.trim());
                        }
                        return null;
                    });


                    await retry(
                        () => Promise.all([
                            page.waitForSelector('#rptStandardActions_ctl00_lstActionType', {visible: true})
                        ]), IS_PROD ? 5 : 2
                    );
                    
                    rateRuleRecord.Rate_Rule_Conditions__r = [];

                    // Booking Dates
                    let bookingDates = await scrapeRateRuleBookingDatesSection(page, rateRuleConditionRecordTypesMap);
                    bookingDates.forEach(record => {
                        rateRuleRecord.Rate_Rule_Conditions__r.push(record);
                    });

                    // Stay Dates
                    let stayDates = await scrapeRateRuleStayDatesSection(page, rateRuleConditionRecordTypesMap);
                    stayDates.forEach(record => {
                        rateRuleRecord.Rate_Rule_Conditions__r.push(record);
                    });

                    // Stay Date Restrictions
                    let stayDateRestrictions = await scrapeRateRuleStayDateRestrictionsSection(page, rateRuleConditionRecordTypesMap);
                    stayDateRestrictions.forEach(record => {
                        rateRuleRecord.Rate_Rule_Conditions__r.push(record);
                    });

                    // Stay Days Section
                    let stayDays = await scrapeRateRuleStayDaysSection(page, rateRuleConditionRecordTypesMap);
                    stayDays.forEach(record => {
                        rateRuleRecord.Rate_Rule_Conditions__r.push(record);
                    });

                    // Other Section
                    let otherConditions = await scrapeRateRuleOtherSection(page, rateRuleConditionRecordTypesMap);
                    otherConditions.forEach(record => {
                        rateRuleRecord.Rate_Rule_Conditions__r.push(record);
                    });
                    

                    rateRuleRecord.Rate_Rule_Actions__r = [];

                    // Actions Section
                    let rateRuleActions = await scrapeRateRuleActionsSection(page, rateRuleActionRecordTypesMap);
                    rateRuleActions.forEach(record => {
                        rateRuleRecord.Rate_Rule_Actions__r.push(record);
                    })


                    rateRuleRecords.push(rateRuleRecord);
                }
            }
            catch (e) {
                scrapingErrors.push(e);
            }
        });

        await logoutOfEscapia(page);
    }
    catch (e) {
        scrapingErrors.push(e);
    }
    return;
};

const getRateRuleRecords = async (pricingPlanIdToRateRuleIdsMap) => {
    let rateRuleRecords = [];
    let scrapingErrors = [];

    // concurrency cluster
    const cluster = await Cluster.launch({
        concurrency: Cluster.CONCURRENCY_CONTEXT,
        maxConcurrency: 6,
        puppeteerOptions: {
            headless: IS_PROD,
        },
        timeout: 60 * 1000 * 45
    });

    // convert map to individual maps and add to array
    let mapArray = [];
    Array.from(pricingPlanIdToRateRuleIdsMap.keys()).forEach(pricingPlanId => {
        let newMap = new Map();
        newMap.set(pricingPlanId, pricingPlanIdToRateRuleIdsMap.get(pricingPlanId));
        mapArray.push(newMap);
    });

    let idMapArrayArray = convertArrayToAnArrayOfArrays(mapArray, 50); // an array of arrays

    for (let i = 0; i < idMapArrayArray.length; i++) {
        cluster.queue({ idMapArray: idMapArrayArray[i], rateRuleRecords: rateRuleRecords, scrapingErrors: scrapingErrors }, scrapeRateRuleRecords);
    }

    await cluster.idle();
    await cluster.close();

    if (scrapingErrors.length > 0) {
        throw scrapingErrors[0];
    }

    return rateRuleRecords;
};


module.exports = async function(escapiaCollectionName) {
    let browser;
    let page;
    try {
        // load browser
        browser = await getBrowser();
        if (!browser) {
            return;
        }

        // authenticate with escapia
        setAndGetConfigEscapia('1655'); // only running against the main instance
        page = await authenticateWithEscapia(await browser.newPage());
        if (!page) {
            await browser.close();
            return;
        }

        // scrape rate rule ids (only those associated to coupon pricing plans)
        let pricingPlanIdToRateRuleIdsMap = await scrapeRateRuleIds(page, escapiaCollectionName);

        logger.verbose(`pricingPlanIdToRateRuleIdsMap.length: ${pricingPlanIdToRateRuleIdsMap.size}`);
        logger.verbose(`pricingPlanIdToRateRuleIdsMap record 0: ${pricingPlanIdToRateRuleIdsMap.get(Array.from(pricingPlanIdToRateRuleIdsMap.keys())[0])}`);

        let rateRuleRecords = await getRateRuleRecords(pricingPlanIdToRateRuleIdsMap);

        // push these records to mongodb
        await MongoDbCouponCodes.insertMany(escapiaCollectionName, rateRuleRecords);

        await logoutOfEscapia(page);
        await browser.close();
        logger.verbose('done scraping rate rules');
        await MongoDbCouponCodes.close();
        return resultsEnum.SUCCESS;
    }
    catch (e) {
        try {
            await logoutOfEscapia(page);
            await browser.close();
        }
        finally {
            logger.error(e);
            await MongoDbCouponCodes.close();
            return resultsEnum.FAILURE;
        }
    }
}