const logger = require('@turnkeyvr/logger');
const MongoDbCouponCodes = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('coupon-codes-prod');
const eachSeries = require('async/eachSeries');
const moment = require('moment');
const { Cluster } = require('puppeteer-cluster');

// global utility methods
const { getBrowser,
    getSelectedValueFromSelectHandle,
    getSelectValue,
    getCheckboxValue,
    getInputValue } = require('../../util/puppeteerUtil');
const { authenticateWithEscapia, setAndGetConfigEscapia, logoutOfEscapia } = require('../../util/escapiaUtil');
const { sleep, resultsEnum, convertArrayToAnArrayOfArrays } = require('../../util/misc');
const { retry } = require('../../util/retry');

// local utility methods
const { IS_PROD } = require('../util/constants');

const scrapeSeasonIds = async (page) => {
    let seasonIds = [];

    await page.goto('https://turnkeyvr.escapia.com/webagency/Admin/Seasons/ManageSeasons.aspx', { waitUntil: 'networkidle0' });

    let gatheredAllResults = false;
    while (!gatheredAllResults) {
        let elementHandleArray = await page.$$('.listRow');
        let firstIdValue = await page.evaluate(e => e.getAttribute('rowarg'), elementHandleArray[0]);
        if (seasonIds.includes(firstIdValue)) {
            gatheredAllResults = true;
        }
        else {
            await eachSeries(elementHandleArray, async e => {
                let idValue = await page.evaluate(e => e.getAttribute('rowarg'), e);
                seasonIds.push(idValue);
            });

            // click next
            let paginationSpanHandle = await page.$('#ctlDataPager');
            // await paginationSpanHandle.$$eval('a', nodes => nodes.map(n => {
            //     if (n.textContent !== 'NEXT') {
            //         return;
            //     }
            //     n.click();
            // }));

            // await page.waitForNavigation();

            await Promise.all([
                paginationSpanHandle.$$eval('a', nodes => nodes.map(n => {
                    if (n.textContent !== 'NEXT') {
                        return;
                    }
                    n.click();
                })),
                page.waitForNavigation()
            ]);
        }
    }

    return seasonIds;
};

const getSeasonAndDateRangeRecords = async (seasonIds) => {
    let seasonRecords = [];
    let scrapingErrors = [];

    // concurrency cluster
    const cluster = await Cluster.launch({
        concurrency: Cluster.CONCURRENCY_CONTEXT,
        maxConcurrency: 10,
        puppeteerOptions: {
            headless: IS_PROD,
        },
        timeout: 60 * 1000 * 30
    });

    logger.debug('totalSeasons: ', seasonIds.length);

    let seasonIdArrayArray = convertArrayToAnArrayOfArrays(seasonIds, 200); // an array of arrays

    const scrapeSeasonsAndDateRanges = async ({ page, data }) => {
        let { seasonIds, seasonRecords, scrapingErrors } = data;

        try {
            await retry(
                () => Promise.all([
                    setAndGetConfigEscapia('1655'),
                    authenticateWithEscapia(page)
                ]), 5
            );

            await eachSeries(seasonIds, async seasonId => {
                try {
                    let seasonRecord = {
                        Escapia_Id__c: seasonId,
                        Coupon_Code_Season_Date_Ranges__r: []
                    };
                    await page.goto('https://turnkeyvr.escapia.com/webagency/Admin/Seasons/OpenSeason.aspx?id=' + seasonId, { waitUntil: 'networkidle0', timeout: 0 });
                    seasonRecord.Name = await getInputValue(page, '#txtSeasonName');
                    await page.waitForSelector('#scrollDiv > table', { visible: true });
                    let tableHandle = await page.$('#scrollDiv > table');
                    let rowCount = await page.evaluate(table => table.rows.length, tableHandle);
                    if (rowCount > 0) {
                        // notice offset for loop; done so string expressions are simpler
                        for (let i = 1; i < rowCount + 1; i++) {
                            let dateRangeRecord = {};
                            let rowNumber = i < 10 ? `0${i}` : `${i}`;
                            dateRangeRecord.Start_Date__c = await getInputValue(page, `input[name="rptPeriod$ctl${rowNumber}$dpStartDate"`);
                            if (dateRangeRecord.Start_Date__c && dateRangeRecord.Start_Date__c !== '') {
                                dateRangeRecord.Start_Date__c = moment(dateRangeRecord.Start_Date__c).format('YYYY-MM-DD');
                            }
                            dateRangeRecord.End_Date__c = await getInputValue(page, `input[name="rptPeriod$ctl${rowNumber}$dpEndDate"`);
                            if (dateRangeRecord.End_Date__c && dateRangeRecord.End_Date__c !== '') {
                                dateRangeRecord.End_Date__c = moment(dateRangeRecord.End_Date__c).format('YYYY-MM-DD');
                            }
                            seasonRecord.Coupon_Code_Season_Date_Ranges__r.push(dateRangeRecord);
                        }
                    }
                    seasonRecords.push(seasonRecord);
                }
                catch (e) {
                    scrapingErrors.push(e);
                }
            });

            await logoutOfEscapia(page);
        }
        catch (e) {
            scrapingErrors.push(e);
        }
    };

    for (let i = 0; i < seasonIdArrayArray.length; i++) {
        cluster.queue({ seasonIds: seasonIdArrayArray[i], seasonRecords: seasonRecords, scrapingErrors: scrapingErrors }, scrapeSeasonsAndDateRanges);
    }

    await cluster.idle();
    await cluster.close();

    if (scrapingErrors.length > 0) {
        throw scrapingErrors[0];
    }

    return seasonRecords;
};

module.exports = async function(escapiaCollectionName) {
    let browser;
    let page;
    try {
        // load browser
        browser = await getBrowser();
        if (!browser) {
            return;
        }

        // authenticate with escapia
        setAndGetConfigEscapia('1655'); // only running against the main instance
        page = await authenticateWithEscapia(await browser.newPage());
        if (!page) {
            await browser.close();
            return;
        }

        let seasonIds = await scrapeSeasonIds(page);

        // temp testing
        // let tempIds = [];
        // for (let i = 0; i < 5; i++) {
        //     tempIds.push(seasonIds[i]);
        // }
        // seasonIds = tempIds;

        // scrape Coupon_Code_Season__c and Coupon_Code_Season_Date_Range__c records
        let seasonAndDateRangeRecords = await getSeasonAndDateRangeRecords(seasonIds);
        
        // add sObjectType
        seasonAndDateRangeRecords = seasonAndDateRangeRecords.map(season => {
            season.sObjectType = 'Coupon_Code_Season__c';
            return season;
        });

        logger.verbose('seasonRecords.length: ', seasonAndDateRangeRecords.length);
        logger.verbose('first season record: ', seasonAndDateRangeRecords[0]);

        // push these records to mongodb
        await MongoDbCouponCodes.insertMany(escapiaCollectionName, seasonAndDateRangeRecords);

        await logoutOfEscapia(page);
        await browser.close();
        logger.verbose('done scraping seasons');
        await MongoDbCouponCodes.close();
        return resultsEnum.SUCCESS;
    }
    catch (e) {
        try {
            await logoutOfEscapia(page);
            await browser.close();
        }
        finally {
            logger.error(e);
            await MongoDbCouponCodes.close();
            return resultsEnum.FAILURE;
        }
    }
}