const logger = require('@turnkeyvr/logger');
const MongoDbCouponCodes = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('coupon-codes-prod');
const eachSeries = require('async/eachSeries');
const moment = require('moment');
const { Cluster } = require('puppeteer-cluster');

// global utility methods
const { getBrowser,
    getSelectedValueFromSelectHandle,
    getSelectValue,
    getCheckboxValue,
    getInputValue } = require('../../util/puppeteerUtil');
const { authenticateWithEscapia, setAndGetConfigEscapia, logoutOfEscapia } = require('../../util/escapiaUtil');
const { sleep, resultsEnum, convertArrayToAnArrayOfArrays } = require('../../util/misc');
const { retry } = require('../../util/retry');

// local utility methods
const { IS_PROD } = require('../util/constants');

const scrapeUnitGroupIds = async (page) => {
    let unitGroupIds = [];

    await page.goto('https://turnkeyvr.escapia.com/webagency/Property/ManageUnitGroups.aspx?ugt=1&ugmt=1', { waitUntil: 'networkidle0', timeout: 0 });

    let gatheredAllResults = false;
    while (!gatheredAllResults) {
        let elementHandleArray = await page.$$('.listRow');
        let firstIdValue = await page.evaluate(e => e.getAttribute('rowarg'), elementHandleArray[0]);
        if (unitGroupIds.includes(firstIdValue)) {
            gatheredAllResults = true;
        }
        else {
            await eachSeries(elementHandleArray, async e => {
                let idValue = await page.evaluate(e => e.getAttribute('rowarg'), e);
                unitGroupIds.push(idValue);
            });

            // click next
            let paginationSpanHandle = await page.$('#ctlDataPager');
            // await paginationSpanHandle.$$eval('a', nodes => nodes.map(n => {
            //     if (n.textContent !== 'NEXT') {
            //         return;
            //     }
            //     n.click();
            // }));

            // await page.waitForNavigation();
            await Promise.all([
                paginationSpanHandle.$$eval('a', nodes => nodes.map(n => {
                    if (n.textContent !== 'NEXT') {
                        return;
                    }
                    n.click();
                })),
                page.waitForNavigation()
            ]);
        }
    }

    return unitGroupIds;
};

const getUnitAndUnitGroupRecords = async (unitGroupIds) => {
    let unitGroupRecords = [];
    let scrapingErrors = [];

    // concurrency cluster
    const cluster = await Cluster.launch({
        concurrency: Cluster.CONCURRENCY_CONTEXT,
        maxConcurrency: 6,
        puppeteerOptions: {
            headless: IS_PROD,
        },
        timeout: 60 * 1000 * 30
    });

    let unitGroupIdArrayArray = convertArrayToAnArrayOfArrays(unitGroupIds, 50); // an array of arrays

    const scrapeUnitAndUnitGroupRecords = async ({ page, data }) => {
        let { unitGroupIds, unitGroupRecords, scrapingErrors } = data;

        try {
            setAndGetConfigEscapia('1655'); // only running against the main instance
            page = await authenticateWithEscapia(page);

            await eachSeries(unitGroupIds, async unitGroupId => {
                try {
                    let unitGroupRecord = {
                        Escapia_Id__c: unitGroupId,
                        Coupon_Code_Units__r: []
                    };
                    try {
                        await retry(
                            () => Promise.all([
                                page.goto('https://turnkeyvr.escapia.com/WebAgency/Property/OpenUnitGroup.aspx?id=' + unitGroupId, { waitUntil: 'networkidle0', timeout: 60 * 1000 * 2 })
                            ]), 5
                        );
                    }
                    catch (e) {
                        logger.error('failed to open unit group: ', unitGroupId);
                        logger.error(e.message);
                        logger.error(e.stack);
                    }

                    unitGroupRecord.Name = await getInputValue(page, '#txtName');
                    unitGroupRecord.Status__c = await getSelectValue(page, '#lstStatus');

                    // get all the values in the "Chosen Units" select section
                    await retry(
                        () => Promise.all([
                            page.waitForSelector('#dlstUnits_RightBox', { visible: true })
                        ]), 5
                    );
                    let selectHandle = await page.$('#dlstUnits_RightBox');
                    unitGroupRecord.Coupon_Code_Units__r = await selectHandle.$$eval('option', nodes => {
                        return nodes.map(n => n.textContent.split(' - ')[0].trim());
                    });
                    // format to match sfdc
                    unitGroupRecord.Coupon_Code_Units__r = unitGroupRecord.Coupon_Code_Units__r.map(rentalName => {
                        return { Rental_Name__c: rentalName };
                    });

                    unitGroupRecords.push(unitGroupRecord);
                }
                catch (e) {
                    scrapingErrors.push(e);
                }
            });

            await logoutOfEscapia(page);
        }
        catch (e) {
            scrapingErrors.push(e);
        }
    };

    for (let i = 0; i < unitGroupIdArrayArray.length; i++) {
        cluster.queue({ unitGroupIds: unitGroupIdArrayArray[i], unitGroupRecords: unitGroupRecords, scrapingErrors: scrapingErrors }, scrapeUnitAndUnitGroupRecords);
    }

    await cluster.idle();
    await cluster.close();

    if (scrapingErrors.length > 0) {
        throw scrapingErrors[0];
    }

    return unitGroupRecords;
};

module.exports = async function(escapiaCollectionName) {
    let browser;
    let page;
    try {
        // load browser
        browser = await getBrowser();
        if (!browser) {
            return;
        }

        // authenticate with escapia
        setAndGetConfigEscapia('1655'); // only running against the main instance
        page = await authenticateWithEscapia(await browser.newPage());
        if (!page) {
            await browser.close();
            return;
        }

        let unitGroupIds = await scrapeUnitGroupIds(page);

        // temp testing
        // let tempIds = [];
        // for (let i = 0; i < 5; i++) {
        //     tempIds.push(unitGroupIds[i]);
        // }
        // unitGroupIds = tempIds;


        let unitAndUnitGroupRecords = await getUnitAndUnitGroupRecords(unitGroupIds);

        // add sObjectType
        unitAndUnitGroupRecords = unitAndUnitGroupRecords.map(unitGroup => {
            unitGroup.sObjectType = 'Coupon_Code_Unit_Group__c';
            return unitGroup;
        });

        logger.verbose(`unitAndUnitGroupRecords.length: ${unitAndUnitGroupRecords.length}`);
        logger.verbose(`unitAndUnitGroupRecords record 1: ${JSON.stringify(unitAndUnitGroupRecords[1])}`);


        // push these records to mongodb
        await MongoDbCouponCodes.insertMany(escapiaCollectionName, unitAndUnitGroupRecords);


        await logoutOfEscapia(page);
        await browser.close();
        logger.verbose('done scraping unit groups');
        await MongoDbCouponCodes.close();
        return resultsEnum.SUCCESS;
    }
    catch (e) {
        try {
            await logoutOfEscapia(page);
            await browser.close();
        }
        finally {
            logger.error(e);
            await MongoDbCouponCodes.close();
            return resultsEnum.FAILURE;
        }
    }
};