const sfgen = require('@turnkeyvr/salesforcegeneral');
const logger = require('@turnkeyvr/logger');
const MongoDbCouponCodes = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('coupon-codes-prod');
const eachSeries = require('async/eachSeries');
const util = require('util');

// global utilities
const { sleep, resultsEnum, convertArrayToAnArrayOfArrays } = require('../../util/misc');

// local utilities
const { bulkDeleteSF, bulkInsertSF, upsertRecordUsingExternalId } = require('../util/sfdc');

const updateSF = util.promisify(sfgen.update);
const insertSF = util.promisify(sfgen.insert);
const customRequestSF = util.promisify(sfgen.customRequest);

const bulkInsertSFFromMongo = async (sobjectType, mongoRecords) => {
    let records = mongoRecords.map(mongoRecord => {
        let record = mongoRecord.record;
        record.attributes = { type: sobjectType, referenceId: mongoRecord._id};
        return record;
    });

    logger.verbose(`first mongo _id: ${records[0].attributes.referenceId}`);

    let recordArrays = convertArrayToAnArrayOfArrays(records, 200); // array of arrays
    let allResults = [];
    await eachSeries(recordArrays, async recordArray => {
        let results = await customRequestSF({
            method: 'POST', 
            url: '/services/data/v42.0/composite/sobjects',
            body: {
                allOrNone: true,
                records: recordArray
            }
        });

        // stupid hack for now -- turns out sfdc composite doesnt return the referenceId
        // so.... setting to allOrNone, checking first result, then manually processing results
        if (results[0].success) {
            results = recordArray.map(r => {
                return {
                    success: true,
                    referenceId: r.attributes.referenceId
                };
            })
        }
        else {
            results = recordArray.map(r => {
                return {
                    success: false,
                    referenceId: r.attributes.referenceId
                };
            })
        }

        allResults.push.apply(allResults, results);
    });

    logger.verbose(`all results 0: ${JSON.stringify(allResults[0])}`);

    return allResults;
};


const syncDelete = async (syncOperationsCollectionName, mongoRecords) => {
    let sfdcIdToMongoRecordMap = new Map();
    mongoRecords.forEach(mongoRecord => {
        sfdcIdToMongoRecordMap.set(mongoRecord.record.Id, mongoRecord);
    });

    let sfdcDeleteResults = await bulkDeleteSF(mongoRecords.map(r => r.record.Id));

    let deleteSuccessMongoIdArray = [];
    let deleteFailureMongoIdArray = [];
    sfdcDeleteResults.forEach(result => {
        if (result.success) {
            deleteSuccessMongoIdArray.push(sfdcIdToMongoRecordMap.get(result.id)._id); // push mongo id
            return;
        }
        logger.error(`failed to delete: ${JSON.stringify(result)}`);
        deleteFailureMongoIdArray.push(sfdcIdToMongoRecordMap.get(result.id)._id); // push mongo id
    });

    if (deleteSuccessMongoIdArray.length > 0) {
        await MongoDbCouponCodes.updateManyById(syncOperationsCollectionName, deleteSuccessMongoIdArray, { status: 'COMPLETE'}, { upsert: true });
    }
    if (deleteFailureMongoIdArray.length > 0) {
        await MongoDbCouponCodes.updateManyById(syncOperationsCollectionName, deleteFailureMongoIdArray, { status: 'FAILED'}, { upsert: true });
        return resultsEnum.FAILURE;
    }
    return resultsEnum.SUCCESS;
};


const syncInsert = async (syncOperationsCollectionName, sObjectType, mongoRecords) => {
    let sfdcInsertResults = await bulkInsertSFFromMongo(sObjectType, mongoRecords);

    let insertSuccessMongoIdArray = [];
    let insertFailureMongoIdArray = [];
    sfdcInsertResults.forEach(result => {
        if (result.success) {
            insertSuccessMongoIdArray.push(result.referenceId); // push mongo id
            return;
        }
        insertFailureMongoIdArray.push(result.referenceId); // push mongo id
    });

    if (insertSuccessMongoIdArray.length > 0) {
        await MongoDbCouponCodes.updateManyById(syncOperationsCollectionName, insertSuccessMongoIdArray, { status: 'COMPLETE'}, { upsert: true });
    }
    if (insertFailureMongoIdArray.length > 0) {
        await MongoDbCouponCodes.updateManyById(syncOperationsCollectionName, insertFailureMongoIdArray, { status: 'FAILED'}, { upsert: true });
        return resultsEnum.FAILURE;
    }
    return resultsEnum.SUCCESS;
};

// todo: better error handling
const syncUpsertWithExternalId = async (syncOperationsCollectionName, sObjectType, externalIdField, mongoRecords) => {
    let upsertSuccessMongoIdArray = [];
    let upsertFailureMongoIdArray = [];
    for (let i = 0; i < mongoRecords.length; i++) {
        let mongoId = mongoRecords[i]._id;
        let record = mongoRecords[i].record;
        let externalIdValue = record[externalIdField];
        delete record[externalIdField];

        let result;
        try {
            result = await upsertRecordUsingExternalId(sObjectType, externalIdField, externalIdValue, record);
        }
        catch (e) {
            logger.error(`failed to upsert mongo record id: ${mongoId}`);
            upsertFailureMongoIdArray.push(mongoId);
            continue;
        }
        
        if (result.success) {
            upsertSuccessMongoIdArray.push(mongoId);
            continue;
        }
        upsertFailureMongoIdArray.push(mongoId);
    }

    if (upsertSuccessMongoIdArray.length > 0) {
        await MongoDbCouponCodes.updateManyById(syncOperationsCollectionName, upsertSuccessMongoIdArray, { status: 'COMPLETE'}, { upsert: true });
    }

    if (upsertFailureMongoIdArray.length > 0) {
        await MongoDbCouponCodes.updateManyById(syncOperationsCollectionName, upsertFailureMongoIdArray, { status: 'FAILED'}, { upsert: true });
        return resultsEnum.FAILURE;
    }

    return resultsEnum.SUCCESS;
};

const syncSeasons = async (syncOperationsCollectionName) => {
    let dmlOperations = await MongoDbCouponCodes.find(syncOperationsCollectionName, {sObjectType: { '$in': [ 'Coupon_Code_Season__c', 'Coupon_Code_Season_Date_Range__c' ] }});

    // divide by action type
    let deleteRecords = dmlOperations.filter(r => r.action === 'DELETE' && r.status === 'PENDING');
    let insertRecords = dmlOperations.filter(r => r.action === 'INSERT' && r.status === 'PENDING');
    let treeInsertRecords = dmlOperations.filter(r => r.action === 'TREE INSERT' && r.status === 'PENDING');
    let updateRecords = dmlOperations.filter(r => r.action === 'UPDATE' && r.status === 'PENDING');

    // delete Date Ranges prior to Seasons
    if (deleteRecords.length > 0) {
        let dateRanges = deleteRecords.filter(r => r.sObjectType === 'Coupon_Code_Season_Date_Range__c');

        // mongo record statuses are updated in the syncDelete method
        let dateRangeDeleteResult = await syncDelete(syncOperationsCollectionName, dateRanges);
        if (dateRangeDeleteResult === resultsEnum.FAILURE) {
            logger.error('Date Range delete failed');
        }

        let seasons = deleteRecords.filter(r => r.sObjectType === 'Coupon_Code_Season__c');

        // mongo record statuses are updated in the syncDelete method
        let seasonDeleteResult = await syncDelete(syncOperationsCollectionName, seasons);
        if (seasonDeleteResult === resultsEnum.FAILURE) {
            logger.error('Season delete failed');
        }
    }

    if (insertRecords.length > 0) {
        let dateRanges = insertRecords.filter(r => r.sObjectType === 'Coupon_Code_Season_Date_Range__c');

        // mongo record statuses are updated in the syncDelete method
        let dateRangeInsertResult = await syncInsert(syncOperationsCollectionName, 'Coupon_Code_Season_Date_Range__c', dateRanges);
        if (dateRangeInsertResult === resultsEnum.FAILURE) {
            logger.error('Date Range insert failed');
        }
    }

    if (treeInsertRecords.length > 0) {
        // doing dumb for now, initial sync will be slow but negligible after that
        for (let i = 0; i < treeInsertRecords.length; i++) {
            let parentRecord;
            try {
                parentRecord = treeInsertRecords[i];
                let sfdcParentRecord = await insertSF('Coupon_Code_Season__c', {
                    Name: parentRecord.record.Name,
                    Escapia_Id__c: parentRecord.record.Escapia_Id__c
                });
                let sfdcChildRecords = [];
                for (let i = 0; i < parentRecord.record.Coupon_Code_Season_Date_Ranges__r.length; i++) {
                    let childRecord = parentRecord.record.Coupon_Code_Season_Date_Ranges__r[i];
                    sfdcChildRecords.push({
                        Coupon_Code_Season__c: sfdcParentRecord.Id,
                        Start_Date__c: childRecord.Start_Date__c,
                        End_Date__c: childRecord.End_Date__c
                    });
                }
                if (sfdcChildRecords.length > 0) {
                    let childInsertResults = await bulkInsertSF('Coupon_Code_Season_Date_Range__c', sfdcChildRecords, true);
                    // set allOrNone to true, so if first result is a failure then all of them failed
                    if (childInsertResults[0].success === false) {
                        await MongoDbCouponCodes.updateOneById(syncOperationsCollectionName, parentRecord._id, { status: 'FAILED'});
                        continue;
                    }
                }

                await MongoDbCouponCodes.updateOneById(syncOperationsCollectionName, parentRecord._id, { status: 'COMPLETE'});
            }
            catch (e) {
                await MongoDbCouponCodes.updateOneById(syncOperationsCollectionName, parentRecord._id, { status: 'FAILED'});
            }
        }
    }

    // this would only be season records
    if (updateRecords.length > 0) {
        logger.verbose(`season updaterecord 0: ${JSON.stringify(updateRecords[0])}`);
        for (let i = 0; i < updateRecords.length; i++) {
            let updateRecord = updateRecords[i];
            try {
                await updateSF('Coupon_Code_Season__c', {
                    Id: updateRecord.record.Id,
                    Name: updateRecord.record.Name
                });
            }
            catch (e) {
                await MongoDbCouponCodes.updateOneById(syncOperationsCollectionName, updateRecord._id, { status: 'FAILED'});
                continue;
            }
            await MongoDbCouponCodes.updateOneById(syncOperationsCollectionName, updateRecord._id, { status: 'COMPLETE'});
        }
    }
};

const syncUnitGroups = async (syncOperationsCollectionName) => {
    let dmlOperations = await MongoDbCouponCodes.find(syncOperationsCollectionName, {sObjectType: { '$in': [ 'Coupon_Code_Unit_Group__c', 'Coupon_Code_Unit__c' ] }});

    // divide by action type
    let deleteRecords = dmlOperations.filter(r => r.action === 'DELETE' && r.status === 'PENDING');
    let insertRecords = dmlOperations.filter(r => r.action === 'INSERT' && r.status === 'PENDING');
    let treeInsertRecords = dmlOperations.filter(r => r.action === 'TREE INSERT' && r.status === 'PENDING');
    let updateRecords = dmlOperations.filter(r => r.action === 'UPDATE' && r.status === 'PENDING');

    // delete Units prior to Unit Groups
    if (deleteRecords.length > 0) {
        let units = deleteRecords.filter(r => r.sObjectType === 'Coupon_Code_Unit__c');

        // mongo records are updated as success/failed within the syncDelete method
        let unitDeleteResult = await syncDelete(syncOperationsCollectionName, units);
        if (unitDeleteResult === resultsEnum.FAILURE) {
            logger.error('Unit delete failed');
        }

        let unitGroups = deleteRecords.filter(r => r.sObjectType === 'Coupon_Code_Unit_Group__c');

        // mongo records are updated as success/failed within the syncDelete method
        let unitGroupDeleteResult = await syncDelete(syncOperationsCollectionName, unitGroups);
        if (unitGroupDeleteResult === resultsEnum.FAILURE) {
            logger.error('Unit Group delete failed');
        }
    }

    if (insertRecords.length > 0) {
        let units = insertRecords.filter(r => r.sObjectType === 'Coupon_Code_Unit__c');

        // mongo records are updated as success/failed within the syncInsert method
        let unitInsertResult = await syncInsert(syncOperationsCollectionName, 'Coupon_Code_Unit__c', units);
        if (unitInsertResult === resultsEnum.FAILURE) {
            logger.error('Unit insert failed');
        }
    }

    if (treeInsertRecords.length > 0) {
        // doing dumb for now, initial sync will be slow but negligible after that
        for (let i = 0; i < treeInsertRecords.length; i++) {
            let parentRecord;
            try {
                parentRecord = treeInsertRecords[i];
                let sfdcParentRecord = await insertSF('Coupon_Code_Unit_Group__c', {
                    Name: parentRecord.record.Name,
                    Escapia_Id__c: parentRecord.record.Escapia_Id__c,
                    Status__c: parentRecord.record.Status__c
                });
                let sfdcChildRecords = [];
                for (let i = 0; i < parentRecord.record.Coupon_Code_Units__r.length; i++) {
                    let childRecord = parentRecord.record.Coupon_Code_Units__r[i];

                    sfdcChildRecords.push({
                        Coupon_Code_Unit_Group__c: sfdcParentRecord.Id,
                        Rental_Name__c: childRecord.Rental_Name__c
                    });
                }
                if (sfdcChildRecords.length > 0) {
                    let childRecordResults = await bulkInsertSF('Coupon_Code_Unit__c', sfdcChildRecords, true);
                    // bulkInsertSF is set to allOrNone === true, so if the first record failed then they all did
                    if (childRecordResults[0].success === false) {
                        await MongoDbCouponCodes.updateOneById(syncOperationsCollectionName, parentRecord._id, { status: 'FAILED'});
                        continue;
                    }
                }
                await MongoDbCouponCodes.updateOneById(syncOperationsCollectionName, parentRecord._id, { status: 'COMPLETE'});
            }
            catch (e) {
                await MongoDbCouponCodes.updateOneById(syncOperationsCollectionName, parentRecord._id, { status: 'FAILED'});
            }
        }
    }

    // this would only be unit groups
    if (updateRecords.length > 0) {
        logger.verbose(`unit group updaterecord 0: ${JSON.stringify(updateRecords[0])}`);
        for (let i = 0; i < updateRecords.length; i++) {
            let updateRecord = updateRecords[i];
            try {
                await updateSF('Coupon_Code_Unit_Group__c', updateRecord.record);
            }
            catch (e) {
                await MongoDbCouponCodes.updateOneById(syncOperationsCollectionName, updateRecord._id, { status: 'FAILED'});
                continue;
            }

            await MongoDbCouponCodes.updateOneById(syncOperationsCollectionName, updateRecord._id, { status: 'COMPLETE'});
        }
    }
};


const syncCouponCodesAndPricingPlans = async (syncOperationsCollectionName) => {
    let dmlOperations = await MongoDbCouponCodes.find(syncOperationsCollectionName, {sObjectType: { '$in': [ 'Coupon_Code__c', 'Pricing_Plan__c' ] }});

    // divide by action type
    let deleteRecords = dmlOperations.filter(r => r.action === 'DELETE' && r.status === 'PENDING');
    let upsertWithExternalIdRecords = dmlOperations.filter(r => r.action === 'UPSERT W/EXTERNAL ID' && r.status === 'PENDING');

    // delete coupon codes first
    if (deleteRecords.length > 0) {
        let couponCodes = deleteRecords.filter(r => r.sObjectType === 'Coupon_Code__c');

        // mongo records are updated as success/failed within the syncDelete method
        let couponCodeDeleteResult = await syncDelete(syncOperationsCollectionName, couponCodes);
        if (couponCodeDeleteResult === resultsEnum.FAILURE) {
            logger.error('Coupon Code delete failed');
        }

        let pricingPlans = deleteRecords.filter(r => r.sObjectType === 'Pricing_Plan__c');

        // mongo records are updated as success/failed within the syncDelete method
        let pricingPlanDeleteResult = await syncDelete(syncOperationsCollectionName, pricingPlans);
        if (pricingPlanDeleteResult === resultsEnum.FAILURE) {
            logger.error('Pricing Plan delete failed');
        }
    }


    // upsert pricing plans first
    if (upsertRecordUsingExternalId.length > 0) {
        let pricingPlans = upsertWithExternalIdRecords.filter(r => r.sObjectType === 'Pricing_Plan__c');
        // mongo records are updated as success/failed within the syncUpsertWithExternalId method
        let pricingPlansUpsertResult = await syncUpsertWithExternalId(syncOperationsCollectionName, 'Pricing_Plan__c', 'Escapia_Id__c', pricingPlans);
        if (pricingPlansUpsertResult === resultsEnum.FAILURE) {
            logger.error('pricing plan upsert failed');
        }

        let couponCodes = upsertWithExternalIdRecords.filter(r => r.sObjectType === 'Coupon_Code__c');
        // mongo records are updated as success/failed within the syncUpsertWithExternalId method
        let couponCodesUpsertResults = await syncUpsertWithExternalId(syncOperationsCollectionName, 'Coupon_Code__c', 'Escapia_Id__c', couponCodes);
        if (couponCodesUpsertResults === resultsEnum.FAILURE) {
            logger.error('coupon code upsert failed');
        }
    }
};

// this one is special b/c I'm relying on custom SFDC REST endpoints to handle the bulk of the logic
const syncRateRules = async (syncOperationsCollectionName) => {
    let successMongoIdArray = [];
    let failureMongoIdArray = [];
    let dmlOperations = await MongoDbCouponCodes.find(syncOperationsCollectionName, {sObjectType: 'Rate_Rule__c'});

    // divide by action type
    let deleteRecords = dmlOperations.filter(r => r.action === 'RATE RULE DELETE' && r.status === 'PENDING');

    // temp testing
    let upsertRecords = dmlOperations.filter(r => r.action === 'RATE RULE UPSERT' && r.status === 'PENDING');

    // delete rate rules first
    if (deleteRecords.length > 0) {
        for (let i = 0; i < deleteRecords.length; i++) {
            let deleteRecord = deleteRecords[i];

            try {
                let deleteResults = await customRequestSF({
                    method: 'POST', 
                    url: '/services/apexrest/DeleteRateRule',
                    body: {recordId: deleteRecord.record.Id}
                });

                if (!deleteResults[0].success) {
                    failureMongoIdArray.push(deleteRecord._id);
                    continue;
                }
            }
            catch (e) {
                failureMongoIdArray.push(deleteRecord._id);
            }

            successMongoIdArray.push(deleteRecord._id);
        }
    }

    // then do syncs
    if (upsertRecords.length > 0) {
        for (let i = 0; i < upsertRecords.length; i++) {
            let upsertRecord = upsertRecords[i];

            // format records
            let rateRuleActions = upsertRecord.record.Rate_Rule_Actions__r;
            delete upsertRecord.record.Rate_Rule_Actions__r;

            let rateRuleConditions = upsertRecord.record.Rate_Rule_Conditions__r;
            delete upsertRecord.record.Rate_Rule_Conditions__r;

            rateRuleConditions = rateRuleConditions.map(r => {
                if (r.Coupon_Code_Units__r && r.Coupon_Code_Units__r.length === 0) {
                    delete r.Coupon_Code_Units__r;
                }
                else if (r.Coupon_Code_Units__r) {
                    r.Coupon_Code_Units__r = {
                        totalSize: r.Coupon_Code_Units__r.length,
                        done: true,
                        records: r.Coupon_Code_Units__r
                    };
                }

                if (r.Rate_Rule_Condition_Unit_Groups__r && r.Rate_Rule_Condition_Unit_Groups__r.length === 0) {
                    delete r.Rate_Rule_Condition_Unit_Groups__r;
                }
                else if (r.Rate_Rule_Condition_Unit_Groups__r) {
                    r.Rate_Rule_Condition_Unit_Groups__r = {
                        totalSize: r.Rate_Rule_Condition_Unit_Groups__r.length,
                        done: true,
                        records: r.Rate_Rule_Condition_Unit_Groups__r
                    };
                }

                return r;
            });

            delete upsertRecord.record._id;

            let upsertResults
            try {
                upsertResults = await customRequestSF({
                    method: 'POST', 
                    url: '/services/apexrest/SyncRateRule',
                    body: {
                        rule: upsertRecord.record,
                        conditions: rateRuleConditions,
                        actions: rateRuleActions
                    }
                });

                if (!upsertResults[0].success) {
                    failureMongoIdArray.push(upsertRecord._id);
                    continue;
                }

                successMongoIdArray.push(upsertRecord._id);
            }
            catch (e) {
                logger.error(`rate rule sync failed: ${e.message}`);
                failureMongoIdArray.push(upsertRecord._id);
            }
        }
    }

    if (successMongoIdArray.length > 0) {
        await MongoDbCouponCodes.updateManyById(syncOperationsCollectionName, successMongoIdArray, { status: 'COMPLETE' }, { upsert: true });
    }

    if (failureMongoIdArray.length > 0) {
        await MongoDbCouponCodes.updateManyById(syncOperationsCollectionName, failureMongoIdArray, { status: 'FAILED' }, { upsert: true });
        return resultsEnum.FAILURE;
    }

    return resultsEnum.SUCCESS;
};

// the strategy for sync error handling is to sync as much as possible first, then check to see if all records were successfully synced
module.exports = async function(syncOperationsCollectionName) {
    try {
        await syncSeasons(syncOperationsCollectionName);
        await syncUnitGroups(syncOperationsCollectionName);
        await syncCouponCodesAndPricingPlans(syncOperationsCollectionName);
        await syncRateRules(syncOperationsCollectionName);

        // query for FAILED or PENDING sync operation records to determine if successful
        let failedDMLOperations = await MongoDbCouponCodes.find(syncOperationsCollectionName, {status: 'FAILED'});
        let pendingDMLOperations = await MongoDbCouponCodes.find(syncOperationsCollectionName, {status: 'PENDING'});

        if (failedDMLOperations.length > 0 || pendingDMLOperations.length > 0) {
            return resultsEnum.FAILURE;
        }
    }
    catch (e) {
        logger.error(e);
        await MongoDbCouponCodes.close();
        return resultsEnum.FAILURE;
    }

    await MongoDbCouponCodes.close();
    return resultsEnum.SUCCESS;
}