const sfgen = require('@turnkeyvr/salesforcegeneral');
const eachSeries = require('async/eachSeries');
const util = require('util');
const { sleep, resultsEnum, convertArrayToAnArrayOfArrays } = require('../../util/misc');

const querySF = util.promisify(sfgen.query);
const updateSF = util.promisify(sfgen.update);
const insertSF = util.promisify(sfgen.insert);
const deleteSF = util.promisify(sfgen.delete);
const customRequestSF = util.promisify(sfgen.customRequest);


// todo: add error handling
const upsertRecordUsingExternalId = async (sObjectType, externalIdField, externalIdValue, record) => {
    // delete record._id;
    // delete record.Escapia_Id__c;
    let results = await customRequestSF({
        method: 'PATCH', 
        url: `/services/data/v47.0/sobjects/${sObjectType}/${externalIdField}/${externalIdValue}`,
        body: record
    });
    return results;
};

const processRecordsForTreeVsBulkInsert = (insertRecords, childRelationshipName) => {
    let bulkInsertRecords = [];
    let treeArrayOfArrays = [];
    let treeArray = [];
    let recordCount = 0;
    for (let i = 0; i < insertRecords.length; i++) {
        let insertRecord = insertRecords[i];

        // first check to see if the parent + child records length is > 200; if so, normal bulk insert
        if (insertRecord[childRelationshipName] && insertRecord[childRelationshipName].records && insertRecord[childRelationshipName].records.length + 1 > 200) {
            bulkInsertRecords.push(insertRecord);
            continue;
        }

        // otherwise, logic for tree arrays
        recordCount += 1;
        if (insertRecord[childRelationshipName] && insertRecord[childRelationshipName].records) {
            recordCount += insertRecord[childRelationshipName].records.length;
        }
        if (recordCount > 200) {
            // reset values and reprocess this record
            treeArrayOfArrays.push(treeArray);
            treeArray = [];
            recordCount = 0;
            i -= 1;
            continue;
        }
        treeArray.push(insertRecord);
    }
    if (treeArray.length > 0) {
        treeArrayOfArrays.push(treeArray);
    }

    return { bulkInsertRecords: bulkInsertRecords, treeArrayOfArrays: treeArrayOfArrays };
};

// https://developer.salesforce.com/blogs/tech-pubs/2017/01/simplify-your-api-code-with-new-composite-resources.html
// https://salesforce.stackexchange.com/questions/118454/writing-nested-custom-object-records
// https://developer.salesforce.com/forums/?id=906F0000000kDSUIA2
// https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_composite_sobject_tree_create.htm
const bulkInsertCompositeTreeSF = async (sobjectType, recordArrays) => {
    // note: records have to be properly formatted before passing in
    //let recordArrays = convertArrayToAnArrayOfArrays(records, 200); // array of arrays
    let allResults = [];
    await eachSeries(recordArrays, async recordArray => {
        let results = await customRequestSF({
            method: 'POST', 
            url: `/services/data/v47.0/composite/tree/${sobjectType}/`,
            body: {
                records: recordArray
            }
        });
        allResults.push.apply(allResults, results);
    });
    return allResults;
};

const bulkInsertSF = async (sobjectType, records, allOrNone) => {
    records = records.map(record => {
        record.attributes = {type: sobjectType};
        return record;
    });
    let recordArrays = convertArrayToAnArrayOfArrays(records, 200); // array of arrays
    let allResults = [];
    await eachSeries(recordArrays, async recordArray => {
        let results = await customRequestSF({
            method: 'POST', 
            url: '/services/data/v42.0/composite/sobjects',
            body: {
                allOrNone: (allOrNone ? allOrNone : false),
                records: recordArray
            }
        });
        allResults.push.apply(allResults, results);
    });
    return allResults;
};

const bulkDeleteSF = async (recordIds) => {
    let idStrings = [];
    let idString = '';
    for (let i = 0; i < recordIds.length; i++) {
        if (i === 0) {
            idString = recordIds[i];
            continue;
        }
        if (i % 200 === 0) {
            idStrings.push(idString);
            idString = recordIds[i];
            continue;
        }
        idString += `,${recordIds[i]}`;
    }
    if (idString !== '') {
        idStrings.push(idString);
    }
    let allResults = [];
    await eachSeries(idStrings, async idString => {
        let results = await customRequestSF({method: 'DELETE', url: `/services/data/v42.0/composite/sobjects?ids=${idString}`});
        allResults.push.apply(allResults, results);
    });
    return allResults;
};

const deleteAllCouponRelatedRecordsInSFDC = async () => {
    let sfdcCouponCodeRecords = (await querySF('SELECT Id FROM Coupon_Code__c')).records;
    let sfdcRateRuleRecords = (await querySF('SELECT Id FROM Rate_Rule__c')).records;
    let sfdcPricingPlanRecords = (await querySF('SELECT Id FROM Pricing_Plan__c')).records;
    let sfdcSeasonDateRangeRecords = (await querySF('SELECT Id FROM Coupon_Code_Season_Date_Range__c')).records;
    let sfdcSeasonRecords = (await querySF('SELECT Id FROM Coupon_Code_Season__c')).records;
    let sfdcCouponCodeUnitRecords = (await querySF('SELECT Id FROM Coupon_Code_Unit__c')).records;
    let sfdcCouponCodeUnitGroupRecords = (await querySF('SELECT Id FROM Coupon_Code_Unit_Group__c')).records;

    if (sfdcCouponCodeRecords.length > 0) {
        await bulkDeleteSF(sfdcCouponCodeRecords.map(record => record.Id));
    }
    if (sfdcRateRuleRecords.length > 0) {
        await bulkDeleteSF(sfdcRateRuleRecords.map(record => record.Id));
    }
    if (sfdcPricingPlanRecords.length > 0) {
        await bulkDeleteSF(sfdcPricingPlanRecords.map(record => record.Id));
    }
    if (sfdcSeasonDateRangeRecords.length > 0) {
        await bulkDeleteSF(sfdcSeasonDateRangeRecords.map(record => record.Id));
    }
    if (sfdcSeasonRecords.length > 0) {
        await bulkDeleteSF(sfdcSeasonRecords.map(record => record.Id));
    }
    if (sfdcCouponCodeUnitRecords.length > 0) {
        await bulkDeleteSF(sfdcCouponCodeUnitRecords.map(record => record.Id));
    }
    if (sfdcCouponCodeUnitGroupRecords.length > 0) {
        await bulkDeleteSF(sfdcCouponCodeUnitGroupRecords.map(record => record.Id));
    }
};

module.exports = {
    querySF,
    updateSF,
    insertSF,
    deleteSF,
    customRequestSF,
    bulkInsertSF,
    bulkInsertCompositeTreeSF,
    bulkDeleteSF,
    deleteAllCouponRelatedRecordsInSFDC,
    processRecordsForTreeVsBulkInsert,
    upsertRecordUsingExternalId
};