const env = require('@turnkeyvr/env');
const logger = require('@turnkeyvr/logger');
const AWS = require('aws-sdk');
const uid = require('uid-safe');
const fs = require('fs-extra');
const util = require('util');
const sane = require('sane');
const { getBrowser } = require('../util/puppeteerUtil');
const { authenticateWithEscapia, setAndGetConfigEscapia } = require('../util/escapiaUtil');
const { resultsEnum } = require('../util/misc');
const { downloadFolioBalanceReport } = require('./reports/folioBalanceReport');
const { downloadRateActionGridReport } = require('./reports/rateActionGridReport');
const { downloadChargeTemplateUnitScopeReport } = require('./reports/chargeTemplateUnitScopeReport');
const { downloadOwnerLedgerReport } = require('./reports/ownerLedgerReport');

const DEFAULT_ESCAPIA_INSTANCE = '1655';
const SUPPORTED_ESCAPIA_INSTANCES = ['1655', '2685'];

AWS.config.region = 'us-east-2';
AWS.config.apiVersions = {
    sns: '2010-03-31'
};

const s3 = new AWS.S3({
    credentials: {
        accessKeyId: env.config.get('robotify.aws.accessKeyId'),
        secretAccessKey: env.config.get('robotify.aws.secretAccessKey'),
    }
});
s3.putObject = util.promisify(s3.putObject);

const sns = new AWS.SNS();

const tempPath = '/tmp'; // aws lambda only allows writing to this directory

const reportTypes = ['Folio Balance', 'Rate Action Grid', 'Charge Template Unit Scope'];

/**
 * request sns message example
{
    "transactionId": "qWSN1UHlEkESAJAbO_wV-ey4",
    "callingProcess": "process_name",
    "reportType": "Folio Balance",
    "fileFormat": "CSV",
    "reportParams": {
        "startDate": "6/6/2019",
        "endDate": "7/16/2019"
    },
    "escapiaInstanceId": "1655"
}
**/

/**
 * response sns message example
{
    "transactionId": "qWSN1UHlEkESAJAbO_wV-ey4",
    "callingProcess": "process_name",
    "reportType": "Folio Balance",
    "success": true,
    "s3URL": "https://robotify-escapia-reports.s3.us-east-2.amazonaws.com/qWSN1UHlEkESAJAbO_wV-ey4/Folio+Balance+Report.csv"
}
**/

const handler = async (event, context) => {
    logger.debug('Received event:', JSON.stringify(event, null, 2));

    const message = JSON.parse(event.Records[0].Sns.Message);

    // test message
    // let message = {
    //     transactionId: "AAA",
    //     callingProcess: "my kewl process",
    //     reportType: "Owner Ledger",
    //     fileFormat: "CSV",
    //     reportParams: {
    //         startDate: "7/6/2020",
    //         endDate: "8/28/2020"
    //     },
    //     escapiaInstanceId: "1655"
    // };

    
    // let message = {
    //     transactionId: "abcdefghijklmnop",
    //     callingProcess: "my kewl process",
    //     reportType: 'Charge Template Unit Scope',
    //     fileFormat: "CSV",
    //     reportParams: {}
    // };
    

    logger.debug('message: ', message);

    // todo: make sure report type is valid
    if (!message.reportType || !reportTypes.includes(message.reportType)) {
        // todo: return failure (or sns notify failure)
    }

    // only supported escapia instances
    if (message.escapiaInstanceId && !SUPPORTED_ESCAPIA_INSTANCES.includes(message.escapiaInstanceId)) {
        // todo: return failure (or sns notify failure)
    }

    let snsResponseParams = {
        TopicArn: 'arn:aws:sns:us-east-2:673853704728:escapia-robotify-report-download-results'
    };

    let snsResponseMessage = {
        transactionId: message.transactionId ? message.transactionId : await uid(18),
        callingProcess: message.callingProcess ? message.callingProcess : '',
        reportType: message.reportType,
        success: false,
        s3Bucket: null,
        s3Key: null
    };

    return new Promise(async (resolve) => {
        logger.debug('transactionId: ', snsResponseMessage.transactionId);
    
        let browser = await getBrowser();
        if (!browser) {
            return;
        }
    
        let escapiaInstanceId = message.escapiaInstanceId && SUPPORTED_ESCAPIA_INSTANCES.includes(message.escapiaInstanceId) ? 
                                message.escapiaInstanceId : 
                                DEFAULT_ESCAPIA_INSTANCE;
        setAndGetConfigEscapia(escapiaInstanceId);
        let page = await authenticateWithEscapia(await browser.newPage());
        if (!page) {
            await browser.close();
            return;
        }
    
        let downloadPath = `${tempPath}/${snsResponseMessage.transactionId}`;
    
        // create directory before monitoring
        fs.mkdirSync(downloadPath);

        let watcher = sane(downloadPath, {glob: ['**/*.csv', '**/*.xls']});
        watcher.on('add', async function (filepath, root, stat) { 
            logger.debug('SANE download complete: ', filepath);
            
            watcher.close();
            await browser.close();

            let data = fs.readFileSync(downloadPath + '/' + filepath);
            var dataBuffer = Buffer.from(data, 'binary');

            try {
                let s3Object = {
                    Bucket: 'robotify-escapia-reports',
                    Key: `${snsResponseMessage.transactionId}/${filepath}`,
                    Body: dataBuffer
                };
                let results = await s3.putObject(s3Object);
                logger.debug('s3 upload complete: ', results);

                snsResponseMessage.s3Bucket = s3Object.Bucket;
                snsResponseMessage.s3Key = s3Object.Key;
                snsResponseMessage.success = true;
            }
            catch (e) {
                logger.error('failed to upload to s3: ', e.message);
            }

            // publish sns message and exit
            snsResponseParams.Message = JSON.stringify(snsResponseMessage);
            sns.publish(snsResponseParams, (err, data) => {
                if (err) {
                    logger.debug(err);
                }
                logger.debug('done');
                resolve();
            });
        });


        let downloadResults;
        if (message.reportType.toLowerCase() === 'folio balance') {
            downloadResults = await downloadFolioBalanceReport(browser, downloadPath, message.reportParams);
        }
        else if (message.reportType.toLowerCase() === 'rate action grid') {
            downloadResults = await downloadRateActionGridReport(browser, downloadPath, message.reportParams);
        }
        else if (message.reportType.toLowerCase() === 'charge template unit scope') {
            downloadResults = await downloadChargeTemplateUnitScopeReport(browser, downloadPath);
        }
        else if (message.reportType.toLowerCase() === 'owner ledger') {
            downloadResults = await downloadOwnerLedgerReport(browser, downloadPath, message.reportParams);
        }
        
        if (downloadResults === resultsEnum.FAILURE) {
            logger.error('failed to download report');
            watcher.close();
            await browser.close();

            // publish sns message and exit
            snsResponseParams.Message = JSON.stringify(snsResponseMessage);
            sns.publish(snsResponseParams, (err, data) => {
                if (err) {
                    logger.debug(err);
                }
                logger.debug('done');
            });
        }
    });

};

const downloadResults = async (event, context) => {
    logger.debug('Received event:', JSON.stringify(event, null, 2));

    const message =  JSON.parse(event.Records[0].Sns.Message);
    logger.debug('message: ', message);
};

// report name
// report format
const downloadReport = (event, context, callback) => {
    let messagePayload = {
        meow: 'mix'
    };

    let response = {
        statusCode: 200,
        headers: {
          "Access-Control-Allow-Origin": "*",
          "Access-Control-Allow-Credentials": true
        }
    };

    let params = {
        Message: JSON.stringify(messagePayload),
        TopicArn: 'arn:aws:sns:us-east-2:673853704728:escapia-robotify-report-download-request'
    };
    sns.publish(params, (err, data) => {
        if (err) {
            logger.debug(err);
            response.body = JSON.stringify({ success: false, message: err.message });
        }
        else {
            response.body = JSON.stringify({ success: true });
        }
        callback(null, response);
    });
};



exports.handler = handler;
exports.downloadReport = downloadReport;
exports.downloadResults = downloadResults;