const { getBrowser, getInputValue } = require('../util/puppeteerUtil');
const { authenticateWithEscapia,
        setAndGetConfigEscapia,
        logoutOfEscapia } = require('../util/escapiaUtil');
const { getFUNBackingMarketingContentRecords } = require('../util/sfUtil');
const { sleep, resultsEnum } = require('../util/misc');
const { retry } = require('../util/retry');
const logger = require('@turnkeyvr/logger');
const sfGen = require('@turnkeyvr/salesforcegeneral');
const util = require('util');

const updateSF = util.promisify(sfGen.update);
const isProd = (process.env.NODE_ENV.toLowerCase() === 'prod' || process.env.NODE_ENV.toLowerCase() === 'production');

const gotoUnitInEscapia = async (page, rentalEscapiaId) => {
    await retry(
        () => Promise.all([
            page.goto(`https://turnkeyvr.escapia.com/WebAgency/Property/EditProperty.aspx?id=${rentalEscapiaId}&ReturnUrl=/Webagency/Property/ManageProperty.aspx`, { waitUntil: 'networkidle0' })
        ]), 5
    );
};

const handler = async () => {
    // only runs in production
    if (!isProd) {
        logger.verbose('FUN only runs in production');
        return;
    }

    let browser;
    let page;
    try {
        let records = await getFUNBackingMarketingContentRecords();

        // no results to process; early exit
        if (!records || records.length === 0) {
            logger.verbose('No FUN records to process');
            return;
        }

        // also need to check to see if the Rental subquery returned any results
        let funRecord = records[0];       

        if (!funRecord.Opportunities__r || funRecord.Opportunities__r.totalSize === 0) {
            throw 'No rental records to process';
        }

        logger.verbose(`got records: ${JSON.stringify(records[0])}`);

        let rentalRecord = funRecord.Opportunities__r.records[0];
        let rentalEscapiaId = rentalRecord.Escapia_Internal_Id__c.split('-')[1];

        if (!rentalEscapiaId || rentalEscapiaId === '') {
            throw 'Failed to get rentalEscapiaId';
        }

        // load browser
        browser = await getBrowser();
        if (!browser) {
            throw 'Failed to load browser';
        }

        // authenticate with escapia
        setAndGetConfigEscapia('1655'); // only running against the main instance
        page = await authenticateWithEscapia(await browser.newPage());
        if (!page) {
            throw 'Failed to authenticate with Escapia';
        }

        // go to record in escapia
        await gotoUnitInEscapia(page, rentalEscapiaId);

        // update and press save
        await retry(
            () => Promise.all([
                page.waitForSelector('input#PropertyWizard_ctl01_txtPropertyName', { visible: true })
            ]), 5
        );
        await page.$eval('input#PropertyWizard_ctl01_txtPropertyName', (node, fun) => {
            node.value = fun;
        }, funRecord.Friendly_Unit_Name__c);

        await page.click('#PropertyWizard_SaveBtn', {waitUntil: 'load'});

        // revisit record and make sure it was successfully updated
        await gotoUnitInEscapia(page, rentalEscapiaId);
        let funValue = await getInputValue(page, 'input#PropertyWizard_ctl01_txtPropertyName');
        if (funValue !== funRecord.Friendly_Unit_Name__c) {
            throw `Failed to update friendly unit name: ${funRecord.Id}`;
        }

        // update sfdc record
        await updateSF('Backing_Marketing_Content__c', {
            Id: funRecord.Id,
            Sync_Friendly_Unit_Name__c: false
        });

        // logout of escapia then exit browser
        await logoutOfEscapia(page);

        await browser.close();
        logger.verbose('done');
    }
    catch (e) {
        try {
            await browser.close();
        }
        finally {
            logger.alert(e);
        }
    }
};

exports.handler = handler;