const { getBrowser } = require('../util/puppeteerUtil');
const { authenticateWithEscapia,
        setAndGetConfigEscapia,
        applyHostFeeCreditToFolio,
        handleAirbnbEscapiaPercentComparison,
        removeGuestPhoneNumberFromReservationNote,
        removePendingReconciliationReservationNote,
        getReservationNumberFromReservation,
        getIsCancelledFromReservation,
        checkForPendingReconciliationReservationNote,
        logoutOfEscapia,
        handleTaxEdits } = require('../util/escapiaUtil');
const { getRateReconciliationCases, 
        getEscapiaUnitReservationIdFromCaseDescription, 
        getEscapiaQuotePercentComparisonFromCaseDescription, 
        getAirbnbAmountFromCaseDescription,
        getListingHostFeeFromCaseDescription,
        getGuestPhoneNumberFromCaseDescription,
        getWebsiteAmountFromCaseDescription,
        determineRequiredTaxEdits,
        updateCase,
        chatterNew } = require('../util/sfUtil');
const { sleep, resultsEnum } = require('../util/misc');
const eachSeries = require('async/eachSeries');
const logger = require('@turnkeyvr/logger');
const MongoDb = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('rate-reconciliation-prod');
const moment = require('moment');

const isProd = (process.env.NODE_ENV.toLowerCase() === 'prod' || process.env.NODE_ENV.toLowerCase() === 'production');


const handler = async () => {
    // only runs in production
    if (!isProd) {
        logger.debug('rate reconciliation only runs in production');
        return;
    }

    let cases = await getRateReconciliationCases('google');

    // no results to process; early exit
    if (!cases || cases.length === 0) {
        logger.info('No cases to process');
        return;
    }

    // load browser
    let browser = await getBrowser();
    if (!browser) {
        return;
    }

    // authenticate with escapia
    setAndGetConfigEscapia('1655'); // only running against the main instance
    let page = await authenticateWithEscapia(await browser.newPage());
    if (!page) {
        await browser.close();
        return;
    }

    // mark all cases as in progress
    let inProgressCases = [];
    await eachSeries(cases, async (caseRecord) => {
        // mark the Case as "In Progress"
        let inProgressResults = await updateCase({
            Id: caseRecord.Id,
            Status: 'In Progress'
        });
        if (inProgressResults === resultsEnum.FAILURE) {
            await handleCaseProcessingFailure(caseRecord, 'Failed to mark Case ' + caseRecord.Id + ' as In Progress');
            return;
        }
        inProgressCases.push(caseRecord);
    });

    // process each case
    await eachSeries(inProgressCases, async (caseRecord) => await processCase(browser, page, caseRecord));

    await MongoDb.close();

    // logout of escapia then exit browser
    await logoutOfEscapia(page);

    await browser.close();
    logger.debug('done');
};

const processCase = async (browser, page, caseRecord) => {
    logger.info('caseRecord Id: ', caseRecord.Id);

    let processCaseSuccess = resultsEnum.SUCCESS; // default to success
    let processCaseNotes = '';

    // if it's a "Marketing Only" Case, let Smiles process it -- higher likelihood of having a security deposit
    if (caseRecord.OpportunityCase__r.Is_Marketing_Only__c) {
        await handleCaseProcessingFailure(caseRecord, 'Robotify does not process Cases for marketing-only Rentals');
        return;
    }

    let dataResults = await getRequiredDataToProcessCase(page, caseRecord);
    if (dataResults.success === resultsEnum.FAILURE) {
        await handleCaseProcessingFailure(caseRecord, 'Failed to retrieve data for Case ' + caseRecord.Id + ': ' + dataResults.notes);
        return;
    }
    logger.debug(dataResults.data);


    // if the reservation in Escapia is cancelled, flag it as an error
    if (dataResults.data.isCancelled) {
        await handleCaseProcessingFailure(caseRecord, 'Reservation is cancelled in Escapia');
        return;
    }

    let { escapiaUnitReservationId, websiteAmount, escapiaQuotePercent, reservationNumber } = dataResults.data;

    // if the reservation in Escapia doesn't have a pending reconiliation note, flag it for human review
    let pendingReconciliationNotePresentResults = await checkForPendingReconciliationReservationNote(page, escapiaUnitReservationId)
    if (pendingReconciliationNotePresentResults === resultsEnum.FAILURE) {
        await handleCaseProcessingFailure(caseRecord, 'No "Pending Reconciliation" note Escapia, please review');
        return;
    }

    // handle escapia % comparison from case (also removes security deposit if present)
    // UPDATE FEB 5, 2020: For website rate reconciliation Cases we are always going to follow the >=85% logic where we adjust the rent amount/booking total
    //                     to match the website quote. As path of least resistance, I'm just passing in 100 to the function.
    if (escapiaQuotePercent < 85) {
        logger.error('Website rate reconciliation comparison percentage < 85% for Case Id: ', caseRecord.Id);
    }
    let handleAirbnbEscapiaPercentComparisonResults = await handleAirbnbEscapiaPercentComparison(browser, page, escapiaUnitReservationId, 100, websiteAmount);
    if (handleAirbnbEscapiaPercentComparisonResults === resultsEnum.FAILURE) {
        await handleCaseProcessingFailure(caseRecord, 'Unable to handle Google + Escapia percent comparison logic for Case ' + caseRecord.Id);
        return;
    }
    processCaseNotes += '- Google/Escapia percent comparison logic complete\n';

    // remove the pending reconciliation note from escapia reservation notes
    let removePendingReconciliationReservationNoteResults = await removePendingReconciliationReservationNote(page, escapiaUnitReservationId);
    if (removePendingReconciliationReservationNoteResults === resultsEnum.FAILURE) {
        await handleCaseProcessingFailure(caseRecord, 'Unable to remove "Pending Reconciliation" note from Escapia Reservation notes for Case ' + caseRecord.Id);
        return;
    }
    processCaseNotes += '- "Pending Reconciliation" note removed from reservation in Escapia\n';

    // make sure pending reconciliation note was remove from escapia reservation
    // if the reservation in Escapia doesn't have a pending reconiliation note, flag it for human review
    let pendingReconciliationNotePresentFinalResults = await checkForPendingReconciliationReservationNote(page, escapiaUnitReservationId)
    if (pendingReconciliationNotePresentFinalResults !== resultsEnum.FAILURE) {
        await handleCaseProcessingFailure(caseRecord, 'Unable to remove "Pending Reconciliation" note from Escapia Reservation notes for Case ' + caseRecord.Id);
        return;
    }

    // add case comment
    let chatterData = { 
        body: {
            messageSegments: [{
                type: 'Text',
                text: 'INT Completed ' + reservationNumber
            }]
        },
        feedElementType: 'FeedItem',
        subjectId: caseRecord.Id
    };
    let chatterResults = await chatterNew(chatterData);
    if (chatterResults === resultsEnum.FAILURE) {
        await handleCaseProcessingFailure(caseRecord, 'Failed to leave chatter message on Case ' + caseRecord.Id);
        return;
    }
    processCaseNotes += '- Chatter comment added to Case\n';

    // update the case for success/review required
    let caseUpdateRecord = {
        Id: caseRecord.Id
    };
    if (processCaseSuccess === resultsEnum.REVIEW_REQUIRED) {
        caseUpdateRecord.Subject = caseRecord.Subject += ' - - REVIEW REQUIRED';
        caseUpdateRecord.Description = caseRecord.Description += '\n\n##### ROBOTIFY NOTES #####\n' + processCaseNotes;
        caseUpdateRecord.Status = 'New';
    }
    else {
        caseUpdateRecord.Description = caseRecord.Description += '\n\n##### ROBOTIFY NOTES #####\n' + processCaseNotes;
        caseUpdateRecord.Status = 'Closed';

        // add mongo record for later recheck logic
        let newRecord = {
            reconciliationDate: moment().toDate(),
            escapiaUnitReservationId: escapiaUnitReservationId,
            sfdcCaseId: caseRecord.Id,
            attempts: 0,
            source: 'Website'
        };
        await MongoDb.insertMany('recheck', [newRecord]);
    }
    
    await updateCase(caseUpdateRecord);
};

const getRequiredDataToProcessCase = async (page, caseRecord) => {
    let results = {
        success: resultsEnum.SUCCESS,
        notes: '',
        data: {
            escapiaUnitReservationId: null,
            websiteAmount: null,
            escapiaQuotePercent: null,
            reservationNumber: null,
            isCancelled: false
        }
    };

    try {
        results.data.escapiaUnitReservationId = getEscapiaUnitReservationIdFromCaseDescription(caseRecord.Description);
        if (!results.data.escapiaUnitReservationId) {
            results.success = resultsEnum.FAILURE;
            results.notes += 'Unable to extract escapiaUnitReservationId from Case Description';
            return results;
        }
    
        results.data.websiteAmount = getWebsiteAmountFromCaseDescription(caseRecord.Description);
        if (!results.data.websiteAmount) {
            results.success = resultsEnum.FAILURE;
            results.notes += 'Unable to retrieve website amount from Case Description';
            return results;
        }
    
        results.data.escapiaQuotePercent = getEscapiaQuotePercentComparisonFromCaseDescription(caseRecord.Description);
        if (!results.data.escapiaQuotePercent) {
            results.success = resultsEnum.FAILURE;
            results.notes += 'Unable to retrieve escapiaQuotePercent from Case Description';
            return results;
        }
    
        results.data.reservationNumber = await getReservationNumberFromReservation(page, results.data.escapiaUnitReservationId);
        if (!results.data.reservationNumber) {
            results.success = resultsEnum.FAILURE;
            results.notes += 'Unable to retrieve reservationNumber';
            return results;
        }
    
        results.data.isCancelled = await getIsCancelledFromReservation(page, results.data.escapiaUnitReservationId);
        return results;
    }
    catch (e) {
        logger.alert(`Website reconciliation failed: ${e}`);
        results.success = resultsEnum.FAILURE;
        return results;
    }
};

const handleCaseProcessingFailure = async (caseRecord, errorMessage) => {
    logger.error(errorMessage);

    let caseUpdateRecord = {
        Id: caseRecord.Id,
        Status: 'New',
        Subject: caseRecord.Subject + ' - - ROBOTIFY FAILURE',
        Description: caseRecord.Description + '\n\n##### ROBOTIFY NOTES #####\n' + errorMessage
    };

    await updateCase(caseUpdateRecord);
};

exports.handler = handler;