const { getBrowser } = require('../util/puppeteerUtil');
const { authenticateWithEscapia,
        setAndGetConfigEscapia,
        handleAirbnbEscapiaPercentComparison,
        getReservationNumberFromReservation,
        getIsCancelledFromReservation,
        logoutOfEscapia, 
        handleTaxEdits,
        applyCommissionChargeToFolio,
        checkForProcessedNoteInReservation,
        addProcessedNoteToReservation } = require('../util/escapiaUtil');
const { updateCase,
        chatterNew } = require('../util/sfUtil');
const env = require('@turnkeyvr/env');
const sfGen = require('@turnkeyvr/salesforcegeneral').promise;
const logger = require('@turnkeyvr/logger');
const { sleep, resultsEnum } = require('../util/misc');
const eachSeries = require('async/eachSeries');

const SUPPORTED_BOOKING_CHANNELS = ['Airbnb', 'OTA: Expedia Merchant', 'OTA: Booking.com', 'OTA: Expedia ETP'];


// processed cases for testing:
// https://tkvr.my.salesforce.com/5003w00001PX8rX
// https://tkvr.my.salesforce.com/5003w00001PX8rmAAD
// https://tkvr.my.salesforce.com/5003w00001PX8vyAAD
// https://tkvr.my.salesforce.com/5003w00001PX8rw
// https://tkvr.my.salesforce.com/5003w00001PX8tdAAD
// https://tkvr.my.salesforce.com/5003w00001PX8v1AAD // tax hit folio credit
// https://tkvr.my.salesforce.com/5003w00001PX8qKAAT // no info


const reconcileCases = async () => {
    // only run in prod
    if (!env.is.prod) {
        logger.debug('rate reconciliation only runs in production');
        return;
    }

    while (true) {
        logger.info('Lexicon Rate Reconciliation is running'); // used for monitoring

        // query lexicon pending rec cases in sfdc that don't have a linked reservation
        const reconcileCaseQuery = `select id, subject, description, ContactId, Reservation__c, Reservation__r.EscapiaNETUnitReservationID__c from case where subject LIKE 'Finalize Lexicon reservation in Escapia%' AND (not subject like '%ROBOTIFY FAILURE%') AND Status = 'New' AND Reservation__c != null AND Record_Type_Developer_Name__c = 'Convert_Hold_To_Reservation' LIMIT 10`;
        let cases = await sfGen.query(reconcileCaseQuery);

        // no results to process; early exit
        if (!cases || cases.length === 0) {
            logger.info('No cases to process');
            await sleep(1000 * 60 * 10); // 10 minutes
            continue;
        }

        // load browser
        let browser = await getBrowser();
        if (!browser) {
            await sleep(1000 * 60 * 10); // 10 minutes
            continue;
        }

        // authenticate with escapia
        setAndGetConfigEscapia('2685'); // only running against the main instance
        let page = await authenticateWithEscapia(await browser.newPage());
        if (!page) {
            await browser.close();
            await sleep(1000 * 60 * 10); // 10 minutes
            continue;
        }

        // mark all cases as in progress
        let inProgressCases = [];
        await eachSeries(cases, async (caseRecord) => {
            // mark the Case as "In Progress"
            let inProgressResults = await updateCase({
                Id: caseRecord.Id,
                Status: 'In Progress'
            });
            if (inProgressResults === resultsEnum.FAILURE) {
                await handleCaseProcessingFailure(caseRecord, 'Failed to mark Case ' + caseRecord.Id + ' as In Progress');
            }
            else {
                inProgressCases.push(caseRecord);
            }
        });

        // process each case
        await eachSeries(inProgressCases, async (caseRecord) => await processCase(browser, page, caseRecord));

        // logout of escapia then exit browser
        await logoutOfEscapia(page);

        await browser.close();

        // breaking for test
        // break;
    }
};

const processCase = async (browser, page, caseRecord) => {
    try {
        logger.info('Lexicon Rate Reconciliation is running'); // used for monitoring
        logger.info('caseRecord Id: ', caseRecord.Id);

        let processCaseSuccess = resultsEnum.SUCCESS; // default to success
        let processCaseNotes = '';

        let dataResults = await getRequiredDataToProcessCase(page, caseRecord);
        if (dataResults.success === resultsEnum.FAILURE) {
            await handleCaseProcessingFailure(caseRecord, 'Failed to retrieve data for Case ' + caseRecord.Id + ': ' + dataResults.notes);
            return;
        }
        logger.debug(dataResults.data);

        // if the reservation in Escapia is cancelled, flag it as an error
        if (dataResults.data.isCancelled) {
            await handleCaseProcessingFailure(caseRecord, 'Reservation is cancelled in Escapia');
            return;
        }

        let { escapiaUnitReservationId, lexiconAmount, commissionFolioCharge, bookingChannel, reservationNumber } = dataResults.data;

        // check to see if [RECONCILIATION COMPLETE] note in reservation notes in Expedia; if it is, flag the case for review
        let checkForReconciledNoteInEscapiaResults = await checkForProcessedNoteInReservation(page, escapiaUnitReservationId);
        if (checkForReconciledNoteInEscapiaResults === resultsEnum.SUCCESS) {
            await handleCaseProcessingFailure(caseRecord, 'Reservation is flagged as already reconciled in Escapia');
            return;
        }

        // if airbnb first override taxes
        if (bookingChannel === 'Airbnb') {
            let taxEdits = { action: 'Partial Taxes', taxesToBeAdjusted: [{ name: 'FL Dept of Revenue', subcategory: '3000.1000' }] };
            let handleTaxEditsResults = await handleTaxEdits(browser, page, escapiaUnitReservationId, taxEdits);
            logger.debug('handleTaxEditsResults', handleTaxEditsResults);
            if (handleTaxEditsResults === resultsEnum.FAILURE) {
                await handleCaseProcessingFailure(caseRecord, 'Tax edits for Case ' + caseRecord.Id + ' failed');
                return;
            }
            else if (handleTaxEditsResults === resultsEnum.REVIEW_REQUIRED) {
                processCaseSuccess = resultsEnum.REVIEW_REQUIRED;
                processCaseNotes += '- Tax edits for Case ' + caseRecord.Id + ' require review\n';
                logger.warn('Tax edits for Case ' + caseRecord.Id + ' require review');
            }
            else {
                processCaseNotes += '- Tax edits complete\n';
            }
        }

        // handle escapia % comparison from case (also removes security deposit if present)
        // For lexicon rate reconciliation Cases we are always going to follow the >=85% logic where we adjust the rent amount/booking total
        // to match the quote. As path of least resistance, I'm just passing in 100 to the function.
        let handleAirbnbEscapiaPercentComparisonResults = await handleAirbnbEscapiaPercentComparison(browser, page, escapiaUnitReservationId, 100, lexiconAmount);
        if (handleAirbnbEscapiaPercentComparisonResults === resultsEnum.FAILURE) {
            await handleCaseProcessingFailure(caseRecord, 'Unable to balance booking total for Case ' + caseRecord.Id);
            return;
        }
        processCaseNotes += '- Balance booking total comparison logic complete\n';

        // add folio charge if needed
        if (bookingChannel === 'OTA: Booking.com' || bookingChannel === 'OTA: Expedia ETP') {
            let commissionFolioChargeResults = await applyCommissionChargeToFolio(page, escapiaUnitReservationId, commissionFolioCharge)
            if (commissionFolioChargeResults !== resultsEnum.SUCCESS) {
                await handleCaseProcessingFailure(caseRecord, 'Unable to apply commission folio charge for Case ' + caseRecord.Id);
                return;
            }
            processCaseNotes += '- Folio charge added\n';
        }

        // add note to reservation notes section in escapia that this was reconciled (to check for later)
        let addedProcessedNoteResults = await addProcessedNoteToReservation(page, escapiaUnitReservationId);
        if (addedProcessedNoteResults !== resultsEnum.SUCCESS) {
            await handleCaseProcessingFailure(caseRecord, 'Unable to add processed note to Escapia Reservation for Case ' + caseRecord.Id);
            return;
        }
        processCaseNotes += '- Processed note added in Escapia Reservation notes\n';

        // add case comment
        let chatterData = { 
            body: {
                messageSegments: [{
                    type: 'Text',
                    text: 'Completed ' + reservationNumber
                }]
            },
            feedElementType: 'FeedItem',
            subjectId: caseRecord.Id
        };
        let chatterResults = await chatterNew(chatterData);
        if (chatterResults === resultsEnum.FAILURE) {
            await handleCaseProcessingFailure(caseRecord, 'Failed to leave chatter message on Case ' + caseRecord.Id);
            return;
        }
        processCaseNotes += '- Chatter comment added to Case\n';

        // update the case for success/review required
        let caseUpdateRecord = {
            Id: caseRecord.Id
        };
        if (processCaseSuccess === resultsEnum.REVIEW_REQUIRED) {
            caseUpdateRecord.Subject = caseRecord.Subject += ' - - REVIEW REQUIRED';
            caseUpdateRecord.Description = caseRecord.Description += '\n\n##### ROBOTIFY NOTES #####\n' + processCaseNotes;
            caseUpdateRecord.Status = 'New';
        }
        else {
            caseUpdateRecord.Description = caseRecord.Description += '\n\n##### ROBOTIFY NOTES #####\n' + processCaseNotes;
            caseUpdateRecord.Status = 'Closed';
        }
        await updateCase(caseUpdateRecord);
    }
    catch (e) {
        logger.error(`Process individual case exception: ${JSON.stringify(e)}`);

        // attempt to update the case as failed
        await handleCaseProcessingFailure(caseRecord, 'Generic error for Case Id: ' + caseRecord.Id);
    }
};

const getRequiredDataToProcessCase = async (page, caseRecord) => {
    let results = {
        success: resultsEnum.SUCCESS,
        notes: '',
        data: {
            escapiaUnitReservationId: null,
            lexiconAmount: null,
            commissionFolioCharge: null,
            bookingChannel: null,
            reservationNumber: null,
            isCancelled: false
        }
    };

    results.data.escapiaUnitReservationId = getLexiconEscapiaUnitIdFromReservation(caseRecord.Reservation__r);
    if (!results.data.escapiaUnitReservationId) {
        results.success = resultsEnum.FAILURE;
        results.notes += 'Unable to retrieve escapiaUnitReservationId from Reservation';
        return results;
    }

    results.data.lexiconAmount = getLexiconAmountFromCaseDescription(caseRecord.Description);
    if (!results.data.lexiconAmount) {
        results.success = resultsEnum.FAILURE;
        results.notes += 'Unable to retrieve lexiconAmount from Case Description';
        return results;
    }

    results.data.bookingChannel = getBookingChannelFromCaseDescription(caseRecord.Description);
    if (!results.data.bookingChannel) {
        results.success = resultsEnum.FAILURE;
        results.notes += 'Unable to retrieve bookingChannel from Case Description';
        return results;
    }
    if (!SUPPORTED_BOOKING_CHANNELS.includes(results.data.bookingChannel)) {
        results.success = resultsEnum.FAILURE;
        results.notes += 'Unknown booking channel: ' + results.data.bookingChannel;
        return results;
    }

    results.data.commissionFolioCharge = getCommissionFolioChargeFromCaseDescription(caseRecord.Description);
    if (!results.data.commissionFolioCharge) {
        results.success = resultsEnum.FAILURE;
        results.notes += 'Unable to retrieve commissionFolioCharge from Case Description';
        return results;
    }
    if (results.data.bookingChannel === 'OTA: Booking.com' || results.data.bookingChannel === 'OTA: Expedia ETP') {
        if (results.data.commissionFolioCharge === 'N/A') {
            results.success = resultsEnum.FAILURE;
            results.notes += 'Missing commissionFolioCharge from Case Description';
            return results;
        }
    }

    results.data.reservationNumber = await getReservationNumberFromReservation(page, results.data.escapiaUnitReservationId);
    if (!results.data.reservationNumber) {
        results.success = resultsEnum.FAILURE;
        results.notes += 'Unable to retrieve reservationNumber';
        return results;
    }

    results.data.isCancelled = await getIsCancelledFromReservation(page, results.data.escapiaUnitReservationId);

    return results;
};

const linkCasesToReservations = async () => {
    // only run in prod
    if (!env.is.prod) {
        logger.debug('rate reconciliation only runs in production');
        return;
    }

    try {
        // query lexicon pending rec cases in sfdc that don't have a linked reservation
        const unlinkedCaseQuery = `select id, subject, description, ContactId, Reservation__c from case where subject LIKE 'Finalize Lexicon reservation in Escapia%' AND (not subject like '%ROBOTIFY FAILURE%') AND IsClosed = false AND Reservation__c = null AND Record_Type_Developer_Name__c = 'Convert_Hold_To_Reservation' LIMIT 50`;
        let unlinkedCaseRecords = await sfGen.query(unlinkedCaseQuery);
        let lexiconRezIdToCaseMap = new Map();

        for (let i = 0; i < unlinkedCaseRecords.length; i++) {
            let caseRecord = unlinkedCaseRecords[i];

            // if "Lexicon Reservation ID" not available, mark case as failure for manual intervention
            let lexiconRezId = getLexiconReservationIdFromCaseDescription(caseRecord.Description);
            if (!lexiconRezId) {
                await handleCaseProcessingFailure(caseRecord, `Unable to parse Lexicon Reservation ID value in Case Description | Case Id: ${caseRecord.Id}`);
                continue;
            }

            logger.info(`lexiconRezId: ${lexiconRezId}`);

            lexiconRezIdToCaseMap.set(lexiconRezId, caseRecord);
        }

        if (lexiconRezIdToCaseMap.size < 1) {
            return;
        }

        // now need to query for Reservations
        // TODO: consider restricting the number of reservations that can be queried; there are soql limits to how many characters can be in a IN statement
        const rezQuery = `SELECT Id, Channel_PMS_Booking_Number__c, Rental__c, Rental__r.Name, GuestContact__c FROM Reservation__c WHERE Channel_PMS_Booking_Number__c IN (${Array.from(lexiconRezIdToCaseMap.keys()).map(id => `'${id}'`)})`;
        let rezs = await sfGen.query(rezQuery);

        if (rezs.length < 1) {
            return;
        }

        let caseUpdateList = [];
        for (let i = 0; i < rezs.length; i++) {
            let rez = rezs[i];
            let caseRecord = lexiconRezIdToCaseMap.get(rez.Channel_PMS_Booking_Number__c);
            caseRecord.Reservation__c = rez.Id;
            caseRecord.OpportunityCase__c = rez.Rental__c;

            // try to associate guest to case, but not the end of the world if we don't
            if (rez.GuestContact__c) {
                caseRecord.ContactId = rez.GuestContact__c;
            }

            // if PENDING flag missing mark as failure for manual review; shouldn't really happen
            if (caseRecord.Description.indexOf('Escapia Id: PENDING') < 0) {
                // extra space in "Escapia Id:  PENDING" in error message below to prevent some weirdo reprocessesing scenario
                await handleCaseProcessingFailure(caseRecord, `'Escapia Id:  PENDING' note missing from Case Description when expected | Case Id: ${caseRecord.Id}`);
                continue;
            }

            // need to replace Escapia Id value with rental name in case description
            caseRecord.Description = caseRecord.Description.replace('Escapia Id: PENDING', `Escapia Id: ${rez.Rental__r.Name}`);
            caseUpdateList.push(caseRecord);
        }

        await sfGen.update('Case', caseUpdateList);
    }
    catch (e) {
        logger.error(`linkCasesToReservations error: ${e.message}`);
        
        // todo: consider marking Cases as failures, on the fence...
    }

    return;
};

const getLexiconEscapiaUnitIdFromReservation = (reservation) => {
    if (!reservation || !reservation.EscapiaNETUnitReservationID__c) {
        return null;
    }
    const matches = reservation.EscapiaNETUnitReservationID__c.match('2685-(\\d+)')
    return matches ? matches[1] : null;
};

const getLexiconReservationIdFromCaseDescription = (description) => {
    if (description.indexOf('Lexicon Reservation ID:') === -1) {
        return null;
    }
    const matches = description.match(/Lexicon Reservation ID:\s*(.*)/i);
    return matches ? matches[1] : null;
};

const getLexiconAmountFromCaseDescription = (description) => {
    if (description.indexOf('Lexicon Amount:') === -1) {
        return null;
    }
    const matches = description.match(/Lexicon Amount:\s*(.*)/i);
    return matches ? Number(parseFloat(matches[1].replace(',', '')).toFixed(2)) : null;
};

const getCommissionFolioChargeFromCaseDescription = (description) => {
    if (description.indexOf('Commission Folio Charge:') === -1) {
        return null;
    }
    if (description.indexOf('Commission Folio Charge: N/A') !== -1) {
        return 'N/A';
    }
    const matches = description.match(/Commission Folio Charge:\s*(.*)/i);
    return matches ? Number(parseFloat(matches[1].replace(',', '')).toFixed(2)) : null;
};

const getBookingChannelFromCaseDescription = (description) => {
    if (description.indexOf('Lexicon Booking Channel:') === -1) {
        return null;
    }
    const matches = description.match(/Lexicon Booking Channel:\s*(.*)/i);
    return matches ? matches[1] : null;
};

const handleCaseProcessingFailure = async (caseRecord, errorMessage) => {
    logger.error(errorMessage);

    let caseUpdateRecord = {
        Id: caseRecord.Id,
        Status: 'New',
        Subject: caseRecord.Subject + ' - - ROBOTIFY FAILURE',
        Description: caseRecord.Description + '\n\n##### ROBOTIFY NOTES #####\n' + errorMessage
    };

    await updateCase(caseUpdateRecord);
};

module.exports = {
    linkCasesToReservations,
    reconcileCases
};