const { getBrowser } = require('../util/puppeteerUtil');
const { setAndGetConfigEscapia, authenticateWithEscapia, logoutOfEscapia } = require('../util/escapiaUtil');
const moment = require('moment');
const _ = require('lodash');
const awsSns = require('@turnkeyvr/aws_sns');
const { promisify } = require("util");
const promiseSendSns = promisify(awsSns.publishEmail);
const env = require('@turnkeyvr/env');
const { queryBq } = require('../util/bigQueryUtil');
const { setRateRule } = require('../util/rateRuleUtil');
const logger = require('@turnkeyvr/logger');
const sfgen = require('@turnkeyvr/salesforcegeneral').promise;

// mainHandler runs hourly
// workerHandler seems to take around 120 seconds
// workerHandler has concurrency of 3
const CALL_LIMIT = 200;

// Main runs every 60 minutes, so worker should discard events older than the last run.
// But allow for worker operation taking up to 120 seconds
// This avoids duplicate processing, and a buildup of sns messages.
const DISCARD_EVENTS_OLDER_THAN_MINUTES = 57;

// Mark the rate in mongo as error, if the object's start timestamp is older than this, and it did not complete or error yet
const OPERATION_MAX_DURATION_MINUTES = 120;

const mongoDbRateManagement = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('rate-management-prod')

const mainHandler = async () => {
    logger.info("start");

    await cleanupPendingItems();

    logger.info("get rates from bq...");
    const newRateSets = await getRates();

    logger.info("get rates from mongo...");
    const knownRates = await mongoDbRateManagement.find('currentRates', {status: {$ne: 'Expire'}});

    const knownRatesByRateId = {};
    knownRates.forEach(r => knownRatesByRateId[r._id] = r);

    const workItems = getWorkItems(newRateSets, knownRates);

    workItems.forEach(wi => {
        const kr = knownRatesByRateId[wi.rateId];
        wi.previousCallHadError = kr && kr.syncErrorTimestamp;
        wi.pending = kr && kr.syncStartedTimestamp && !kr.syncCompletedTimestamp && !kr.syncErrorTimestamp;
    });
    
    try {
        await processAll(workItems);        
    }
    catch (ex) {
        logger.error("error in process all", ex);
    }

    try {        
        await mongoDbRateManagement.close();
    }
    catch (ex) {
        logger.error("error closing mongo", ex);
    }

    return true
}
exports.mainHandler = mainHandler

const cleanupPendingItems = async () => {
    logger.debug("cleanupPendingItems: start");

    const pendingItems = await mongoDbRateManagement.find('currentRates', {
        syncStartedTimestamp: {$ne: null},
        syncCompletedTimestamp: {$eq: null},
        syncErrorTimestamp: {$eq: null},
    });

    if (pendingItems.length > 0) {
        logger.warn(`pending items found: ${pendingItems.length}`);
    }

    const didNotComplete = pendingItems.filter(item => moment().diff(item.syncStartedTimestamp, "minutes") > OPERATION_MAX_DURATION_MINUTES);

    for (let i = 0; i < didNotComplete.length; i++) {
        const item = didNotComplete[i];
        logger.warn(`update mongo with "operation did not complete error" for ${item.date} ${item.minstay}`);
        item.syncError = `operation did not complete after ${OPERATION_MAX_DURATION_MINUTES} minutes`;
        item.syncErrorAfterSave = false;
        item.syncErrorTimestamp = new Date();
        item.syncErrorCount = (item.syncErrorCount || 0) + 1;
        item.syncCompletedTimestamp = null;
        let updateResult = await mongoDbRateManagement.updateOneById('currentRates', item._id, _.pick(item, [
            "syncError",
            "syncErrorAfterSave",
            "syncErrorTimestamp",
            "syncErrorCount",
            "syncCompletedTimestamp"
        ]));
        logger.debug(JSON.stringify(updateResult));
    }

    logger.debug("cleanupPendingItems: done");
}

const processAll = async (workItems) => {
    const adds = workItems.filter(wi => wi.Status == "Add");
    logger.info("Total adds needed", adds.length);
    const updates = workItems.filter(wi => wi.Status == "Update");
    logger.info("Total updates needed", updates.length);
    const expires = workItems.filter(wi => wi.Status == "Expire");
    logger.info("Total expires needed", expires.length);
    const previousCallHadError = workItems.filter(wi => wi.previousCallHadError);
    logger.info("Total items that previously failed", previousCallHadError.length);
    const pending = workItems.filter(wi => wi.pending);
    logger.info("Total pending items", pending.length);
    logger.info('throttled - but possible ' + workItems.length + ' new workitems!')

    let calls = workItems.filter(wi => !wi.previousCallHadError && !wi.pending);
    logger.info("Total calls needed (non-retry)", calls.length);
    let retryCalls = workItems.filter(wi => wi.previousCallHadError && !wi.pending);
    logger.info("Total retry calls needed", retryCalls.length);
    calls = calls.slice(0, CALL_LIMIT);
    logger.info("Calls (non-retry)", calls.length);
    const retryCallLimit = CALL_LIMIT - calls.length;
    retryCalls = retryCalls.slice(0, retryCallLimit);
    logger.info("Retry calls", retryCalls.length);
    calls = calls.concat(retryCalls);
    logger.info("Total calls this run", calls.length);

    for (let i = 0; i < calls.length; i++) {
        const item = calls[i];
        try {
            item.syncQueuedTimestamp = new Date();
            // testing
            // await workerHandler({Records: [{Sns: {MessageId: '1234', Message: JSON.stringify(item)}}]}, null)
            // prod
            await promiseSendSns(env.config.get('snsTopics.setRateRule'), 'RATESET', JSON.stringify(item))
        }
        catch (e) {
            logger.error(`${item.date} ${item.minstay}: unhandled error`, e);
        }
    }
}

const workerHandler = async (event, context) => {
    let snsEvent = event.Records[0].Sns
    let rateChangeSet = JSON.parse(snsEvent.Message)
    logger.info('workerHandler doing ' + rateChangeSet.Status + ' for ' + rateChangeSet.Name)

    // avoid uncaught exceptions because they will trigger automatic lambda retry
    try {
        await processOne(rateChangeSet);
    }
    catch(ex) {
        logger.error(`unhandled error for ${rateChangeSet.Status} ${rateChangeSet.Name}:`, ex);
    }

    try {        
        await mongoDbRateManagement.close();
    }
    catch (ex) {
        logger.error("error closing mongo", ex);
    }
}
exports.workerHandler = workerHandler

// as of today - each call to processOne sets one rate set in escapia 
// [min stay / date] and a collection of rentals and their rates for that date
const processOne = async (rateChangeSet) => {
    if (moment().diff(rateChangeSet.syncQueuedTimestamp, "minutes") > DISCARD_EVENTS_OLDER_THAN_MINUTES) {
        logger.debug(`${rateChangeSet.Name}: syncQueuedTimestamp too old: ${rateChangeSet.syncQueuedTimestamp}`);
        throw "syncQueuedTimestamp too old";
    }

    let mongoRateId = null;
    let syncErrorCount = null;

    const existingItems = await mongoDbRateManagement.find('currentRates', {
        date: rateChangeSet.date,
        minstay: rateChangeSet.minstay
    });

    if (existingItems.length > 0) {
        logger.debug(`${rateChangeSet.Name}: existing items found: ${existingItems.length}`);
        const existing = existingItems[0];
        mongoRateId = existing._id;
        syncErrorCount = existing.syncErrorCount;

        if (existing.syncStartedTimestamp && !existing.syncErrorTimestamp && !existing.syncCompletedTimestamp) {
            throw "item is already pending";
        }
    }

    let mongoRate = {
        date: rateChangeSet.date,
        minstay: rateChangeSet.minstay,
        syncAction: rateChangeSet.Status,
        syncStartedTimestamp: new Date(),
        syncDataPending: { // this will get removed on successful sync
            rentals: rateChangeSet.rentals
        },
        syncWarnings: [],
        syncError: null,
        syncErrorAfterSave: null,
        syncErrorTimestamp: null,
        syncCompletedTimestamp: null
    };
    
    if (mongoRateId) {
        logger.info(`update mongo for ${rateChangeSet.Name}`);
        let updateResult = await mongoDbRateManagement.updateOneById('currentRates', mongoRateId, mongoRate);
        logger.debug(JSON.stringify(updateResult));
    }
    else {
        mongoRate.syncCreatedTimestamp = new Date();
        logger.info(`insert into mongo for ${rateChangeSet.Name}`);
        let insertNewResult = await mongoDbRateManagement.insertMany('currentRates', [mongoRate]);
        mongoRateId = mongoRate._id;
        logger.debug(JSON.stringify(insertNewResult));
    }

    let error = null;
    try {
        await doEscapiaWork(rateChangeSet);
    }
    catch(ex) {
        error = ex;
    }

    if (error) {
        logger.error(`error processing robot rate [${error}]: ${rateChangeSet.Status} ${rateChangeSet.Name}`)
    }
    else {
        logger.info(rateChangeSet.Name + ': done with rate - success');

        const notFound = rateChangeSet.rentals.filter(r => r.notFound).map(r => r.name);
        if (notFound.length > 0 && rateChangeSet.Status !== 'Expire') {
            mongoRate.syncWarnings.push({
                problem: "rentalsNotFound",
                message: "One or more rentals were not found in Escapia.",
                rentalNames: notFound
            });

            const invalidRentals = await getInvalidRentalsFromSf(notFound);
            const nonCanceledRentalsNotFound = notFound.filter(r => !invalidRentals.includes(r));
            if (nonCanceledRentalsNotFound.length > 0) {
                mongoRate.syncWarnings.push({
                    problem: "nonCanceledRentalsNotFound",
                    message: "One or more rentals were not found in Escapia. These rentals are not canceled in Salesforce.",
                    rentalNames: nonCanceledRentalsNotFound
                });
            }        
        }

        if (rateChangeSet.inactiveRentalsThatCannotBeRemoved.length > 0) {
            mongoRate.syncWarnings.push({
                problem: "inactiveRentalsThatCannotBeRemoved",
                message: "Saved with inactive rentals that should not be in the rule but cannot be removed.",
                rentalNames: rateChangeSet.inactiveRentalsThatCannotBeRemoved
            });
        }
    }
    
    if (mongoRate.syncWarnings.length > 0) {
        logger.warn(`${rateChangeSet.Name}: syncWarnings: ${JSON.stringify(mongoRate.syncWarnings)}`);
    }

    // update mongo
    if (error) {
        mongoRate.syncError = error;
        mongoRate.syncErrorAfterSave = rateChangeSet.saved;
        mongoRate.syncErrorTimestamp = new Date();
        mongoRate.syncErrorCount = (syncErrorCount || 0) + 1;
        mongoRate.syncCompletedTimestamp = null;
    }
    else {
        mongoRate.syncError = null;
        mongoRate.syncErrorAfterSave = null;
        mongoRate.syncErrorTimestamp = null;
        mongoRate.syncErrorCount = 0;
        mongoRate.syncCompletedTimestamp = new Date();
        mongoRate.rateRuleName = rateChangeSet.rateRuleName; // current escapia state
        mongoRate.status = rateChangeSet.Status; // current escapia state
        mongoRate.rentals = rateChangeSet.rentals; // current escapia state
        mongoRate.syncDataPending = {};
    }

    let updateResult = await mongoDbRateManagement.updateOneById('currentRates', mongoRateId, mongoRate);
    logger.debug(JSON.stringify(updateResult));
    logger.info(`updated mongo after processing, for ${rateChangeSet.Name}`);

    const logItem = Object.assign({}, mongoRate);
    logItem.rateId = mongoRateId;
    delete logItem._id;
    let insertNewResult = await mongoDbRateManagement.insertMany('rateLog', [logItem]);
    logger.debug(JSON.stringify(insertNewResult));
    logger.info(`inserted into mongo log for ${rateChangeSet.Name}`);
}

const doEscapiaWork = async (rateChangeSet) => {
    logger.info('doEscapiaWork ' + rateChangeSet.Status + ' for ' + rateChangeSet.Name)

    logger.debug("get browser");
    let browser = await getBrowser();
    if (!browser) {
        throw "cannot get browser";
    }

    logger.debug("authenticate with escapia");
    const instanceId = "1655";
    setAndGetConfigEscapia(instanceId); // only running against the main instance
    let page = await authenticateWithEscapia(await browser.newPage());
    if (!page) {
        await browser.close();
        throw("cannot authenticate");
    }

    try {
        await setRateRule(page, rateChangeSet);
    }
    finally {
        logger.debug("logout of escapia then exit browser");
        await logoutOfEscapia(page);
        await browser.close();
    }
}

const getWorkItems = (newRateSets, knownRates) => {
    let newChangeSet = []
    for ( let dateGroup in newRateSets ) {
        for ( let minStayGroup in newRateSets[dateGroup] ) {
            let changeSetObj = {date: dateGroup, minstay: minStayGroup, rentals: []}
            for ( let rental in newRateSets[dateGroup][minStayGroup] ) {
                changeSetObj.rentals.push({name: rental, rate: newRateSets[dateGroup][minStayGroup][rental]})
            }

            newChangeSet.push(changeSetObj)
        }
    }

    let workItems = []

    for ( let idx=0; idx<newChangeSet.length; idx++ ) {
        let name = newChangeSet[idx].date + '_' + newChangeSet[idx].minstay

        let setStatus = 'New' // Known, Updated
        for ( let kidx=0; kidx<knownRates.length; kidx++) {
            if ( knownRates[kidx].date == newChangeSet[idx].date && knownRates[kidx].minstay == newChangeSet[idx].minstay ) {
                knownRates[kidx].found = true
                setStatus = 'Known'
                if (!knownRates[kidx].rentals) {
                    knownRates[kidx].rentals = [];
                }
                if (knownRates[kidx].syncError) {
                    logger.debug(`getWorkItems: ${name} - Previously had an error`)
                    setStatus = 'Updated'
                }
                else if ( newChangeSet[idx].rentals.length != knownRates[kidx].rentals.length ) {
                    // logger.info('diff rental cnt: [' + knownRates[kidx].date + ' ** ' + knownRates[kidx].minstay + ']' + newChangeSet[idx].rentals.length + ';' + knownRates[kidx].rentals.length)
                    setStatus = 'Updated'
                }
                else {
                    let matchesFound = 0
                    for ( let csrIdx=0; csrIdx<newChangeSet[idx].rentals.length; csrIdx++ ) {
                        for ( let krIdx=0; krIdx<knownRates[kidx].rentals.length; krIdx++ ) {
                            if ( newChangeSet[idx].rentals[csrIdx].name === knownRates[kidx].rentals[krIdx].name &&
                                newChangeSet[idx].rentals[csrIdx].rate == knownRates[kidx].rentals[krIdx].rate) {
                                matchesFound++
                            }
                        }
                    }

                    if ( newChangeSet[idx].rentals.length != matchesFound ) {
                        // logger.info('rental/rate mismatch [' + knownRates[kidx].date + ' ** ' + knownRates[kidx].minstay + ']' + newChangeSet[idx].rentals.length + '/' + matchesFound)
                        setStatus = 'Updated'
                    }
                }

                // update in Escapia if any rental is out of sync,
                // or if the rate rule in Escapia has been expired but is now needed again,
                // or if the rate rule was left in an error state, even if it is in sync with the new rate specifies
                if ( setStatus === 'Updated') {
                    newChangeSet[idx].Name = name
                    newChangeSet[idx].Status = 'Update'
                    newChangeSet[idx].StatusDate = new Date()
                    newChangeSet[idx].rateId = knownRates[kidx]._id
                    workItems.push(newChangeSet[idx])
                }
                else {
                    // logger.info('exact match! ' + knownRates[kidx].date + ' ** ' + knownRates[kidx].minstay)
                }

                break;
            }
        }

        if ( setStatus === 'New' ) {
            newChangeSet[idx].Name = name
            newChangeSet[idx].Status = 'Add'
            newChangeSet[idx].StatusDate = new Date()
            workItems.push(newChangeSet[idx])
        }
    }

    for ( let kidx=0; kidx<knownRates.length; kidx++) {
        if ( typeof knownRates[kidx].found == 'undefined') {
            // logger.info('dead known rate? ' + JSON.stringify(knownRates[kidx]))
            let deleteRate = {}
            deleteRate.Status = 'Expire'
            deleteRate.StatusDate = new Date()
            deleteRate.Name = knownRates[kidx].date + '_' + knownRates[kidx].minstay
            deleteRate.rateId = knownRates[kidx]._id
            deleteRate.date = knownRates[kidx].date
            deleteRate.minstay = knownRates[kidx].minstay
            deleteRate.rentals = knownRates[kidx].rentals
            workItems.push(deleteRate)
        }
    }

    return workItems;
}

const getInvalidRentalsFromSf = async(rentalNames) => {
    const query = `
        Select
            Id,
            Name,
            Stage_Type__c
        From
            Opportunity
        Where
            (Stage_Type__c = 'Canceled'
            Or Escapia_Internal_Id__c = null)
            And Name In (${rentalNames.map(r => `'${r}'`).join(", ")})
    `;
    
    const rentalResults = await sfgen.query(query);

    return rentalResults.map(r => r.Name);
}

async function getRates() {
    let today = moment()
    let bqResults = await queryBq("SELECT date, name, mnst, rate, cadensa_pricing FROM `turnkey-data-warehouse.escapia_bot.robot_input` where mnst is not null");

    // Date property to object -> minstay number to array of rental objects with rates
    // ex.  { 2020-01-01: { 1: [{rental:"123RENTAL", rate:234}, {rental:"124RENTAL", rate:556}] }
    let rateGroups = {}
    bqResults.forEach(row => {
        let mDate = moment(row.date.value)

        if ( mDate.isAfter(today, 'day') ) {
            let dateStr = mDate.format('YYYY-MM-DD')
            // let rate = row.cadensa_pricing == null ? row.rate : row.cadensa_pricing
            let rate = row.rate

            if (typeof rateGroups[dateStr] == 'undefined')
                rateGroups[dateStr] = {}

            if (typeof rateGroups[dateStr][row.mnst] == 'undefined')
                rateGroups[dateStr][row.mnst] = []

            if (typeof rateGroups[dateStr][row.mnst][row.name] == 'undefined') {
                rateGroups[dateStr][row.mnst][row.name] = rate
            }
            else {
                if (rate != rateGroups[dateStr][row.mnst][row.name]) {
                    // logger.info('rate row dup and conflict ' + row.name + ' ' + dateStr + ':' + rate + '/' + rateGroups[dateStr][row.mnst][row.name])
                }
            }
        }
    })

    logger.info('done getting rates')
    return rateGroups
}