const { getBrowser } = require('../util/puppeteerUtil');
const { setAndGetConfigEscapia, authenticateWithEscapia, logoutOfEscapia } = require('../util/escapiaUtil');
const sfgen = require('@turnkeyvr/salesforcegeneral').promise;
const awsSns = require('@turnkeyvr/aws_sns');
const { promisify } = require("util");
const { sleep } = require('../util/misc');
const { createServiceOrder } = require('../util/serviceOrderUtil');
const env = require('@turnkeyvr/env');

const logger = require('@turnkeyvr/logger');
const promiseSendSns = promisify(awsSns.publishEmail);

const isProd = (process.env.NODE_ENV.toLowerCase() === 'prod' || process.env.NODE_ENV.toLowerCase() === 'production');

const mainHandler = async (event, context) => {
    logger.info('manageServiceOrders main running!')

    let sfWorkOrderCases = await sfgen.query(`Select Id, Owner.Name, CaseNumber, 
        OpportunityCase__r.Id, OpportunityCase__r.Escapia_Instance_Id__c, 
        OpportunityCase__r.Name, Amount_invoiced_to_TK_for_Service__c, 
        Amount_to_bill_HomeOwner_for_service__c, Description, Origin,
        Bill_Owner_For__c, If_other_please_explain__c
        from Case where Status='Field Ops Complete - Need to Enter Bill' and 
        Owner.Name = 'Smiles - Service Orders'`)

    /*
    let sfWorkOrderCases = await sfgen.query(`Select Id, CaseNumber, OpportunityCase__r.Id, 
        OpportunityCase__r.Name, 
        Amount_invoiced_to_TK_for_Service__c, Amount_to_bill_HomeOwner_for_service__c,
        Bill_Owner_For__c, If_other_please_explain__c
        from Case where CaseNumber='06578565'`)
    */

    logger.info('service order cases?' + sfWorkOrderCases.length)
    if ( sfWorkOrderCases.length > 0 ) {
        for (let idx=0; idx<sfWorkOrderCases.length; idx++) {
            let item = sfWorkOrderCases[idx]
            
            if ( item.Amount_to_bill_HomeOwner_for_service__c == null ) {
                if ( item.Description.indexOf('The system is not processing') == -1 ) {
                    let newDescription = '[The system is not processing this case because it lacks an amount to bill to homeowner]\n\n' + item.Description 
                                        
                    await doCaseUpdate(item, {Id: item.Id, Description:newDescription})
                    // await sfgen.update("Case", {Id: item.Id, Description:newDescription})
                }
            }
            else if ( item.OpportunityCase__r.Escapia_Instance_Id__c != null && item.OpportunityCase__r.Escapia_Instance_Id__c.length > 0 ) {
                logger.info('service order case:' + JSON.stringify(item))

                let reasonForRejection = ''
                if ( item.OpportunityCase__r == null ) 
                    reasonForRejection += 'No opportunity assigned to case.\n'
                if ( item.Amount_invoiced_to_TK_for_Service__c == null ) 
                    reasonForRejection += 'Amount invoiced to TK is empty.\n'
                if ( item.Bill_Owner_For__c == null ) 
                    reasonForRejection += 'Reason for billing owner has not been provided.\n'
                if ( item.Bill_Owner_For__c == 'Other' && item.If_other_please_explain__c == null)
                    reasonForRejection += 'Other reason for billing owner has not been provided.\n'

                if ( reasonForRejection.length > 0 ) {
                    logger.info(`service order can't be processed:` + item.Id + ':' + reasonForRejection)
                    reasonForRejection = 'The system could not process this bill to owner case because:\n\n' + reasonForRejection
                    await setCaseError(item, reasonForRejection)
                }
                else {
                    await doCaseUpdate(item, {Id: item.Id, OwnerId:'005i0000008nXsuAAE'})
                    // await sfgen.update("Case", {Id: item.Id, OwnerId:'005i0000008nXsuAAE'})
                    await promiseSendSns(env.config.get('snsTopics.processServiceOrder'), 'PROCESSWORKORDER', JSON.stringify(item))
                    // dev only
                    // await workerHandler({Records: [{Sns: {MessageId: '1234', Message: JSON.stringify(item)}}]}, null)
                }    
            }
            else {
                await setCaseError(item, 'This rental does not exist in Escapia - so the robot cannot process the service order.')
            }    
        }
    }
}

exports.mainHandler = mainHandler

const workerHandler = async (event, context) => {
    let snsEvent = event.Records[0].Sns
    let serviceOrderCase = JSON.parse(snsEvent.Message)
    logger.info('workerHandler doing ' + serviceOrderCase.Id + ';' + serviceOrderCase.Type)

    // this will fail...
    let successServiceOrderNumber = false
    let browser = null
    try {
        let escapiaConfig = setAndGetConfigEscapia(serviceOrderCase.OpportunityCase__r.Escapia_Instance_Id__c)

        browser = await getBrowser();
        if (!browser) {
            return;
        }

        logger.info('GOING TO LOGIN?')
        // authenticate with escapia
        let page = await authenticateWithEscapia(await browser.newPage());
        if (!page) {
            logger.info('Something bad happened?')
            await browser.close();
            return;
        }

        logger.info('YIPPEE I SURVIVED LOGGING IN!')

        // do the deed
        try {
            successServiceOrderNumber = await createServiceOrder(escapiaConfig, page, serviceOrderCase) 
        }
        catch (ex) {
            if ( ex == '[NO_RENTAL_FOUND]' ) 
                await setCaseError(serviceOrderCase, 'Rental not found in Escapia!')
            else 
                throw ex
        }

        // logout of escapia then exit browser
        await logoutOfEscapia(page);
        await browser.close();
    }
    catch (ex) {
        if ( !successServiceOrderNumber ) {
            logger.error('processing service order [' + ex + ']: ' + serviceOrderCase.Id)
        }
        
        if ( browser ) 
            await browser.close();
    }

    if ( successServiceOrderNumber )  {
        // close case
        // let caseCloseResult = await sfgen.update("Case", {Id: serviceOrderCase.Id, Status:'Closed', Escapia_Service_Number__c:successServiceOrderNumber})
        let caseCloseResult = await doCaseUpdate(serviceOrderCase, {Id: serviceOrderCase.Id, Status:'Closed', Escapia_Service_Number__c:successServiceOrderNumber})
        logger.info('caseCloseResult? ' + JSON.stringify(caseCloseResult))

        let commentObj = {
            CommentBody__c: successServiceOrderNumber,
            Comment_Created_By__c: '005i0000008nXsuAAE',
            Guest_Impacting__c: false,
            Hide_Comment_from_Owner__c: true,
            Is_case_resolution__c: false,
            Parent_Case__c: serviceOrderCase.Id
        }

        let caseCommentInsertResult = await sfgen.insert('Case_Comment_2__c', commentObj)
        logger.info('caseCommentInsertResult? ' + JSON.stringify(caseCommentInsertResult))

        // TODO ADD CaseComment
        logger.info(`${snsEvent.MessageId}: done with service order - success [REZID: ${serviceOrderCase.Id}]`);
    }
    else {
        // re-assign case?
        logger.info(`${snsEvent.MessageId}: done with service order - error [REZID: ${serviceOrderCase.Id}]`);
        // 00Gi0000004b3qFEAQ: 'Smiles - Service Orders'
        await doCaseUpdate(serviceOrderCase, {Id: serviceOrderCase.Id, OwnerId:'00Gi0000004b3qFEAQ'})
    }

    return
}

exports.workerHandler = workerHandler

const doCaseUpdate = async (oldCase, updateObj) => {
    if ( oldCase.Origin.indexOf('Owner') > -1 )
        updateObj.Origin = 'Internal'

    let result = await sfgen.update("Case", updateObj)
    return result
}

const setCaseError = async (errCase, caseCommentForError) => {
    // 00Gi0000001o799EAA - "Homeowner Support" Queue Id
    // let caseReassignResult = await sfgen.update("Case", {Id: errCase.Id, OwnerId:'00Gi0000001o799EAA'})
    let caseReassignResult = await doCaseUpdate(errCase, {Id: errCase.Id, OwnerId:'00Gi0000001o799EAA'})
    logger.info('caseReassignResult? ' + JSON.stringify(caseReassignResult))

    let commentObj = {
        CommentBody__c: caseCommentForError,
        Comment_Created_By__c: '005i0000008nXsuAAE', // API USER
        Guest_Impacting__c: false,
        Hide_Comment_from_Owner__c: true,
        Is_case_resolution__c: false,
        Parent_Case__c: errCase.Id
    }

    let caseCommentInsertResult = await sfgen.insert('Case_Comment_2__c', commentObj)
    logger.info('caseCommentInsertResult? ' + JSON.stringify(caseCommentInsertResult))
}