const { getBrowser } = require('../util/puppeteerUtil');
const { retry } = require('../util/retry');
const { setAndGetConfigEscapia, authenticateWithEscapia } = require('../util/escapiaUtil');
const logger = require('@turnkeyvr/logger');
const async = require('async')
const sfgen = require('@turnkeyvr/salesforcegeneral').promise;
const awsSns = require('@turnkeyvr/aws_sns');
const { promisify } = require("util");
const env = require('@turnkeyvr/env');

const promiseSendSns = promisify(awsSns.publishEmail);

const isProd = (process.env.NODE_ENV.toLowerCase() === 'prod' || process.env.NODE_ENV.toLowerCase() === 'production');

const handler = async () => {
    if (!isProd) {
        logger.debug('monitor amenities only runs in production');
        return;
    }
    
    let browser = await getBrowser();
    if (!browser) {
        return;
    }

    // authenticate with escapia
    setAndGetConfigEscapia('1655'); // only running against the main instance
    let page = await authenticateWithEscapia(await browser.newPage());
    if (!page) {
        await browser.close();
        return;
    }

    await page.goto('https://turnkeyvr.escapia.com/webagency/Property/EditProperty.aspx?id=142632', {waitUntil: 'load'});
    await page.waitForSelector('#ui-id-5');
    await page.click('#ui-id-5');

    let fromEscapiaAmenityArr = await getAmenityList(page, 1)
    // logger.info('amenities: ' + amenityArr.length)

    let tabs = await getAmenityTabs(page)
    for ( let idx=0; idx<tabs.length; idx++ ) {
        // logger.info('tab:' + tabs[idx])
        await retry(
            () => Promise.all([
                page.click("a[href*='" + tabs[idx] + "']"),
                page.waitForNavigation({waitUntil:'networkidle0'})
            ]), 5
        );

        // await page.click("a[href*='" + tabs[idx] + "']");
        // await page.waitForNavigation({waitUntil:'networkidle0'})

        let appendAmenityArr = await getAmenityList(page, (idx+2))
        // logger.info('appendAmenityArr: ' + appendAmenityArr.length)
        fromEscapiaAmenityArr.push.apply(fromEscapiaAmenityArr, appendAmenityArr);
    }

    logger.info('amenityArr size: ' + fromEscapiaAmenityArr.length)
    // logger.debug(JSON.stringify(amenityArr))
    await browser.close();

    let sfAmenities = await sfgen.query(`Select Id, Amenity_Tab_Number__c, Feature_ID__c, Feature_Name__c
        from Escapia_Amenity_Reference__c`)

    let newAmenityArr = []
    for ( let eidx=0; eidx<fromEscapiaAmenityArr.length; eidx++) {
        let foundAmenity = false
        let eAmenity = fromEscapiaAmenityArr[eidx]
        for ( let idx=0; idx<sfAmenities.length; idx++ ) {
            let sfAmenity = sfAmenities[idx]
            if ( sfAmenity.Feature_ID__c == eAmenity.id ) {
                foundAmenity = true
                break
            }
        }

        if ( !foundAmenity ) {
            logger.debug(eAmenity.id + ': did not find amenity!')
            newAmenityArr.push(eAmenity)

            let amenityObj = {
                Amenity_Tab_Number__c: eAmenity.tabNumber,
                Feature_Name__c: eAmenity.name,
                Feature_ID__c: eAmenity.id
            }

            await sfgen.insert('Escapia_Amenity_Reference__c', amenityObj)
        }
    }

    if (newAmenityArr.length > 0 ) {
        logger.info('will try to send an update email!');

        let emailBody = `The following amenities have been added to SF and might need to be mapped to TurnKey amenities:`
        emailBody += `\n\nTab\tName`

        for ( let idx=0; idx<newAmenityArr.length; idx++ ) {
            emailBody += `\n` + newAmenityArr[idx].tabNumber +
                `\t` + newAmenityArr[idx].name
        }

        let subject = 'New Escapia Amenities Discovered!'
        await promiseSendSns(env.config.get('snsTopics.monitorAmenitiesEmail'), subject, emailBody)
    }

    logger.info('monitorAmenityList done');

    return
}

async function getAmenityTabs(page) {
    const tabs = await page.evaluate(() => {
        let tabTable = document.querySelector("#PropertyWizard_ctl09_lstTabs")
        let tabNodes = tabTable.querySelectorAll("a")

        let retArr = []
        for (let idx = 0; idx < tabNodes.length; idx++) {
            let firstIdx = tabNodes[idx].href.indexOf("'") + 1
            let lastIdx = tabNodes[idx].href.indexOf("$ctl00'", firstIdx)

            retArr.push(tabNodes[idx].href.substring(firstIdx, lastIdx))
        }

        return retArr
    });

    return tabs;
}

async function getAmenityList(page, tabNumber) {
    const amenities = await page.evaluate(tabNumber => {

        let nodeList = document.querySelectorAll("input[id^='lstAmenities_']")

        let retArr = []
        for (let idx = 0; idx < nodeList.length; idx++) {
            let labelNode = document.querySelector('label[for="' + nodeList[idx].id + '"]')

            let noiseIdx = labelNode.innerText.indexOf(':') + 1
            if ( noiseIdx > -1 ) {
                labelNode.innerText = labelNode.innerText.substring(noiseIdx).trim()
            }

            retArr.push({
                id: nodeList[idx].id.replace('lstAmenities_', ''),
                name: labelNode.innerText,
                tabNumber: tabNumber
            })
        }

        return retArr
    }, tabNumber);

    return amenities;
}

exports.handler = handler