const moment = require('moment');
const eachSeries = require('async/eachSeries');
const logger = require('@turnkeyvr/logger');
const MongoDb = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('rate-reconciliation-prod');
const sfGen = require('@turnkeyvr/salesforcegeneral');
const util = require('util');
const { getBrowser } = require('../util/puppeteerUtil');
const { authenticateWithEscapia,
        setAndGetConfigEscapia,
        removePendingReconciliationReservationNote,
        checkForPendingReconciliationReservationNote,
        logoutOfEscapia } = require('../util/escapiaUtil');
const { resultsEnum, sleep } = require('../util/misc');

const querySF = util.promisify(sfGen.query);

const isProd = (process.env.NODE_ENV.toLowerCase() === 'prod' || process.env.NODE_ENV.toLowerCase() === 'production');

const checkForTimeouts = async () => {
    // only runs in production
    if (!isProd) {
        logger.debug('rate reconciliation only runs in production');
        return;
    }

    // pending reconciliation case query; looking for records that are 2+ days old and stuck as "In Progress"
    let caseQuery = 'SELECT Id, '
        + 'Subject, '
        + 'CreatedDate '
        + 'FROM Case '
        + 'WHERE Subject IN (\'Finalize Airbnb reservation in Escapia\', \'Finalize Website reservation in Escapia\') '
        + 'AND Status = \'In Progress\' '
        + 'AND OpportunityCase__r.Escapia_Instance_Id__c = \'1655\' '
        + 'AND CreatedDate < Last_N_Days:2';

    let caseRecords = await querySF(caseQuery);

    if (caseRecords.records.length > 0) {
        caseRecords.records.forEach(caseRecord => {
            logger.alert(`Warning: pending rate reconciliation Case >2 days old: ${JSON.stringify(caseRecord)}`);
        });
    }
};

const handler = async () => {
    // only runs in production
    if (!isProd) {
        logger.debug('rate reconciliation only runs in production');
        return;
    }

    try {
        let today = moment().toDate();
        let yesterday = moment().add(-1, 'days').toDate();
        let results = await MongoDb.find('recheck', { reconciliationDate: { $lte: yesterday } }, { limit: 3 });
        logger.debug(`results: ${JSON.stringify(results)}`);

        if (results.length === 0) {
            return;
        }

        // load browser
        let browser = await getBrowser();
        if (!browser) {
            return;
        }

        // authenticate with escapia
        setAndGetConfigEscapia('1655'); // only running against the main instance
        let page = await authenticateWithEscapia(await browser.newPage());
        if (!page) {
            await browser.close();
            return;
        }

        await eachSeries(results, async (record) => {
            // see if it's in progress
            let pendingReconciliationNotePresentResults = await checkForPendingReconciliationReservationNote(page, record.escapiaUnitReservationId);

            // no flag present -- delete mongo record
            if (pendingReconciliationNotePresentResults === resultsEnum.FAILURE) {
                logger.debug('no flag present');
                await MongoDb.deleteOneById('recheck', record._id);
                return;
            }

            // throw alert if attempts > 3
            if (record.attempts > 3) {
                logger.alert(`removing pending rec flag failed > 3: ${record.escapiaUnitReservationId}`);
            }

            // otherwise, remove from page + create a new mongo record to check again tomorrow

            // remove the pending reconciliation note from escapia reservation notes
            let removePendingReconciliationReservationNoteResults = await removePendingReconciliationReservationNote(page, record.escapiaUnitReservationId);
            if (removePendingReconciliationReservationNoteResults === resultsEnum.FAILURE) {
                logger.error(`unable to remove pending reconciliation flag from Escapia: ${record.escapiaUnitReservationId}`);
                return;
            }

            // make sure pending reconciliation note was remove from escapia reservation
            let pendingReconciliationNotePresentFinalResults = await checkForPendingReconciliationReservationNote(page, record.escapiaUnitReservationId)
            if (pendingReconciliationNotePresentFinalResults !== resultsEnum.FAILURE) {
                logger.error(`postcheck: unable to remove pending reconciliation flag from Escapia: ${record.escapiaUnitReservationId}`)
                return;
            }

            // create new mongo record to recheck
            let newRecord = {
                reconciliationDate: today,
                escapiaUnitReservationId: record.escapiaUnitReservationId,
                sfdcCaseId: record.sfdcCaseId,
                attempts: record.attempts + 1,
                source: record.source
            };
            await MongoDb.insertMany('recheck', [newRecord]);

            // delete old mongo record
            await MongoDb.deleteOneById('recheck', record._id);
        });

        await MongoDb.close();

        // logout of escapia then exit browser
        await logoutOfEscapia(page);
        await browser.close();
    }
    catch (e) {
        try {
            await MongoDb.close();
        }
        finally {
            logger.error('Rate Reconciliation Recheck Failed');
            logger.error(e.message);
            return;
        }
    }

    logger.debug('done');
};



exports.checkForTimeouts = checkForTimeouts;
exports.handler = handler;