const { getBrowser } = require('../util/puppeteerUtil');
const { retry } = require('../util/retry');
const { setAndGetConfigEscapia, authenticateWithEscapia, logoutOfEscapia } = require('../util/escapiaUtil');
const logger = require('@turnkeyvr/logger');
const sfgen = require('@turnkeyvr/salesforcegeneral').promise;
const env = require('@turnkeyvr/env');
const awsSns = require('@turnkeyvr/aws_sns');
const { promisify } = require("util");
const promiseSendSns = promisify(awsSns.publishEmail);

// mainHandler runs hourly
// workerHandler seems to take around 40 seconds
// workerHandler has concurrency of 3
// so mainHandler should be able to handle 270 listings each run? let's set to 200...
// const LISTING_COUNT = 200;
const LISTING_COUNT = 10; // TODO: Switch to 200 later. setting this lower initially, just to be safe.

const isProd = (process.env.NODE_ENV.toLowerCase() === 'prod' || process.env.NODE_ENV.toLowerCase() === 'production');

const mainHandler = async (event, context) => {
    if (!isProd) {
        logger.debug('set amenities only runs in production');
        return;
    }
    
    let listings = await getListingsToProcessFromSf();
    logger.info(`${listings.length} listings to process`);
    if (listings.length > 0) {
        logger.info(listings);
    }

    for (let i = 0; i < listings.length; i++) {
        let listing = listings[i];
        await promiseSendSns(env.config.get('snsTopics.setAmenities'), 'SETAMENITIES', JSON.stringify(listing));
        // TESTING ONLY!
        // await workerHandler({Records: [{Sns: {MessageId: '1234', Message: JSON.stringify(listing)}}]}, null)
    }
}
exports.mainHandler = mainHandler;

const getListingsToProcessFromSf = async() => {
    let rentalResults = await sfgen.query(`
        Select
            Id,
            Name,
            Marketing_Content__c,
            Escapia_Internal_Id__c
        From
            Opportunity
        Where
            Escapia_Internal_Id__c != null
            And Escapia_Instance_Id__c = '1655'
            And Stage_Type__c not in ('Live', 'Canceled')
            And Marketing_Content__r.Amenity_Sync_Pending__c = true
        Limit ${LISTING_COUNT}
    `);
    
    
    // The initial rollout is for new listings that are being worked by the Onboarding team
    // For full rollout, modify the following part of the query:
    //      And Stage_Type__c not in ('Live', 'Canceled')

    return rentalResults
    .filter(res => {
        return !!res.Escapia_Internal_Id__c && res.Escapia_Internal_Id__c.indexOf('TMP_') == -1;
    })
    .map(res => {
        return {
            listingContentId: res.Marketing_Content__c,
            escapiaId: res.Escapia_Internal_Id__c.substr(5) // 1655-103920 --> 103920
        };
    });
}

const workerHandler = async (event, context) => {
    let syncDate = new Date();

    let snsEvent = event.Records[0].Sns;
    let setAmenitiesArgs = JSON.parse(snsEvent.Message);
    const {listingContentId, escapiaId} = setAmenitiesArgs;
    
    logger.info(`listingContentId: ${listingContentId}, escapiaId: ${escapiaId}: start`);
    let browser = await getBrowser();
    if (!browser) {
        throw "cannot get browser";
    }

    // authenticate with escapia
    setAndGetConfigEscapia('1655'); // only running against the main instance
    let page = await authenticateWithEscapia(await browser.newPage());
    if (!page) {
        await browser.close();
        throw("cannot authenticate");
    }

    let success = false;
    try {
        await updateAmenities(page, listingContentId, escapiaId);
        success = true;
        await logoutOfEscapia(page);
    }
    finally {
        await browser.close();
    }

    if (success) {
        await updateSyncDateInSalesforce(listingContentId, syncDate);
    }

    logger.info(`${listingContentId}: done`);
}
exports.workerHandler = workerHandler;

const updateSyncDateInSalesforce = async(listingContentId, syncDate) => {
    await sfgen.update("Backing_Marketing_Content__c", {Id: listingContentId,
        Date_Amenities_Synced_To_Escapia__c: syncDate
    });
}

const updateAmenities = async(page, listingContentId, escapiaId) => {
    // does not throw an exception if page not found, because we get redirected to an error page.
    // However the next call (waitForSelector will throw an exception)
    await page.goto(`https://turnkeyvr.escapia.com/webagency/Property/EditProperty.aspx?id=${escapiaId}`, {waitUntil: 'load'});

    // throws an exception after the default timeout of 30000 ms if selector not found
    await page.waitForSelector('#ui-id-5');

    // throws an exception if selector not found
    await page.click('#ui-id-5');

    let amenityIdsByTabId = await getAmenityIdsByTabIdFromSf(listingContentId);

    // skip listing if there are no amenities for it in Salesforce.
    // this is to avoid unchecking all the amenities in Escapia if the amentities have not been set up in Salesforce yet
    if (Object.keys(amenityIdsByTabId).length == 0) {
        logger.warn(`${listingContentId}: no amenities in salesforce.  skipping.`);
        return;
    }

    let totalModified = 0;

    const TAB_IDS = [2,3,4,5,6,7,8,9,10,11,1]; // process tab 1 last because on initial load it is not clickable

    for (let i = 0; i < TAB_IDS.length; i++) {
        let tabId = TAB_IDS[i];
        let modifiedCount = await oneTab(page, tabId, amenityIdsByTabId[tabId] || []);
        totalModified = totalModified + modifiedCount;
    }

    if (totalModified > 0) {
        logger.info(`${listingContentId}: ${totalModified} amenities modified`);
        await saveChanges(page);
    }
    else {
        logger.warn(`${listingContentId}: no changes to save`);
    }
}

const getAmenityIdsByTabIdFromSf = async(listingContentId) => {
    const query = `
    Select        
        TurnKey_Amenity__r.Name,
        Escapia_Amenity_Reference__r.Amenity_Tab_Name__c,
        Escapia_Amenity_Reference__r.Amenity_Tab_Number__c,
        Escapia_Amenity_Reference__r.Feature_ID__c,
        Escapia_Amenity_Reference__r.Feature_Name__c
    From
        TurnKey_Amenity_Escapia_Reference__c
    Where
        TurnKey_Amenity__c In
        (
            Select
                TurnKey_Amenity__c
            From
                TurnKey_Amenity_Listing_Content__c
            Where
                Listing_Content__c = '${listingContentId}'
        )
    `;
    
    let escapiaAmenities = await sfgen.query(query);
    let amenityIdsByTabId = {};
    let totalEscapiaAmenities = 0;
    escapiaAmenities.forEach(result => {
        let tabId = result.Escapia_Amenity_Reference__r.Amenity_Tab_Number__c;
        let featureId = result.Escapia_Amenity_Reference__r.Feature_ID__c;
        if (!amenityIdsByTabId[tabId]) {
            amenityIdsByTabId[tabId] = [];
        }
        // don't include duplicates - 2 TK amenities mapped to the same Escapia amenity, for example
        if (!amenityIdsByTabId[tabId].includes(featureId)) {
            amenityIdsByTabId[tabId].push(featureId);
            totalEscapiaAmenities++;
        }
    });

    logger.info(`${listingContentId}: total escapia amenities: ${totalEscapiaAmenities}`);
    
    return amenityIdsByTabId;
}

const oneTab = async(page, tabId, amenityIds) => {
    logger.debug(`oneTab: tabId: ${tabId}, escapia amenities: ${amenityIds.length}`);

    let urlTabId = tabId - 1;
    let urlTabIdFormatted = ("00" + urlTabId).slice(-2); // format integer as 2 digits
    let tabLinkFragment = `PropertyWizard$ctl09$lstTabs$ctl${urlTabIdFormatted}`;

    await retry(
        () => Promise.all([
            // throws an exception if selector not found
            page.click("a[href*='" + tabLinkFragment + "']"),
            page.waitForNavigation({waitUntil:'networkidle0'})
        ]), 5
    );

    let updateResult = await updateAmenitiesOnOneTab(page, amenityIds);

    if (updateResult.total == 0) {
        throw new Error("no amenity checkboxes found");
    }

    logger.debug(`oneTab: tabId: ${tabId}, modified: ${updateResult.modified} of ${updateResult.total}`);

    return updateResult.modified;
}

async function updateAmenitiesOnOneTab(page, amenityIds) {
    const result = await page.evaluate(amenityIds => {
        let modified = 0;
        let total = 0;
        let nodeList = document.querySelectorAll("input[id^='lstAmenities_']");
        nodeList.forEach(node => {
            let amenityId = node.id.replace('lstAmenities_', '');
            let shouldBeChecked = amenityIds.includes(amenityId);
            let el = document.getElementById(node.id);
            if (el.checked != shouldBeChecked) {
                el.checked = shouldBeChecked;
                modified++;
            }
            total++;
        });
        return {modified, total};
    }, amenityIds);
    return result;
}

const saveChanges = async(page) => {
    // throws an exception if selector not found
    await page.click("input[id^='PropertyWizard_SaveBtn']");
}