const { getBrowser } = require('../util/puppeteerUtil');
const { setAndGetConfigEscapia, authenticateWithEscapia, logoutOfEscapia } = require('../util/escapiaUtil');
const logger = require('@turnkeyvr/logger');
const sfgen = require('@turnkeyvr/salesforcegeneral').promise;
const env = require('@turnkeyvr/env');
const awsSns = require('@turnkeyvr/aws_sns');
const { promisify } = require("util");
const promiseSendSns = promisify(awsSns.publishEmail);
const { getSelectOptions } = require('../util/selectOptionsUtil');
const { sleep } = require('../util/misc');
const moment = require('moment');
const mongoDb = new (require('@turnkeyvr/mongodbhelper').MongoDbHelper)('robotify-rental-guardian-prod');

// mainHandler runs hourly
// workerHandler seems to take around 40 seconds
// workerHandler has concurrency of 7
const CALL_LIMIT = 1000;

// Main runs every 60 minutes, so worker should discard events older than the last run.
// But allow for worker operation taking up to 120 seconds
// This avoids duplicate processing of rentals, and a buildup of sns messages.
const DISCARD_EVENTS_OLDER_THAN_MINUTES = 57;

const tabDetailsSelector = "#ui-id-1";
const tabUnitAndOwnersSelector = "#ui-id-2";
const tabChargesAndFeesSelector = "#ui-id-3";

const mongoCollectionRentals = "rentals";
const mongoCollectionWorkerLog = "workerLog";

const dateFormat = "MM/DD/YYYY";

const mainHandler = async (event, context) => {
    const sfRentals = await getRentalsFromSf();
    logger.info(`${sfRentals.length} Salesforce rentals`);
    
    const mongoRentals = await mongoDb.find(mongoCollectionRentals, {});
    logger.info(`${mongoRentals.length} Mongo rentals`);
    const mongoRentalsBySalesforceRentalId = {};
    mongoRentals.forEach(mongoRental => mongoRentalsBySalesforceRentalId[mongoRental.salesforceRentalId] = mongoRental);

    // errors will normally get retried each run
    // syncErrorTimestamp and syncErrorCount will get updated with each failure, and cleared on success
    // if the problem is fixed in salesforce, for example by moving a rental to canceled stage, then these values will never get cleared
    // warn for rentals that have 20 errors or more in the last 48 hours
    const warnAfterXErrors = 20;
    let unrecoveredErrors = mongoRentals.filter(rental => {
            return  moment().diff(rental.syncErrorTimestamp, "hours") < 48 && rental.syncErrorCount > warnAfterXErrors
        });
    if (unrecoveredErrors.length > 0) {
        logger.warn(`Unrecovered errors: ${unrecoveredErrors.length} rentals that have more than ${warnAfterXErrors} sync errors`);
        logger.warn("Rentals with unrecovered errors (first 100):", unrecoveredErrors.slice(0,100).map(rental => rental.salesforceRentalName));
    }

    const feeNameStartsWith = "TurnKey Rental Umbrella (TRU)";

    let rentalCountWithNoRGService = 0;
    const mongoRentalsWithInvalidContract = [];

    // Compare Salesforce data to Mongo data (which includes Escapia state from previous updates)
    // Determine a sync action - add, update or delete - and prepare data to pass to the worker
    // The worker will make the changes in Escapia and then update mongo to reflect the latest Escapia state

    // The most recent Rental Service object on the Salesforce rental indicates whether the fee should exist in Escapia.
    // Checking the "Enrolled in RentalGuardian" checkbox on the Rental in Salesforce results in a Rental Service object getting created.
    // The Rental Service object will have a Start Date of the first day of the current month, and a null End Date
    //      Action: The fee should be added in Escapia
    // If the checkbox is unchecked, the End Date will get set to the last day of the current month
    //      Action: The fee should be updated in Escapia - update the End Date
    //      Action: Once the End Date has passed, the fee should be deleted in Escapia
    const rentals = sfRentals.map(sfRental => {
        const service = sfRental.Rental_Services__r && sfRental.Rental_Services__r.records && sfRental.Rental_Services__r.records.length > 0
            ? sfRental.Rental_Services__r.records[0]
            : null;
        const mapped = {
            salesforceRentalId: sfRental.Id,
            salesforceRentalName: sfRental.Name,
            Escapia_Internal_Id__c: sfRental.Escapia_Internal_Id__c,
            Escapia_Instance_Id__c: sfRental.Escapia_Instance_Id__c,
            robotData: {
                sfData: {
                    BO_Bedrooms__c: sfRental.BO_Bedrooms__c,
                    rentalGuardianService: service
                        ? {
                            Id: service.Id,
                            Service_Start_Date__c: service.Service_Start_Date__c,
                            Service_End_Date__c: service.Service_End_Date__c,
                            serviceName: service.Service__r.Name,
                            CreatedDate: service.CreatedDate
                        }
                        : null
                }
            }
        };

        mapped.robotData.feeNameStartsWith = feeNameStartsWith;

        if (!mapped.robotData.sfData.rentalGuardianService) {
            mapped.robotData.deleted = true;
            rentalCountWithNoRGService++;
        }
        else {
            mapped.robotData.escapiaFields = mapToEscapiaFields(mapped.robotData);
        }

        const mongoRental = mongoRentalsBySalesforceRentalId[sfRental.Id];

        if (mongoRental) {
            mapped.previousCallHadError = !!mongoRental.syncError;
            mapped.mongoId = mongoRental._id;
            mapped.syncErrorCount = mongoRental.syncErrorCount || 0;

            if (mongoRental.syncError
                && typeof mongoRental.syncError === "string"
                && (mongoRental.syncError.indexOf("Contract not found for") == 0
                    || mongoRental.syncError.indexOf("edit contract screen loaded for") == 0)) {
                mongoRentalsWithInvalidContract.push(mongoRental);
            }
        }

        if (!mongoRental || !mongoRental.escapiaState) {
            if (mapped.robotData.deleted) {
                logger.info(`${mapped.salesforceRentalName}: Fee should be deleted, but it was never added`);
                // delete the fee in escapia in case it is there due to a manual action (related to ownership change, for example)
                // even though this robot never originally added it
                mapped.syncAction = "delete";
            }
            else {
                mapped.syncAction = "add";
            }
        }
        else {
            if (mapped.robotData.deleted && !mongoRental.escapiaState.deleted) {
                mapped.syncAction = "delete";
            }
            else if (!mapped.robotData.deleted && mongoRental.escapiaState.deleted) {
                mapped.syncAction = "add";
            }
            else if (!mapped.robotData.deleted && updateNeeded(mongoRental.escapiaState, mapped.robotData)) {
                mapped.syncAction = "update";
            }
            else {
                // nothing to do
            }
        }

        return mapped;
    });

    logger.info(`${rentalCountWithNoRGService} Salesforce rentals that do not have a rental service`);

    const deletes = rentals.filter(rental => rental.syncAction == "delete");
    if (deletes.length > 0) {
        logger.info("Total deletes needed", deletes.length);
    }
    const adds = rentals.filter(rental => rental.syncAction == "add");
    logger.info("Total adds needed", adds.length);
    const updates = rentals.filter(rental => rental.syncAction == "update");
    logger.info("Total updates needed", updates.length);

    let calls = rentals.filter(rental => rental.syncAction && !rental.previousCallHadError);
    logger.info("Total calls needed (non-retry)", calls.length);
    let retryCalls = rentals.filter(rental => rental.syncAction && rental.previousCallHadError);
    logger.info("Total retry calls needed", retryCalls.length);
    calls = calls.slice(0, CALL_LIMIT);
    logger.info("Calls (non-retry)", calls.length);
    const retryCallLimit = CALL_LIMIT - calls.length;
    retryCalls = retryCalls.slice(0, retryCallLimit);
    logger.info("Retry calls", retryCalls.length);
    calls = calls.concat(retryCalls);
    logger.info("Total calls this run", calls.length);

    if (mongoRentalsWithInvalidContract.length > 0) {
        logger.info(`Invalid contracts in Escapia: ${mongoRentalsWithInvalidContract.map(r => r.salesforceRentalName).join(",")}`);
    }

    if (calls.length > mongoRentalsWithInvalidContract.length) {
        logger.info("Total pending items", (calls.length - mongoRentalsWithInvalidContract.length));
    }

    for (let i = 0; i < calls.length; i++) {
        const call = calls[i];
        call.syncQueuedTimestamp = new Date();
        await promiseSendSns(env.config.get('snsTopics.setRentalGuardianFee'), 'SETRENTALGUARDIANFEE', JSON.stringify(call));
        // TESTING ONLY!
        // await workerHandler({Records: [{Sns: {MessageId: '1234', Message: JSON.stringify(call)}}]}, null);
    }

    await mongoDb.close();
}
exports.mainHandler = mainHandler;

const getRentalsFromSf = async() => {
    const query = `
        Select
            Id,
            Name,
            BO_Bedrooms__c,
            Escapia_Internal_Id__c,
            Escapia_Instance_Id__c,
            (
                Select
                    Id,
                    Service_Start_Date__c,
                    Service_End_Date__c,
                    Service__r.Name,
                    CreatedDate
                From
                    Rental_Services__r
                Where
                    Service__r.Name = 'RentalGuardian'
                Order By
                    CreatedDate Desc
                Limit 1
            )
        From
            Opportunity
        Where
            Escapia_Instance_Id__c = '1655'
            And Escapia_Internal_Id__c != null
            And Stage_Type__c In ('Live', 'Test')
        Order By LastModifiedDate Desc
    `;

    // Order By LastModifiedDate Desc: this prioritizes recently modified rentals,
    // as well as avoiding the same rentals getting processed over and over if there are problems
    
    const rentalResults = await sfgen.query(query);

    return rentalResults.filter(res => {
        return !!res.Escapia_Internal_Id__c && res.Escapia_Internal_Id__c.indexOf('TMP_') == -1;
    });
}

const updateNeeded = (currentEscapiaState, newEscapiaState) => {
    if (newEscapiaState.sfData.rentalGuardianService && !currentEscapiaState.sfData.rentalGuardianService) {
        return true;
    }

    return currentEscapiaState.sfData.BO_Bedrooms__c != newEscapiaState.sfData.BO_Bedrooms__c
        || currentEscapiaState.sfData.rentalGuardianService.Service_Start_Date__c != newEscapiaState.sfData.rentalGuardianService.Service_Start_Date__c
        || currentEscapiaState.sfData.rentalGuardianService.Service_End_Date__c != newEscapiaState.sfData.rentalGuardianService.Service_End_Date__c;
}

const mapToEscapiaFields = (robotData) => {
    const {sfData, feeNameStartsWith} = robotData;

    const maxNightsToCharge = 10;
    const nightlyFee = (sfData.BO_Bedrooms__c > 2) ? 9 : 7;

    return {
        name: `${feeNameStartsWith} $${nightlyFee}`,
        uponCancellation: "Remove",
        effectiveStartDate: moment(sfData.rentalGuardianService.Service_Start_Date__c).format(dateFormat),
        endDate: !sfData.rentalGuardianService.Service_End_Date__c
            ? "01/01/2075"
            : moment(sfData.rentalGuardianService.Service_End_Date__c).format(dateFormat),
        amount: nightlyFee,
        minimumAmount: nightlyFee,
        maximumAmount: nightlyFee * maxNightsToCharge,
        calculationMethod: "Per Night",
        journalDocumentTemplate: `TurnKey Rental Umbrella (TRU) $${nightlyFee} [TurnKey Rental Umbrella (TRU) $${nightlyFee}]`,
        postingPeriod: "Last",
        reservationTypes:  [
            "Renter",
            "Owner Referral",
            "Travel Agent",
            "Group",
            "Long Term"
        ]
    };
}

const workerHandler = async (event, context) => {
    let snsEvent = event.Records[0].Sns;
    let args = JSON.parse(snsEvent.Message);
    const unitCode = args.salesforceRentalName;
    const instanceId = args.Escapia_Instance_Id__c;

    logger.debug("args", args);

    logger.info(`unitCode: ${unitCode}: sync action: ${args.syncAction}: start`);

    if (moment().diff(args.syncQueuedTimestamp, "minutes") > DISCARD_EVENTS_OLDER_THAN_MINUTES) {
        logger.info(`unitCode: ${unitCode}: quit because timestamp too old`, args.syncQueuedTimestamp);
        return;
    }

    if (instanceId != '1655') {
        logger.error(`unitCode: ${unitCode}: quit because escapia instance id ${instanceId} is not supported`);
        return;
    }
    
    args.syncStartedTimestamp = new Date();
    
    let browser = await getBrowser();
    if (!browser) {
        throw "cannot get browser";
    }

    // authenticate with escapia
    setAndGetConfigEscapia(instanceId); // only running against the main instance
    let page = await authenticateWithEscapia(await browser.newPage());
    if (!page) {
        await browser.close();
        throw("cannot authenticate");
    }

    let error = null;
    try {
        if (args.syncAction == "delete") {
            await deleteFee(page, unitCode, args.robotData.feeNameStartsWith);
        }
        else {
            // add, update are handled by addFee.  Any existing fee is first deleted.
            await addFee(page, unitCode, args.robotData.escapiaFields, args.robotData.feeNameStartsWith);
        }
    }
    catch(e){
        error = e;
    }
    finally {
        await logoutOfEscapia(page);
        await browser.close();
    }

    if (error) {
        args.syncError = error;
        args.syncErrorTimestamp = new Date();
        args.syncErrorCount = (args.syncErrorCount || 0) + 1;
        args.syncCompletedTimestamp = null;
    }
    else {
        args.syncError = null;
        args.syncErrorTimestamp = null;
        args.syncErrorCount = 0;
        args.syncCompletedTimestamp = new Date();
    }

    if (error) {
        logger.error(`unitCode: ${unitCode}: sync action: ${args.syncAction}: ERROR`, error);
    }
    else {
        logger.info(`unitCode: ${unitCode}: sync action: ${args.syncAction}: Success`);
    }

    logger.debug("Update mongo");
    
    let mongoId = args.mongoId;
    delete args.mongoId;
    delete args.previousCallHadError;
    const logItem = Object.assign({}, args);
    await mongoDb.insertMany(mongoCollectionWorkerLog, [logItem]);

    if (!error) {
        // escapiaState in mongo is the state in escapia after our changes.
        // If an error occurs and we're not sure what the state is, leave this as-is in mongo
        args.escapiaState = args.robotData;
    }
    delete args.robotData;

    if (!mongoId) {
        const mongoRentals = await mongoDb.find(mongoCollectionRentals, {salesforceRentalId: args.salesforceRentalId});
        if (mongoRentals && mongoRentals.length > 0) {
            mongoId = mongoRentals[0]._id;
            logger.info("found existing mongo document for", args.salesforceRentalId);
        }
    }

    await mongoDb.updateOneById(mongoCollectionRentals, mongoId, args, {upsert: true});

    await mongoDb.close();

    logger.debug("Done updating mongo");
}
exports.workerHandler = workerHandler;

const deleteFee = async(page, unitCode, feeNameStartsWith) => {
    logger.debug(`deleteFee`, unitCode, feeNameStartsWith);

    await goToEditContract(page, unitCode);

    const editContractUrl = page.url();
    logger.debug("editContractUrl", editContractUrl);

    await click(page, tabChargesAndFeesSelector);

    let feeDeleted = await doDeleteFee(page, feeNameStartsWith);

    if (feeDeleted) {
        logger.info("Fee was deleted", unitCode);

        logger.debug("Click the Save button, on the 'Charges & Fees' tab");
        await click(page, "#ctlContractMultiPage_SaveBtn");

        // verification...
    
        logger.debug("Final verification: fee should NOT exist");

        await page.goto(editContractUrl, {waitUntil: 'load'});

        await click(page, tabChargesAndFeesSelector);
    
        let foundAfterSave = await searchFeeNameTableRows(page, feeNameStartsWith, true);
        logger.debug(`${foundAfterSave.length} fees still exist after save`);
        if (foundAfterSave.length == 0) {
            logger.debug("final verification: success");
        }
        else {
            throw(`verification failed after delete`);
        }
    }
    else {
        logger.warn(`Could not delete fee for ${unitCode}`);
    }
}

// Returns true if fees found and deleted, false otherwise
const doDeleteFee = async(page, feeNameStartsWith) => {
    logger.debug("doDeleteFee");
    logger.debug(`Look for fees starting with ${feeNameStartsWith} and delete them`);

    // Assumes we are on the "Charges & Fees" tab
    await page.waitForSelector(tabChargesAndFeesSelector, {visible: true});

    let existingFees = await searchFeeNameTableRows(page, feeNameStartsWith, true);

    logger.debug(`Found ${existingFees.length} fees to delete`);

    if (existingFees.length == 0) {
        logger.debug("Fee not found");
        return false;
    }
    else {
        while (existingFees.length > 0) {
            let tdHandle = existingFees[0];
    
            let tdValueHandle = await tdHandle.getProperty('outerText');
            let tdValue = (await tdValueHandle.jsonValue()).trim();
    
            logger.debug(`Deleting fee ${tdValue}`);
    
            logger.debug("Click the fee table row to launch the edit fee popup");
            const [popup] = await Promise.all([
                new Promise(resolve => page.once('popup', resolve)),
                tdHandle.click()
            ]);
    
            logger.debug("In the edit fee popup, click the 'delete' button, then click 'accept' in the alert dialog");
            await popup.waitForSelector('#btnDelete')
            await Promise.all([
                new Promise(resolve => popup.once('dialog', dialog => {
                    dialog.accept();
                    resolve();
                })),
                popup.click('#btnDelete')
            ]);

            // wait for the popup to go away
            await sleep(5000);
    
            existingFees = await searchFeeNameTableRows(page, feeNameStartsWith, true);
        }

        logger.debug("Done deleting fees");

        return true;
    }
}

const addFee = async(page, unitCode, data, feeNameStartsWith) => {
    logger.debug("addFee", unitCode);

    await goToEditContract(page, unitCode);

    const editContractUrl = page.url();
    logger.debug("editContractUrl", editContractUrl);

    await click(page, tabChargesAndFeesSelector);

    logger.debug(`Delete any existing fees (name starts with "${feeNameStartsWith}")`);

    let feeDeleted = await doDeleteFee(page, feeNameStartsWith);

    if (feeDeleted) {
        logger.debug("Found existing fees and deleted them");

        // Navigate away and come back to the tab, so that page.once will work for the popup
        await click(page, tabUnitAndOwnersSelector);
        await click(page, tabChargesAndFeesSelector);

        // the following doesn't work, because the deleted fee comes back
        // await page.goto(url, {waitUntil: 'load'});
        // await click(page, tabChargesAndFeesSelector);
    }
    else {
        logger.debug("No existing fees");
    }

    logger.debug("Now add the fee. Click the New Contract Charge Template button");
    // TODO: timeout happens between here and the next debug log statement, if there are no fees in the table at this point
    const newContractButtonSelector = "td.tb_Button";
    await page.waitForSelector(newContractButtonSelector);
    const [popup] = await Promise.all([
        new Promise(resolve => page.once('popup', resolve)),
        page.click(newContractButtonSelector)
    ]);

    logger.debug("Fill out fields in the Edit Fee popup");

    await fillOutFormFields(popup, data);

    logger.debug("Click the Save button in the popup");
    await click(popup, "#btnSave");

    // wait for the popup to go away
    await sleep(5000);

    logger.debug("Charges & Fees page should have reloaded. Ensure fee now exists on the Charges & Fees screen");
    let foundAfterAdd = await searchFeeNameTableRows(page, data.name);
    if (foundAfterAdd.length != 1) {
        throw(`incorrect fee count after add: ${foundAfterAdd.length}`);
    }
    else {
        logger.debug("Verified fee now exists, on the Charges & Fees page");
        logger.debug("Click the Save button, on the Charges & Fees page");
        await click(page, "#ctlContractMultiPage_SaveBtn");
    }

    logger.debug("Final verification: fee should exist");

    await page.goto(editContractUrl, {waitUntil: 'load'});

    await click(page, tabChargesAndFeesSelector);

    let foundAfterSave = await searchFeeNameTableRows(page, feeNameStartsWith, true);

    if (foundAfterSave.length == 1) {
        logger.debug("Click the fee table row");
        const [popupVerify] = await Promise.all([
            new Promise(resolve => page.once('popup', resolve)),
            foundAfterSave[0].click()
        ]);

        logger.debug("Verify form fields");
        await verifyFormFields(popupVerify, data);

        logger.debug("Final verification after add: success");
    }
    else {
        throw "verfication failed after add";
    }
}

const fillOutFormFields = async(popup, data) => {
    // Charge Name (Displayed on Owner Statement)
    await enterText(popup, "#txtName", data.name);

    await selectByName(popup, "select#lstCancellationActions", data.uponCancellation);

    await setInputValue(popup, "[name=dpStartDate]", data.effectiveStartDate);
    
    await setInputValue(popup, "[name=dpEndDate]", data.endDate);

    await enterText(popup, "#txtCalculationAmount", data.amount);
    
    await enterText(popup, "#txtMinimumAmount", data.minimumAmount);
    
    await enterText(popup, "#txtMaximumAmount", data.maximumAmount);

    await selectByName(popup, "select#lstCalculationType", data.calculationMethod);

    await selectByName(popup, "select#lstDocumentTemplate", data.journalDocumentTemplate);

    await selectByName(popup, "select#lstPostingPeriod", data.postingPeriod);

    await selectByName(popup, "select#lstReservationType", data.reservationTypes);
}

const verifyFormFields = async(popup, data) => {
    const verifyInputField = async(selector, dataVal) => {
        let fieldVal = await getInputValue(popup, selector);
        if (fieldVal != dataVal) {
            throw `${selector} has incorrect value`;
        }
    }
    await verifyInputField("#txtName", data.name);
    await verifyInputField("[name=dpStartDate]", data.effectiveStartDate);
    await verifyInputField("[name=dpEndDate]", data.endDate);
    await verifyInputField("#txtCalculationAmount", data.amount);
    await verifyInputField("#txtMinimumAmount", data.minimumAmount);
    await verifyInputField("#txtMaximumAmount", data.maximumAmount);
}

const goToEditContract = async(page, unitCode) => {
    logger.debug("goToEditContract", unitCode);

    // Does not throw an exception if page not found, because we get redirected to an error page.
    // However the next call (waitForSelector) will throw an exception.
    await page.goto(`https://turnkeyvr.escapia.com/webagency/Contract/FindContract.aspx`, {waitUntil: 'load'});

    // Enter search term
    await enterText(page, '#txtUnitCode', unitCode);

    // Submit the search form
    await click(page, "#btnSubmit");

    const selector = "td.listData.clickable:nth-child(6)"; // "Unit" column of search results table
    await page.waitForSelector(selector, {visible: true}); // this will time out if no search results
    const searchResults = await findInList(page, selector, unitCode, false, true);

    let foundContract = null;
    if (searchResults.length == 1) {
        foundContract = searchResults[0];
    } else {
        // go through the list of contracts, starting at the bottom
        // find the first that has a contract end date that is in the future
        for (let i = searchResults.length-1; i >=0; i --) {
            const curr = searchResults[i];
            const contractDatesColumnValue = await curr.evaluate(element => element.parentElement.children[3].outerText);
            const datesStr = contractDatesColumnValue.split("-"); // 1/1/2016 - 6/9/2016
            const endDateStr = datesStr.length == 2 ? datesStr[1] : null;
            const isFuture = moment(endDateStr, dateFormat).isAfter(moment())
            if (isFuture) {
                foundContract = curr;
                break;
            }
        }
    }

    if (searchResults.length > 1) {
        logger.info(`Multiple contracts found for unit ${unitCode} - ${searchResults.length}`);
    }

    if (!foundContract) {
        throw `Contract not found for ${unitCode}`;
    }

    await foundContract.click(); // click the last one in the list

    // We should now be on:
    // https://turnkeyvr.escapia.com/webagency/Contract/EditContract.aspx?id=110855&ReturnURL=/webagency/Contract/FindContract.aspx

    // 3 tabs:
    // "#ui-id-1" - Details tab - this loads by default and is not clickable on page load, because it is the current tab
    // "#ui-id-2" - Units & Owners tab - clickable on page load
    // "#ui-id-3" - Charges & Fees tab - clickable on page load

    // Verify unit name on contract
    await click(page,tabUnitAndOwnersSelector);
    let unitTableRows = await searchUnitNameTableRows(page,  unitCode);
    if (unitTableRows.length == 0) {
        throw "edit contract screen loaded for the wrong unit";
    }
    else {
        logger.debug(`Verified that the Edit Contract screen loaded for unit ${unitCode}`);
    }

    // Verify end date on contract is not in the past
    await click(page, tabDetailsSelector);
    const fieldVal = await getInputValue(page, "[name='ctlContractMultiPage$ctl01$dpContractEndDate']");
    const futureEndDate = moment(fieldVal, dateFormat).isAfter(moment());
    if (!futureEndDate) {
        throw `edit contract screen loaded for expired contract with end date ${fieldVal}`;
    }
    else {
        logger.debug(`Verified that the Edit Contract screen loaded for valid end date ${fieldVal}`);
    }
}

const searchUnitNameTableRows = async(page, unitCode) => {
    // Assumes we are on the "Units & Owners" tab
    await page.waitForSelector(tabUnitAndOwnersSelector, {visible: true});
    return await findInList(page, "td.listData.clickable:nth-child(1)", unitCode, false, true);
}

const searchFeeNameTableRows = async(page, feeName, startsWith) => {
    // Assumes we are on the "Charges & Fees" tab
    await page.waitForSelector(tabChargesAndFeesSelector, {visible: true});
    return await findInList(page, "td.listData.clickable:nth-child(2)", feeName, startsWith); // 1st child is a checkbox
}

const findInList = async(page, selector, searchText, startsWith, caseInsensitive) => {
    logger.debug("findInList", selector, searchText, !!startsWith, !!caseInsensitive);

    let tds = await page.$$(selector);
    let found = [];
    for (let i = 0; i < tds.length; i++) {
        let tdHandle = tds[i];
        let tdValueHandle = await tdHandle.getProperty('outerText');
        let tdValue = (await tdValueHandle.jsonValue()).trim();
        if (caseInsensitive) {
            tdValue = tdValue.toLowerCase();
            searchText = searchText.toLowerCase();
        }
        if (tdValue == searchText || (startsWith && tdValue.indexOf(searchText) == 0)) {
            found.push(tdHandle);
        }
    }

    return found;
}

const getInputValue = async(page, selector) => {
    logger.debug("getInputValue", selector);

    await page.waitForSelector(selector, {visible: true});
    let inputHandle = await page.$(selector);
    let inputValueHandle = await inputHandle.getProperty('value');
    let inputValue = (await inputValueHandle.jsonValue()).trim();
    return inputValue;
};

const setInputValue = async(page, selector, value) => {
    logger.debug("setInputValue", selector, value);

    await page.evaluate((selector, value) => {
        document.querySelector(selector).value = value;
    }, selector, value);
}

const selectByName = async(page, selector, names) => {
    if (!Array.isArray(names)) {
        names = [names];
    }

    logger.debug("selectByName", selector, names);

    await page.waitForSelector(selector, {visible: true});
    await page.focus(selector);

    const selectOptionsSelector = `${selector} > option`;
    let options = await getSelectOptions(page, selectOptionsSelector);
    let matchingOptions = names.map(optionName => options.namesAndValues.find(opt => opt.name == optionName));
    let optionValues = matchingOptions.filter(opt => !!opt).map(opt => opt.value);

    const selected = await page.select(selector, ...optionValues);
    if (!selected || selected.length != names.length) {
        throw(`could not select all options for ${selector}`);
    }
}

const enterText = async(page, selector, text) => {
    text = text + ""; // string
    logger.debug("enterText", selector, text);

    await page.waitForSelector(selector, {visible: true});
    await page.focus(selector);
    await page.type(selector, text);
}

const click = async(page, selector) => {
    logger.debug("click", selector);

    await page.waitForSelector(selector, {visible: true});
    await page.focus(selector);
    await page.click(selector, {waitUntil: 'load'});
}