const env = require('@turnkeyvr/env');
const logger = require('@turnkeyvr/logger');
const axios = require('axios');
const util = require('util');
const salesforceGeneral = require('@turnkeyvr/salesforcegeneral');
const { getAirbnbConfirmationCodeFromCaseDescription } = require('./sfUtil');
const { resultsEnum } = require('../util/misc');
const _clientId = env.config.get('airbnb.api.clientId');
const _secret = env.config.get('airbnb.api.secret');
const _clientIdTest = env.config.get('airbnb.api.clientIdTest');
const _secretTest = env.config.get('airbnb.api.secretTest');

const updateSF = util.promisify(salesforceGeneral.update);
//test
// airbnbAccountInfo = {
//     Id: "sfdc record id",
//     access: "normal access token",
//     refresh: "refresh token"
// }
const refreshToken = async (airbnbAccountInfo, isTest) => {
    let postData = {
        refresh_token: airbnbAccountInfo.refresh,
        reset_refresh_token: 0
    };
    try {
        let results = await axios.post('https://api.airbnb.com/v2/oauth2/authorizations?_unwrapped=true', postData, {
            headers: {
                "Authorization":"Basic " + getEncodedAuth(isTest),
                'Content-Type': 'application/json',
                'User-Agent': 'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1.13) Gecko/20080311 Firefox/2.0.0.13'
            }
        });

        // NOTE: this has to be tested in prod
        let updateSFResults = await updateSF('Airbnb_Account__c', {Id: airbnbAccountInfo.Id, airbnb_access_code__c: results.data.access_token});
        
        logger.debug('refreshToken success? : ', results.data);
        return results.data;
    }
    catch (e) {
        logger.error('refreshToken failed: ', e);
    }
};

const checkTokenStatus = async (token, isTest) => {
    try {
        let results = await axios.get('https://api.airbnb.com/v2/oauth2/authorizations/' + token + '?_unwrapped=true', {
            headers: {
                "Authorization":"Basic " + getEncodedAuth(isTest),
                'Content-Type': 'application/json',
                'User-Agent': 'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1.13) Gecko/20080311 Firefox/2.0.0.13'
            }
        });
        logger.debug('checkTokenStatus results: ', results.data);
        return results.data.valid; // todo: improve this
    }
    catch (e) {
        logger.error('checkTokenStatus failed: ', e);
        return false;
    }
}

const getEncodedAuth = (isTest) => {
    let encodedAuth = Buffer.from(_clientId + ":" + _secret).toString('base64');
    if(isTest){
        encodedAuth = Buffer.from(_clientIdTest + ":" + _secretTest).toString('base64');
    }
    return encodedAuth;
}

const getValidCode = async (airbnbAccountInfo, isTest) => {
    try {
        let tokenIsValid = await checkTokenStatus(airbnbAccountInfo.access, false);
        if (tokenIsValid) {
            return airbnbAccountInfo.access;
        }
    }
    catch (e) {
        logger.error('getValidCode failed: ', e);
        return null;
    }

    // if not valid, refresh and return new token
    // results example: 
    // { access_token: 'dibxi4mvby9v32tmbjj77yxc7',
    // expires_at: 1568142782,
    // user_id: 143519492 }
    let results = await refreshToken(airbnbAccountInfo, false);
    logger.debug('is valid after refresh: ', results);
    return results.access_token;
};

const sendMessage = async (airbnbAccountInfo, messageThreadId, message) => {
    try {
        let validAccess = await getValidCode(airbnbAccountInfo, false);

        let postData = {
            'thread_id': messageThreadId,
            'message': message
        };

        let results = await axios.post('https://api.airbnb.com/v2/messages?_format=for_api_partners', postData, {
            headers: {
                "X-Airbnb-API-Key": _clientId,
                "X-Airbnb-Oauth-Token": validAccess,
                'Content-Type': 'application/json',
                'User-Agent': 'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1.13) Gecko/20080311 Firefox/2.0.0.13'
            }
        });

        logger.debug('send message success: ', results.data);
    }
    catch (e) {
        logger.error('send message failed: ', e.message);
        return resultsEnum.FAILURE;
    }
    return resultsEnum.SUCCESS;
};

const getReservation = async (airbnbAccountInfo, airbnbConfirmationCode) => {
    try {
        let validAccess = await getValidCode(airbnbAccountInfo, false);
        let results = await axios.get('https://api.airbnb.com/v2/reservations/' + airbnbConfirmationCode, {
            headers: {
                "X-Airbnb-API-Key": _clientId,
                "X-Airbnb-Oauth-Token": validAccess
            }
        });
        return results.data;
    }
    catch (e) {
        logger.error('getReservation failed: ', e);
    }
    return null;
};

const getAirbnbAccountInfoFromCase = (caseRecord) => {
    let airbnbAccount = null;
    if(!caseRecord.OpportunityCase__r.Rental_Setup__r.air_bnb_Integration_listing_live__c || caseRecord.OpportunityCase__r.Rental_Setup__r.Airbnb_Account__r === null || caseRecord.OpportunityCase__r.Rental_Setup__r.Airbnb_Account__r.airbnb_access_code__c  === null) {
        airbnbAccount = caseRecord.OpportunityCase__r.Rental_Setup__r.Airbnb_Integration_Account__r;
    }
    else {
        airbnbAccount = caseRecord.OpportunityCase__r.Rental_Setup__r.Airbnb_Account__r;
    }

    let accountInfoObj = {
        Id: airbnbAccount.Id,
        access: airbnbAccount.airbnb_access_code__c,
        refresh: airbnbAccount.refresh_token__c
    };

    if (!accountInfoObj.Id || !accountInfoObj.access || !accountInfoObj.refresh) {
        return null;
    }

    return accountInfoObj;
};

const getMessageThreadIdFromReservation = async (caseRecord) => {
    let airbnbAccountInfo = getAirbnbAccountInfoFromCase(caseRecord);
    if (!airbnbAccountInfo) {
        return null;
    }

    let airbnbConfirmationCode = getAirbnbConfirmationCodeFromCaseDescription(caseRecord.Description);
    if (!airbnbConfirmationCode) {
        return null;
    }

    let airbnbReservation = await getReservation(airbnbAccountInfo, airbnbConfirmationCode);
    if (!airbnbReservation || !airbnbReservation.reservation || !airbnbReservation.reservation.thread_id) {
        return null;
    }

    return airbnbReservation.reservation.thread_id;
};

module.exports = {
    sendMessage: sendMessage,
    getAirbnbAccountInfoFromCase: getAirbnbAccountInfoFromCase,
    getMessageThreadIdFromReservation: getMessageThreadIdFromReservation
};