/**
 * Helper to convert escapia's line item object to sf canceled line item object
 */
const moment = require("moment");

const dateFormat = {
    escapia: 'M/D/YYYY',
    sf: 'YYYY-MM-DD'
};

function isEmpty(value) {
    return (value === undefined || value === null || value.length === 0);
}

function formatDate(value) {
    return (value && value.length > 0) ? moment(value, dateFormat.escapia).format(dateFormat.sf) : null;
}

function formatCurrency(value) {
    if (isEmpty(value)) return null;
    var parsed = parseFloat(value.replace(/,/g, ''));
    return (isNaN(parsed)) ? null : parsed;
}

function formatDescription(value) {
    if (isEmpty(value)) return null;
    var updated = value.replace(/,/g, ' ');
    return updated.replace(/["']/, '').trim();
}

function findType(str, arrSearch) {
    var found = arrSearch.find(function (obj) {
        if (!(obj.search instanceof Array)) obj.search = [obj.search];
        var matches = obj.search.filter(function (search) {
            return (str.match(search) !== null);
        });
        return (matches.length === obj.search.length);
    });
    return (!isEmpty(found)) ? found.type : null;
}

function getBookingType(description, amount) {
    var checks = [
        { search: /vendor fee and tax/i, type: 'fee' },
        { search: /tax/i, type: 'tax' },
        { search: /county/i, type: 'tax' },
        { search: /city/i, type: 'tax' },
        { search: /TX - PortA - Sunflower Beach/, type: 'tax' },
        { search: /fee/i, type: 'fee' },
        { search: /clean/i, type: 'fee' },
        { search: /deposit/i, type: 'fee' },
        { search: /vendor/i, type: 'fee' },
        { search: /tablet/i, type: 'fee' },
        { search: /golf/i, type: 'fee' },
        { search: /keep/i, type: 'fee' },
        { search: /TAC - Airbnb/, type: 'fee' },
        { search: /keep/i, type: 'fee' },
        { search: [/from/i, /to/im], type: 'rent' }
    ];
    var type = findType(description, checks);
    if (!isEmpty(type)) return type;
    return (amount > 0) ? 'fee' : 'refund';
}

function getRezType(description, amount) {
    if (amount < 0) return 'payment';
    var checks = [
        { search: /refund/i, type: 'refund' },
        { search: /fee/i, type: 'fee' },
        { search: /charge/i, type: 'fee' },
        { search: /clean/i, type: 'fee' }
    ];
    var type = findType(description, checks);
    return (!isEmpty(type)) ? type : 'refund';
}

function getType(obj, bookingNumber) {
    var description = obj.Description__c;
    if (isEmpty(description)) return null;
    var amount = obj.Amount__c;
    return (!isEmpty(bookingNumber)) ? getBookingType(description, amount) : getRezType(description, amount);
}

/**
 * Maps an escapia line item to sf canceled line item
 * @param escapiaObj - escapia line item obj
 * @param caseId - sf case id
 * @param rezId - sf rez id
 */
exports.mapLineItem = function (escapiaObj, caseId, rezId) {
    const sfObj = {
        Description__c: formatDescription(escapiaObj.DocumentDescription),
        Creation_Date__c: formatDate(escapiaObj.Creation_Date),
        Posting_Date__c: formatDate(escapiaObj.PostingDate),
        Balance__c: formatCurrency(escapiaObj.FolioBalance),
        Amount__c: formatCurrency(escapiaObj.Amount),
    };
    sfObj.Type__c = (sfObj.Amount__c === null) ? 'Unknown' : getType(sfObj, escapiaObj.Booking_Number);
    if (rezId) sfObj.Canceled_Reservation__c = rezId;
    if (caseId) sfObj.Cancel_Case__c = caseId;
    return sfObj;
}
