const logger = require('@turnkeyvr/logger');
const moment = require('moment');
const _ = require('lodash');
const { getSelectOptions } = require('../util/selectOptionsUtil');
const { sleep } = require('../util/misc');
const { retry } = require('../util/retry');

const dateFormat = "M/D/YYYY";
const expiredStatus = '1'; // Expired
const expiredStatusName = "Expired";
const activeStatus = '2'; // Inactive // TODO: Inactive during testing, we'll switch this to Active for rollout
const activeStatusName = "Inactive"; // TODO: Inactive during testing, we'll switch this to Active for rollout

const timeoutMs = 60000; // 60 seconds (default is 30 seconds)

exports.setRateRule = async (page, rateRuleObj) => {
    let rateRuleName = 'TestAutomation - ' + rateRuleObj.date + ' - Min Night: ' + rateRuleObj.minstay
    rateRuleObj.rateRuleName = rateRuleName;

    page.setDefaultNavigationTimeout(timeoutMs);
    page.setDefaultTimeout(timeoutMs);
    
    logger.info("setRateRule:", `${rateRuleObj.Status}: ${rateRuleName}`);

    rateRuleObj.inactiveRentalsThatCannotBeRemoved = [];

    // look for existing rule first, even for Add
    let rateRuleFound = await navigateToExistingRateRule(page, rateRuleName);

    logger.debug("setRateRule:", `rate rule ${rateRuleName} found? ${rateRuleFound}`);

    let ruleRentalNames = [];
    let rentalNamesBeforeSave = [];
    let availRentalNamesBeforeSave = [];
    let cannotRemoveRentalNames = [];

    if ( rateRuleObj.Status === 'Add' ||  rateRuleObj.Status === 'Update') {
        if (!rateRuleFound) {
            if (rateRuleObj.Status === 'Update') {
                logger.warn("setRateRule:", `created new rate rule but status is Update: ${rateRuleName}`);
            }

            logger.debug("setRateRule:", `creating new rate rule ${rateRuleName}...`);
            await createRateRule(page, rateRuleName, rateRuleObj)
            logger.debug("setRateRule:", `created new rate rule ${rateRuleName}`);
        }
        else {
            if (rateRuleObj.Status === 'Add') {
                logger.warn("setRateRule:", `rate rule already exists in Escapia, but status is Add: ${rateRuleName}`);
            }

            // update status if found. this allows us to find an expired rule and update it to active
            // page needs to load and then javascript can take a while to run, so wait until the select box has the expected number of options - 5
            await page.waitForSelector("#lstStatus > option:nth-child(5)", {timeout: 90000});
            await page.select('#lstStatus', activeStatus);
        }

        let cannotRemoveRentals = [];

        // Button should be visible if there are no rentals yet
        // this would be the case for a newly created rule
        // or an existing rule that does not have any rentals yet
        // (which could happen if rentals specified in big query are not active in Escapia)

        if (await page.$('#btnAddBulkAction') !== null) {
            logger.debug("setRateRule:", "click Add Bulk Action button");
            await page.click('#btnAddBulkAction');
        }
        else {
            // if there are rentals, clear them first
            // so that we can get a list of rentals that are still there after clearing all rentals from the rule
            // these will be inactive rentals that cannot be removed from the rule at this stage
            // they seem to persist after save
            // we'll use this list to ensure that the rate rule grid contains all the right rentals
            logger.debug("setRateRule:", "clear all rentals");
            await clearRentals(page);
            logger.debug("setRateRule:", "get existing rentals");
            await sleep(2000); // try to avoid "Cannot find context with specified id" error
            cannotRemoveRentals = await getExistingRentals(page);
        }

        cannotRemoveRentalNames = cannotRemoveRentals.map(r => r.name);
        logger.debug("setRateRule:", "rentals that cannot be removed:", cannotRemoveRentalNames);

        ruleRentalNames = rateRuleObj.rentals.map(r => r.name && r.name + "");
        logger.debug("setRateRule:", "ruleRentalNames:", ruleRentalNames);

        logger.debug("setRateRule:", "setRentals");
        const actualAvailRentalNames = await setRentals(page, ruleRentalNames);
        logger.debug("setRateRule:", "actualAvailRentalNames:", actualAvailRentalNames);

        const updatedRentalNames = await setRateAmounts(page, rateRuleObj);
        logger.debug("setRateRule:", "updatedRentalNames:", updatedRentalNames);

        const rentalsBeforeSave = await getExistingRentals(page);
        rentalNamesBeforeSave = rentalsBeforeSave.map(r => r.name);
        logger.debug("setRateRule:", "rentalNamesBeforeSave:", rentalNamesBeforeSave);
        
        availRentalNamesBeforeSave = _.difference(rentalNamesBeforeSave, cannotRemoveRentalNames);
        logger.debug("setRateRule:", "availRentalNamesBeforeSave:", availRentalNamesBeforeSave);

        const notInRule = _.difference(availRentalNamesBeforeSave, ruleRentalNames);
        if (notInRule.length > 0) {
            logger.warn("setRateRule:", `rental names, before save, that are not in the rule: ${notInRule.join(", ")}`);
            throw `Found ${notInRule.length} rental names, before save, that are not in the rule`;
        }

        // If some of the rentals in the rule are not found in the Add Rentals dialog, or if they fail to get added to the grid:
        // Go ahead and set rate amounts for the rentals we do have, and save.
        // But final verification will fail and this entire method will return an error, so that we can retry later
        const inRuleButNotFound = _.difference(ruleRentalNames, availRentalNamesBeforeSave);
        if (inRuleButNotFound.length > 0) {
            logger.warn("setRateRule:", `rentals not found, before save: ${inRuleButNotFound.join(", ")}`);
        }

        // If not all the rentals that we add finally end up in the grid -
        // that could mean the web app is malfunctioning, or that we have not waited for the page to fully load
        const difference1 = _.difference(actualAvailRentalNames, availRentalNamesBeforeSave);
        if (difference1.length > 0) {
            logger.warn("setRateRule:", "present in actualAvailRentalNames but missing in availRentalNamesBeforeSave:", difference1);
            throw `${difference1.length} rentals added via the Add popup but are not present in the grid`;
        }

        const difference2 = _.difference(availRentalNamesBeforeSave, actualAvailRentalNames);
        if (difference2.length > 0) {
            logger.warn("setRateRule:", "present in availRentalNamesBeforeSave but missing in actualAvailRentalNames:", difference2);
            throw `${difference2.length} rentals in in the grid that were not added via the Add popup`;
        }
    }
    else if ( rateRuleObj.Status === 'Expire' ) {
        logger.debug("setRateRule:", "set to expired");

        if (!rateRuleFound) {
            throw 'did not find rule to expire: ' + rateRuleName;
        }

        const currentStatus = await getStatus(page);
        if (currentStatus == expiredStatus) {
            logger.warn("setRateRule:", "already expired:", rateRuleName);
        }

        await page.waitForSelector('#imgRateDetails')
        await page.click('#imgRateDetails')
        // page needs to load and then javascript can take a while to run, so wait until the select box has the expected number of options - 5
        await page.waitForSelector("#lstStatus > option:nth-child(5)", {timeout: 90000});
        await page.select('#lstStatus', expiredStatus);
    }

    logger.debug("setRateRule:", "ready to save");

    await page.waitForSelector('#btnSave');
    await sleep(5000); // Without the sleep, the rule will sometimes get saved without all the changes made, especially if there are many rentals

    await retry(
        () => Promise.all([
            page.waitForSelector('#btnSave'),
            page.click('#btnSave')
        ]),
        2 // retry twice
    );

    logger.debug("setRateRule:", "saved, now back on rate rules listing page");

    // Verify...

    rateRuleObj.saved = true;

    if (rateRuleObj.Status === 'Expire') {
        await verifyExpiredRule(page, rateRuleName);
    }
    else {
        const savedRentals = await verify(page, rateRuleName, rateRuleObj, availRentalNamesBeforeSave);
        const savedRentalNames = savedRentals.map(r => r.name);
        logger.debug("setRateRule:", "savedRentalNames:", savedRentalNames);

        // If the rentals in the grid before saving don't match those in the grid once we come back to verify after saving, 
        // that could mean the web app is malfunctioning, or that we saved too early
        const difference1 = _.difference(availRentalNamesBeforeSave, savedRentalNames);
        if (difference1.length > 0) {
            logger.warn("setRateRule:", "available rentals did not get saved:", difference1);
            throw `${difference1.length} available rentals did not get saved`;
        }

        const difference2 = _.difference(savedRentalNames, rentalNamesBeforeSave);
        if (difference2.length > 0) {
            logger.warn("setRateRule:", "rentals saved but were not added to the rule:", difference2);
            throw `${difference2.length} rentals saved but were not added to the rule`;
        }

        const difference3 = _.difference(savedRentalNames, ruleRentalNames);
        if (difference3.length > 0) {
            logger.debug("setRateRule:", "rentals saved that are not in the rule definition:", difference3);
            const savedBecauseCannotRemove = _.intersection(difference3, cannotRemoveRentalNames);
            logger.debug("setRateRule:", "savedBecauseCannotRemove:", savedBecauseCannotRemove);
            if (savedBecauseCannotRemove.length > 0) {
                rateRuleObj.inactiveRentalsThatCannotBeRemoved = savedBecauseCannotRemove;
            }
            const savedInError = _.difference(difference3, cannotRemoveRentalNames);
            logger.debug("setRateRule:", "savedInError:", savedInError);
            if (savedInError.length > 0) {
                logger.warn("setRateRule:", "saved in error:", savedInError);
                throw `${savedInError.length} rentals saved in error`;
            }
        }
    }
};

const getStatus = async(page) => {
    const statusSelector = "#lstStatus > option";
    // page needs to load and then javascript can take a while to run, so wait until the select box has the expected number of options - 5
    await page.waitForSelector("#lstStatus > option:nth-child(5)", {timeout: 90000});
    let options = await getSelectOptions(page, statusSelector);
    logger.debug("verifyStatus:", "options:", options);
    let selectedOption = options.namesAndValues.find(opt => opt.selected === true);
    return selectedOption ? selectedOption.name : null;
}

const verifyStatus = async(page, requiredStatus) => {
    let selectedStatusName = await getStatus(page);
    logger.debug("verifyStatus:", "selectedStatusName:", selectedStatusName);
    if (selectedStatusName != requiredStatus) {
        throw `${statusSelector} has incorrect value ${selectedStatusName}. Should be ${requiredStatus}`;
    }
}

const verifyExpiredRule = async(page, rateRuleName) => {
    let rateRuleFound = await navigateToExistingRateRule(page, rateRuleName);

    logger.debug("verifyExpiredRule:", `rate rule ${rateRuleName} found? ${rateRuleFound}`);

    if (!rateRuleFound) {
        throw "Could not find rule for verification";
    }

    await verifyStatus(page, expiredStatusName);
}

async function verify(page, rateRuleName, rateRuleObj) {
    let rateRuleFound = await navigateToExistingRateRule(page, rateRuleName);

    logger.debug("verify:", `rate rule ${rateRuleName} found? ${rateRuleFound}`);

    if (!rateRuleFound) {
        throw "Could not find rule for verification";
    }

    // status...
    await verifyStatus(page, activeStatusName);
    
    // dates, min stay...

    const verifyCheckboxField = async(selector, shouldBe) => {
        const actual = await getCheckboxValue(page, selector);
        logger.debug("verifyCheckboxField:", `${selector} is ${actual}, should be ${shouldBe}`);
        if (actual != shouldBe) {
            throw `${selector} has incorrect value ${actual}. Should be ${shouldBe}`;
        }
    }

    const verifyInputField = async(selector, shouldBe) => {
        const actual = await getInputValue(page, selector);
        logger.debug("verifyInputField:", `${selector} is ${actual}, should be ${shouldBe}`);
        if (actual != shouldBe) {
            throw `${selector} has incorrect value ${actual}. Should be ${shouldBe}`;
        }
    }

    const verifyDateInputField = async(selector, shouldBe) => {
        const actual = await getInputValue(page, selector);
        const actualMoment = moment(actual, dateFormat);
        const shouldBeMoment = moment(shouldBe, dateFormat);
        logger.debug("verifyDateInputField:", `${selector} is ${actualMoment.format(dateFormat)}, should be ${shouldBeMoment.format(dateFormat)}`);
        if (!actualMoment.isSame(shouldBeMoment)) {
            throw `${selector} has incorrect value ${actualMoment.format(dateFormat)}. Should be ${shouldBeMoment.format(dateFormat)}`;
        }
    }

    if (rateRuleObj.minstay != 'SatOnly') {
        await verifyCheckboxField('#chkTravelLengthOfStayIsActive', true);
        await verifyInputField('#txtTravelLengthOfStayFrom', rateRuleObj.minstay);
        await verifyInputField('#txtTravelLengthOfStayTo', "9999");

        await verifyCheckboxField('#chkTravelFirstNightIsActive', false);
        await verifyCheckboxField('#chkTravelLastNightIsActive', false);
        await verifyCheckboxField('#chkTravelFirstNightSaturday', false);
        await verifyCheckboxField('#chkTravelLastNightFriday', false);

    }
    else {
        await verifyCheckboxField('#chkTravelLengthOfStayIsActive', false);
        // don't verify these, because they don't empty out when the checkbox is unchecked
        // await verifyInputField('#txtTravelLengthOfStayFrom', "");
        // await verifyInputField('#txtTravelLengthOfStayTo', "");

        await verifyCheckboxField('#chkTravelFirstNightIsActive', true);
        await verifyCheckboxField('#chkTravelLastNightIsActive', true);
        await verifyCheckboxField('#chkTravelFirstNightSaturday', true);
        await verifyCheckboxField('#chkTravelLastNightFriday', true);
    }

    const dateSelect = moment(rateRuleObj.date).format(dateFormat);

    await verifyDateInputField('input[name="rptTravelDates$ctl01$dpTravelFromDate"]', dateSelect);
    await verifyDateInputField('input[name="rptTravelDates$ctl01$dpTravelToDate"]', dateSelect);

    // rentals..

    const savedRentals = await getExistingRentals(page);

    logger.debug("verify:", `savedRentals: ${savedRentals.length}`);

    const savedRatesByRental = {};
    savedRentals.forEach(r => savedRatesByRental[r.name] = r.rate);

    const incorrect = rateRuleObj.rentals.filter(r => {
        savedRatesByRental[r.name] && savedRatesByRental[r.name] != r.rate
    });

    if (incorrect.length > 0) {
        throw `rentals saved with incorrect rates: ${incorrect}`;
    }

    rateRuleObj.rentals.forEach(r => {
        if (savedRatesByRental[r.name]) {
            delete r.notFound;
        }
        else {
            r.notFound = true;
        }
    });

    return savedRentals;
}

const getInputValue = async(page, selector) => {
    logger.debug("getInputValue", selector);
    await page.waitForSelector(selector);
    const elementHandle = await page.$(selector);
    const valueHandle = await elementHandle.getProperty('value');
    const val = (await valueHandle.jsonValue());
    return val && val.trim();
};

const getCheckboxValue = async(page, selector) => {
    logger.debug("getCheckboxValue", selector);
    await page.waitForSelector(selector);
    const elementHandle = await page.$(selector);
    const valueHandle = await elementHandle.getProperty('checked');
    const val = (await valueHandle.jsonValue());
    return val == true;
};

const navigateToPricePlan = async(page) => {
    logger.debug("navigateToExistingRateRule:", "Go to Manage Pricing Plan page");

    await page.goto('https://turnkeyvr.escapia.com/webagency/Property/Setup/Rates/ManagePricingPlan.aspx', {waitUntil: 'load'});

    logger.debug("navigateToPricePlan:", "Click the first pricing plan in the list. Should be Beyond Pricing");

    const tdBeyondPricing = await page.$$("td.listData.clickable")
    const planName = "Beyond Pricing Plan";
    logger.debug("navigateToPricePlan:", `Verify plan name is ${planName}`);
    const tdValueHandle = await tdBeyondPricing[0].getProperty('outerText');
    const tdValue = (await tdValueHandle.jsonValue()).trim();
    logger.debug("navigateToPricePlan:", `plan name is ${tdValue}`);
    if (tdValue != planName) {
        throw `unexpected plan name ${tdValue}`;
    }
    await tdBeyondPricing[0].click();
}

async function navigateToExistingRateRule(page, rateRuleName) {
    await navigateToPricePlan(page);

    logger.debug("navigateToExistingRateRule:", "Wait for the default search results to load");

    // The search results table takes several seconds to load - "Loading, please wait" displays
    // After that the data is all in browser memory, so searches are fast
    // First, the single cell containing the text "No matching records found" appears for an instant
    // If there are no results, there will be change after that
    // If there are results, the single cell goes away and there will be one or more rows with 5 cells, the first cell containing a checkbox

    // NOTE: The following wait will only work if we already have at least one rule in the plan
    // we can't wait for td:nth-child(1), has to be td:nth-child(3) - because the "No matching records found" always appears before the results appear
    await page.waitForSelector('#RateRuleTable > tbody > tr:nth-child(1) > td:nth-child(3)', {visible: true, timeout:180000});

    logger.debug("navigateToExistingRateRule:", "On the Edit Pricing Plan page, search for rate rule");

    // Look for active, inactive, expired. If we find the rate rule, we'll update the status later, before saving.
    // This avoids a large number of rate rules getting created if we have unexpected issues such as a bug or unexpected data from big query
    await page.waitForSelector('input[placeholder="Search"]', {visible: true})
    // active already checked by default
    await page.click('#chkShowInactiveRR')
    await page.click('#chkShowExpiredRR')
    await page.focus('input[placeholder="Search"]');
    await page.keyboard.press('Backspace');
    await page.keyboard.type(rateRuleName);

    logger.debug("navigateToExistingRateRule:", "Wait for the first result row to show up");

    // NOTE: no network calls here, but sleep needed because javascript needs to run and re-render
    // 500ms seems to work but let's do 5000 to be sure
    // If we don't sleep long enough, the wrong search results are returned, and we'll throw an error
    await sleep(5000);

    const firstRowFirstCellSelector = '#RateRuleTable > tbody > tr:nth-child(1) > td:nth-child(1)';
    await page.waitForSelector(firstRowFirstCellSelector, {visible: true});

    let firstRowFirstCellElem = await page.$(firstRowFirstCellSelector); // contains text or a checkbox
    let firstRowFirstCellText = await page.evaluate(element => element.textContent, firstRowFirstCellElem);
    logger.debug("navigateToExistingRateRule:", "firstRowFirstCellText value:", firstRowFirstCellText);

    if (firstRowFirstCellText == "No matching records found") {
        logger.debug("navigateToExistingRateRule:", "no results");
        return false;
    }

    // look through the results for an exact match
    // if found go to the rate page and verify again
    // for example, search term "TestAutomation - 2020-09-04 - Min Night: 3" will return
    // "TestAutomation - 2020-09-04 - Min Night: 3" and "TestAutomation - 2020-09-04 - Min Night: 30"

    let tableRows = await page.evaluate(() => {
        const trData = [];
        let idx = 0;
        document.querySelectorAll("#RateRuleTable > tbody > tr").forEach(tr => {
            trData.push({
                idx: idx++,
                sequence: tr.cells[1].textContent.trim(),
                name: tr.cells[2].textContent.trim(),
                status: tr.cells[4].textContent.trim(),
            });
        });
        return trData;
    });

    const formatRuleResult = (tableRow) => {
        return `${tableRow.name} [${tableRow.status}]`;
    }

    const formatRules = (rules) => {
        return rules.slice(0, 10).map(r => formatRuleResult(r)).join(", ");
    }; 

    logger.debug("navigateToExistingRateRule: tableRows:", formatRules(tableRows));

    // if any results do not match the search term - this probably means we are not waiting long enough after submitting the search term
    const unexpectedSearchResults = tableRows.filter(r => r.name.indexOf(rateRuleName) == -1);
    if (unexpectedSearchResults.length > 0) {
        logger.warn("navigateToExistingRateRule:", "Results do not match search term: tableRows:", formatRules(tableRows));
        throw `search results do not match the search term (${unexpectedSearchResults.length} results)`;
    }

    const validTableRows = tableRows.filter(r => r.name == rateRuleName);

    if (validTableRows.length == 0) {
        logger.info("navigateToExistingRateRule: Did not find any exact name matches");
    }
    else if (validTableRows.length == 1) {
        logger.debug("navigateToExistingRateRule: Found one rule:", formatRules(validTableRows));
    }
    else {
        logger.warn("navigateToExistingRateRule: Found multiple rules:", formatRules(validTableRows));
    }

    const nonExpired = validTableRows.filter(r => r.status != "Expired");

    if (nonExpired.length > 1) {
        logger.warn("navigateToExistingRateRule:", "Found multiple non-expired rules:", formatRules(nonExpired));
        throw `found multiple non-expired rules: ${nonExpired.length > 1}`;
    }

    const expired = validTableRows.filter(r => r.status == "Expired");
    if (nonExpired.length > 1) {
        logger.warn("navigateToExistingRateRule:", "Found multiple expired rules:", formatRules(expired));
    }

    let foundRule = null;
    if (nonExpired.length > 0) {
        foundRule = nonExpired[nonExpired.length-1];
    }
    else if (expired.length > 0) {
        foundRule = expired[expired.length-1];
    }

    if (foundRule) {
        logger.debug("navigateToExistingRateRule:", `Found a rule in the search results table: ${formatRuleResult(foundRule)}`);

        const rateRuleElem = await page.$('#RateRuleTable > tbody > tr:nth-child(' + (foundRule.idx + 1) + ')');
        await rateRuleElem.click();

        logger.debug("navigateToExistingRateRule:", "Wait for the Open Rate Rule page to open");
        await page.waitForSelector('#txtName', {timeout: 180000})
        const ruleTextElem = await page.$("#txtName");
        logger.debug("navigateToExistingRateRule:", "Now on the Open Rate Rule page");
        text = await page.evaluate(ruleTextElem => ruleTextElem.value, ruleTextElem);
        logger.debug("navigateToExistingRateRule:", `rule name is: ${text}`);

        if (text === rateRuleName) {
            logger.debug("navigateToExistingRateRule:", "Verified rate rule name still matches");
            return true;
        }
        else {
            throw "rate rule name does not match on Open Rate Rule page";
        }
    }
    else {
        logger.debug("navigateToExistingRateRule:", "No exact name match in results");
        return false;
    }
}

// NOTE: could add some waitForSelector calls in this method, if we see issues
async function createRateRule(page, rateRuleName, rateRuleObj) {
    await navigateToPricePlan(page);

    await page.waitForSelector('#butNewRateRule');
    await sleep(2000); // seems like the button click does not always work right away
    await page.click('#butNewRateRule');

    // page needs to load and then javascript can take a while to run, so wait until the select box has the expected number of options - 5
    await page.waitForSelector("#lstStatus > option:nth-child(5)", {timeout: 90000});
    await page.select('#lstStatus', activeStatus);
    
    await page.waitForSelector('#txtName')
    // txtName sometimes ends up empty, resulting in an alert message that we don't currently look for
    // the result is that the page will just hang
    // moving the lstStatus code to before the txtName code probably fixes, but adding sleep just in case
    await sleep(1000);
    logger.debug("createRateRule:", "type in rate rule name");
    await page.focus('#txtName');
    await page.keyboard.type(rateRuleName);

    if ( rateRuleObj.minstay != 'SatOnly' ) {
        const checkbox = await page.$('#chkTravelLengthOfStayIsActive');
        await checkbox.click();
        await sleep(2000);

        await page.evaluate((minstay) => {
            document.querySelector('#txtTravelLengthOfStayFrom').value = minstay
            document.querySelector('#txtTravelLengthOfStayTo').value = 9999
            return true
        }, rateRuleObj.minstay);
    }
    else {
        await page.click('#chkTravelFirstNightIsActive')
        await page.click('#chkTravelLastNightIsActive')
        await sleep(5000)
        await page.click('#chkTravelFirstNightSaturday')
        await page.click('#chkTravelLastNightFriday')
    }

    await page.click('#btnAddTravelDateCondition');
    let dateSelect = moment(rateRuleObj.date).format(dateFormat)

    await page.waitForSelector('#rptTravelDates_ctl01_dpTravelFromDate_Top > input')
    await page.evaluate((dateSelect) => {
        document.querySelector('#rptTravelDates_ctl01_dpTravelFromDate_Top > input').value = dateSelect
        document.querySelector('#rptTravelDates_ctl01_dpTravelToDate_Top > input').value = dateSelect
        return true
    }, dateSelect);

    return
}

async function clearRentals(page) {
    await page.waitForSelector('#btnBulkActionSelectUnits', {visible: true});
    logger.debug("clearRentals:", "selector #btnBulkActionSelectUnits loaded, click the button");

    await sleep(5000); // the button click sometimes hangs without a sleep here

    const [popup] = await Promise.all([
        new Promise(resolve => page.once('popup', resolve)),
        await page.click('#btnBulkActionSelectUnits'),
    ]);

    await popup.waitForSelector('#lstDualList_LeftBox', {visible: true})
    await popup.waitForSelector('#lstDualList_RightBox')

    logger.debug("clearRentals:", "get select options from right box");

    const options = await getSelectOptions(popup, 'select#lstDualList_RightBox > option');
    
    logger.debug("clearRentals:", `clearing ${options.values.length} rentals`);
    
    await popup.select('select#lstDualList_RightBox', ...options.values);
    await popup.click('#lstDualList_MoveLeft');

    logger.debug("clearRentals:", "click Apply button");

    await popup.click('#btnApply');
}

// NOTE: could add some waitForSelector calls in this method, if we see issues
async function setRentals(page, rentalNames) {
    await page.waitForSelector('#btnBulkActionSelectUnits', {visible: true})
    logger.debug("setRentals:", "selector #btnBulkActionSelectUnits loaded, click the button");

    await sleep(5000); // the button click sometimes hangs without a sleep here

    const [popup] = await Promise.all([
        new Promise(resolve => page.once('popup', resolve)),
        await page.click('#btnBulkActionSelectUnits'),
    ]);

    await popup.waitForSelector('#lstDualList_LeftBox', {visible: true})
    await popup.waitForSelector('#lstDualList_RightBox')

    logger.debug("setRentals:", "get select options from right box");
    const options = await getSelectOptions(popup, 'select#lstDualList_RightBox > option' );

    logger.debug("setRentals:", `clearing ${options.values.length} rentals`);

    await popup.select('select#lstDualList_RightBox', ...options.values);

    await popup.click('#lstDualList_MoveLeft');

    // not all rentals that come from data are active in Escapia...
    let actualAvailRentals = [];
    let selectArr = []

    logger.debug("setRentals:", "get select options from left box");
    let newOptions = await getSelectOptions(popup, '#lstDualList_LeftBox > option' );

    for ( let idx=0; idx<newOptions.namesAndValues.length; idx++ ) {
        for ( let nIdx=0; nIdx<rentalNames.length; nIdx++ ) {
            const rentalName = rentalNames[nIdx];
            if ( newOptions.namesAndValues[idx].name.indexOf(`${rentalName} -`) == 0 ) {
                selectArr.push(newOptions.namesAndValues[idx].value)
                actualAvailRentals.push(rentalName);
                break
            }
        }
    }

    logger.debug("setRentals:", `select ${selectArr.length} options in the left box`);
    await popup.select('#lstDualList_LeftBox', ...selectArr);

    logger.debug("setRentals:", "move new options to the right box");
    await popup.click('#lstDualList_MoveRight');

    logger.debug("setRentals:", "click apply button");
    await popup.click('#btnApply');

    return actualAvailRentals
}

async function getExistingRentals(page) {
    logger.debug("getExistingRentals:", "waiting for selector #btnSave");
    // The Save button is the last thing on the page, so if it has loaded then we have all our rental rows loaded too
    await page.waitForSelector('#btnSave', {visible: true, timeout: 180000});
    logger.debug("getExistingRentals:", "selector #btnSave loaded");

    const rentals = await page.evaluate(() => {
        const rentals = [];
        const inputElements = document.querySelectorAll('[id^="rptBulkActions_"]');
        for (let i = 0; i < inputElements.length; i++) {
            const currentInputElement = inputElements[i];
            if (currentInputElement.id.indexOf('_txtName') > -1) {
                const currentRentalName = currentInputElement.value;
                const rateInputElement = inputElements[i + 1];
                if (rateInputElement.id.indexOf('_txtNightlyRate') == -1) {
                    throw `rate input field not found for ${currentRentalName}`;
                }
                rentals.push({
                    name: currentRentalName,
                    rate: rateInputElement.value
                });
            }
        }
        return rentals;
    });

    return rentals;
}

async function setRateAmounts(page, rateRuleObj) {
    logger.debug("setRateAmounts:", "start");

    const ratesByRental = {};
    rateRuleObj.rentals.forEach(r => ratesByRental[r.name] = r.rate);

    // The Save button is the last thing on the page, so if it has loaded then we have all our rental rows loaded too
    await page.waitForSelector('#btnSave', {visible: true, timeout:180000});

    logger.debug("setRateAmounts:", "selector #btnSave loaded, now setting rates");

    const updated = await page.evaluate(({ratesByRental}) => {
        const updated = [];
        const inputElements = document.querySelectorAll('[id^="rptBulkActions_"]');
        for (let i = 0; i < inputElements.length; i++) {
            const currentInputElement = inputElements[i];
            if (currentInputElement.id.indexOf('_txtName') > -1) {
                const currentRentalName = currentInputElement.value;
                const rateInputElement = inputElements[i + 1];
                if (rateInputElement.id.indexOf('_txtNightlyRate') == -1) {
                    throw `rate input field not found for ${currentRentalName}`;
                }
                // if the rental is not specified in the rate rule currently
                // then set its value to empty, so that it is not bookable
                // (the grid can contain "cannotRemove" rentals that were in the rate rule at one time and are now invalid rentals)
                rateInputElement.value = ratesByRental[currentRentalName] || "";
                updated.push(currentRentalName);
            }
        }
        return updated;
    }, {ratesByRental});

    logger.debug("setRateAmounts:", `done, updated ${updated.length} rates in grid`);

    return updated;
}